@@ -9,20 +9,28 @@ setClass("lavData",
         data.type="character",     # "full", "moment" or "none"
         ngroups="integer",         # number of groups
         group="character",         # group variable
+        nlevels="integer",         # number of levels
+        cluster="character",       # cluster variable(s)
         group.label="character",   # group labels
+        level.label="character",   # level labels
         std.ov="logical",          # standardize observed variables?
         nobs="list",               # effective number of observations
         norig="list",              # original number of observations
         ov.names="list",           # variable names (per group)
         ov.names.x="list",         # exo variable names (per group)
+        ov.names.l="list",         # names per level
         #ov.types="list",           # variable types (per group)
         #ov.idx="list",             # column indices (all observed variables)
+        ordered="character",       # ordered variables
+        weights="list",            # sampling weights (per group)
+        sampling.weights="character", # sampling weights variable
         ov="list",                 # variable table
         case.idx="list",           # case indices per group
         missing="character",       # "listwise" or not?
         Mp="list",                 # if not complete, missing patterns
                                    # we need this here, to get nobs right!
         Rp="list",                 # response patterns (categorical only)
+        Lp="list",                 # level patterns
         eXo="list",                # local copy exo only
         X="list"                   # local copy
     )
@@ -31,7 +39,6 @@ setClass("lavData",
 
 setClass("lavSampleStats",         # sample moments
     representation(
-        CAT="list",
         var="list",                # observed variances (per group)
         cov="list",                # observed var/cov matrix (per group)
         mean="list",               # observed mean vector (per group)
@@ -60,7 +67,9 @@ setClass("lavSampleStats",         # sample moments
         cov.log.det="list",        # log det of observed cov (per group)
         res.icov="list",
         res.cov.log.det="list",
-        ridge="numeric",           # ridge constant
+        #ridge.constant="numeric",  # ridge constant (per group)
+        #ridge.constant.x="numeric",# ridge constant (per group) for eXo
+        ridge="numeric",
         WLS.obs="list",            # all relevant observed stats in a vector
         WLS.V="list",              # weight matrix for GLS/WLS
         WLS.VD="list",             # diagonal of weight matrix only
@@ -69,11 +78,15 @@ setClass("lavSampleStats",         # sample moments
 
         missing.flag="logical",    # missing patterns?
         missing="list",            # missingness information
-        missing.h1="list"          # h1 model
+        missing.h1="list",         # h1 model
+
+        YLp = "list",              # cluster/level information
+
+        zero.cell.tables="list"    # bivariate tables with empty cells
     )
 )
 
-setClass("Model",          # MATRIX representation of the sem model
+setClass("lavModel",          # MATRIX representation of the sem model
     representation(
         GLIST="list",              # list of all model matrices (for all groups)
         dimNames="list",           # dim names for the model matrices
@@ -83,11 +96,12 @@ setClass("Model",          # MATRIX representation of the sem model
         representation="character",  # stub, until we define more classes
         meanstructure="logical",
         categorical="logical",
+        multilevel="logical",
         group.w.free="logical",
         link="character",
-        control="list",
 
-        ngroups="integer",
+        nblocks="integer",
+        ngroups="integer",   # only for rsem!! (which uses rsem:::computeDelta)
         nmat="integer",
         nvar="integer",
         num.idx="list",
@@ -135,8 +149,9 @@ setClass("Model",          # MATRIX representation of the sem model
         ov.x.dummy.ov.idx="list",
         ov.x.dummy.lv.idx="list",
         ov.y.dummy.ov.idx="list",
-        ov.y.dummy.lv.idx="list"
+        ov.y.dummy.lv.idx="list",
 
+        estimator="character"
     )
 )
 
@@ -166,6 +181,7 @@ setClass("Fit",
 
 setClass("lavaan",
     representation(
+        version     = "character",       # lavaan version
         call        = "call",            # matched call
         timing      = "list",            # timing information
         Options     = "list",            # lavOptions
@@ -173,14 +189,17 @@ setClass("lavaan",
         pta         = "list",            # parameter table attributes
         Data        = "lavData",         # full data
         SampleStats = "lavSampleStats",  # sample statistics
-        Model       = "Model",           # internal matrix representation
+        Model       = "lavModel",        # internal matrix representation
         Cache       = "list",            # housekeeping stuff
         Fit         = "Fit",             # fitted results 
         boot        = "list",            # bootstrap results
         optim       = "list",            # optimizer results
+        loglik      = "list",            # loglik values and info
         implied     = "list",            # model implied moments
         vcov        = "list",            # vcov
         test        = "list",            # test
+        h1          = "list",            # unrestricted model results
+        baseline    = "list",            # baseline model results
         external    = "list"             # optional slot, for add-on packages
     ) 
 )
@@ -192,7 +211,7 @@ setClass("lavaanList",
         ParTable        = "list",
         pta             = "list",
         Data            = "lavData",  # from first dataset (ngroups!)
-        Model           = "Model",    # based on first dataset
+        Model           = "lavModel", # based on first dataset
         meta            = "list",
 
         timingList      = "list",
