@@ -21,7 +21,7 @@
 ctr_pml_plrt_nested <- function(fit_objH0, fit_objH1) {
 
     # sanity check, perhaps we misordered H0 and H1 in the function call??
-    if(fit_objH1@Fit@test[[1]]$df > fit_objH0@Fit@test[[1]]$df) {
+    if(fit_objH1@test[[1]]$df > fit_objH0@test[[1]]$df) {
         tmp <- fit_objH0
         fit_objH0 <- fit_objH1
         fit_objH1 <- tmp
@@ -35,14 +35,25 @@ ctr_pml_plrt_nested <- function(fit_objH0, fit_objH1) {
     }
   
     nsize <- fit_objH0@SampleStats@ntotal
-    PLRT <- 2 * (fit_objH1@Fit@logl - fit_objH0@Fit@logl)
+    PLRT <- 2 * (fit_objH1@optim$logl - fit_objH0@optim$logl)
 
     # create a new object 'objH1_h0': the object 'H1', but where
     # the parameter values are from H0
     objH1_h0 <- lav_test_diff_m10(m1 = fit_objH1, m0 = fit_objH0, test = FALSE)
   
     # EqMat
-    EqMat <- lav_test_diff_A(m1 = fit_objH1, m0 = fit_objH0)
+    #EqMat <- lav_test_diff_A(m1 = fit_objH1, m0 = fit_objH0)
+    EqMat <- fit_objH0@Model@ceq.JAC
+
+    # DEBUG YR -- eliminate the constraints also present in H1
+    #          -- if we do this, there is no need to use MASS::ginv later
+    #JAC0 <- fit_objH0@Model@ceq.JAC
+    #JAC1 <- fit_objH1@Model@ceq.JAC
+    #unique.idx <- which(apply(JAC0, 1, function(x) { 
+    #                    !any(apply(JAC1, 1, function(y) { all(x == y) })) }))
+    #if(length(unique.idx) > 0L) {
+    #    EqMat <- EqMat[unique.idx,,drop = FALSE]
+    #}
 
     # Observed information (= for PML, this is Hessian / N)
     Hes.theta0 <- lavTech(objH1_h0, "information.observed")
@@ -61,7 +72,8 @@ ctr_pml_plrt_nested <- function(fit_objH0, fit_objH1) {
 
     MInvGtM <- EqMat %*% Inv.G %*% t(EqMat)
     MinvHtM <- EqMat %*% Inv.Hes.theta0 %*% t(EqMat)
-    Inv_MinvHtM <- solve(MinvHtM)
+    #Inv_MinvHtM <- solve(MinvHtM)
+    Inv_MinvHtM <- MASS::ginv(MinvHtM)
     tmp.prod <- MInvGtM %*% Inv_MinvHtM
     tmp.prod2 <- tmp.prod %*% tmp.prod
     sum.eig <- sum(diag(tmp.prod))
@@ -78,7 +90,7 @@ ctr_pml_plrt_nested <- function(fit_objH0, fit_objH1) {
 # for testing: this is the 'original' (using m.el.idx and x.el.idx)
 ctr_pml_plrt_nested2 <- function (fit_objH0, fit_objH1) {
 
-    if (fit_objH1@Fit@test[[1]]$df > fit_objH0@Fit@test[[1]]$df) {
+    if (fit_objH1@test[[1]]$df > fit_objH0@test[[1]]$df) {
         tmp <- fit_objH0
         fit_objH0 <- fit_objH1
         fit_objH1 <- tmp
@@ -90,9 +102,9 @@ ctr_pml_plrt_nested2 <- function (fit_objH0, fit_objH1) {
         equalConstr = FALSE
     }
 
-    nsize <- fit_objH0@Data@nobs[[1]]
-    PLRT <- 2 * nsize * (fit_objH0@Fit@fx - fit_objH1@Fit@fx)
-    Npar <- fit_objH1@Fit@npar
+    nsize <- fit_objH0@SampleStats@ntotal
+    PLRT <- 2 * nsize * (fit_objH0@optim$fx - fit_objH1@optim$fx)
+    Npar <- fit_objH1@optim$npar
     MY.m.el.idx2 <- fit_objH1@Model@m.free.idx
     MY.x.el.idx2 <- fit_objH1@Model@x.free.idx
     MY.m.el.idx <- MY.m.el.idx2
@@ -183,8 +195,8 @@ ctr_pml_plrt_nested2 <- function (fit_objH0, fit_objH1) {
  # an rx1 vector where r are the equality constraints. In the null hypothesis
  # we test H0: g(theta)=0. The matrix of derivatives is of dimension:
  # nrows= number of free non-redundant parameters under H0, namely 
- # NparH0 <- fit_objH0[[1]]@Fit@npar , and ncols= number of free non-redundant
- # parameters under H1, namely NparH1 <- fit_objH0[[1]]@Fit@npar.
+ # NparH0 <- fit_objH0[[1]]@optim$npar , and ncols= number of free non-redundant
+ # parameters under H1, namely NparH1 <- fit_objH0[[1]]@optim$npar.
  # The matrix of derivatives of g(theta) is composed of 0's, 1's, -1's, and 
  # in the rows that refer to odd number of parameters that are equal there is one -2.
  # The 1's, -1's (and possibly -2) are the contrast coefficients of the parameters.
@@ -233,7 +245,8 @@ ctr_pml_plrt_nested2 <- function (fit_objH0, fit_objH1) {
         MY.m.el.idx2 = MY.m.el.idx2, MY.x.el.idx2 = MY.x.el.idx2,
         Npar = Npar, equalConstr = equalConstr)
     Hes.theta0 <- NHes.theta0/nsize
-    Inv.Hes.theta0 <- solve(Hes.theta0)
+    #Inv.Hes.theta0 <- solve(Hes.theta0)
+    Inv.Hes.theta0 <- MASS::ginv(Hes.theta0)
 
     NJ.theta0 <- MYgetVariability(object = obj, MY.m.el.idx = MY.m.el.idx,
         MY.x.el.idx = MY.x.el.idx, equalConstr = equalConstr)
@@ -243,7 +256,8 @@ ctr_pml_plrt_nested2 <- function (fit_objH0, fit_objH1) {
     Inv.G <- Inv.Hes.theta0 %*% J.theta0 %*% Inv.Hes.theta0
     MInvGtM <- EqMat %*% Inv.G %*% t(EqMat)
     MinvHtM <- EqMat %*% Inv.Hes.theta0 %*% t(EqMat)
-    Inv_MinvHtM <- solve(MinvHtM)    #!!! change names
+    #Inv_MinvHtM <- solve(MinvHtM)    #!!! change names
+    Inv_MinvHtM <- MASS::ginv(MinvHtM)
     tmp.prod <- MInvGtM %*% Inv_MinvHtM  #!!! change names
     tmp.prod2 <- tmp.prod %*% tmp.prod
     sum.eig <- sum(diag(tmp.prod))
@@ -392,11 +406,12 @@ MYcomputeGradient <- function (object, GLIST, samplestats = NULL, X = NULL,
         GLIST <- object@GLIST
     }
    Sigma.hat <- computeSigmaHat(object, GLIST = GLIST, extra = (estimator ==  "ML"))
+   Mu.hat <- computeMuHat(object, GLIST = GLIST)
    TH <- computeTH(object, GLIST = GLIST)
    g<-1
-   d1 <- pml_deriv1(Sigma.hat = Sigma.hat[[g]], TH = TH[[g]],
-                             th.idx = th.idx[[g]], num.idx = num.idx[[g]],
-                             X = X[[g]], lavcache = lavcache[[g]])
+   d1 <- pml_deriv1(Sigma.hat = Sigma.hat[[g]], Mu.hat = Mu.hat[[g]], 
+                    TH = TH[[g]], th.idx = th.idx[[g]], num.idx = num.idx[[g]],
+                    X = X[[g]], lavcache = lavcache[[g]])
 
  #!?  if(equalConstr) { #delete the following three commented lines, wrong
  #     Delta <- lavaan:::computeDelta (lavmodel= object, GLIST. = GLIST)
@@ -491,16 +506,17 @@ MYNvcov.first.order <- function (lavmodel, lavsamplestats = NULL,
                                        x.el.idx. = MY.x.el.idx)
   #  }
     Sigma.hat <- computeSigmaHat(lavmodel)
+    Mu.hat <- computeMuHat(lavmodel)
     TH <- computeTH(lavmodel)
     g <-1
 
-    SC <- pml_deriv1(Sigma.hat = Sigma.hat[[g]], TH = TH[[g]],
-                              th.idx = lavmodel@th.idx[[g]], 
-                              num.idx = lavmodel@num.idx[[g]],
-                              X = lavdata@X[[g]], lavcache = lavcache,
-                              scores = TRUE, negative = FALSE)
+    SC <- pml_deriv1(Sigma.hat = Sigma.hat[[g]], TH = TH[[g]], 
+                     Mu.hat = Mu.hat[[g]], th.idx = lavmodel@th.idx[[g]], 
+                     num.idx = lavmodel@num.idx[[g]],
+                     X = lavdata@X[[g]], lavcache = lavcache,
+                     scores = TRUE, negative = FALSE)
     group.SC <- SC %*% Delta[[g]]
-    B0.group[[g]] <- crossprod(group.SC)
+    B0.group[[g]] <- lav_matrix_crossprod(group.SC)
     #!!!! B0.group[[g]] <- B0.group[[g]]/lavsamplestats@ntotal  !!! skip so that the result
     # is in line with the 0.5-18 version of lavaan
 
