@@ -53,11 +53,14 @@ bootstrapLavaan <- function(object,
         stop("lavaan ERROR: this function is not (yet) available if conditional.x = TRUE")
     }
 
+    lavoptions. <- list(parallel = parallel, ncpus = ncpus, cl = cl, 
+                        iseed = iseed)
+
     bootstrap.internal(object          = object,
                        lavdata.        = NULL,
                        lavmodel.       = NULL,
                        lavsamplestats. = NULL,
-                       lavoptions.     = NULL,
+                       lavoptions.     = lavoptions.,
                        lavpartable.    = NULL,
                        R               = R,
                        type            = type.,
@@ -65,10 +68,6 @@ bootstrapLavaan <- function(object,
                        FUN             = FUN,
                        warn            = warn,
                        return.boot     = return.boot,
-                       parallel        = parallel,
-                       ncpus           = ncpus,
-                       cl              = cl,
-                       iseed           = iseed,
                        h0.rmsea        = h0.rmsea,
                        ...)
 }
@@ -87,10 +86,6 @@ bootstrap.internal <- function(object          = NULL,
                                FUN             = "coef",
                                warn            = 0L,
                                return.boot     = FALSE,
-                               parallel        = c("no", "multicore", "snow"),
-                               ncpus           = 1L,
-                               cl              = NULL,
-                               iseed           = NULL,
                                h0.rmsea        = NULL,
                                ...) {
 
@@ -130,13 +125,15 @@ bootstrap.internal <- function(object          = NULL,
         }
     }
 
+    parallel <- lavoptions$parallel
+    ncpus    <- lavoptions$ncpus
+    cl       <- lavoptions$cl
+    iseed    <- lavoptions$iseed
+
     # prepare
     old_options <- options(); options(warn = warn)
-    if (missing(parallel)) parallel <- "no"
     # the next 10 lines are borrowed from the boot package
-    parallel <- match.arg(parallel)
     have_mc <- have_snow <- FALSE
-    ncpus <- ncpus
     if (parallel != "no" && ncpus > 1L) {
         if (parallel == "multicore") have_mc <- .Platform$OS.type != "windows"
         else if (parallel == "snow") have_snow <- TRUE
@@ -277,7 +274,6 @@ bootstrap.internal <- function(object          = NULL,
 
         # verbose
         if(verbose) cat("  ... bootstrap draw number:", sprintf("%4d", b))
-
         bootSampleStats <- try(lav_samplestats_from_data(
                                lavdata       = NULL,
                                DataX         = dataX,
@@ -285,6 +281,7 @@ bootstrap.internal <- function(object          = NULL,
                                DataOv        = lavdata@ov,
                                DataOvnames   = lavdata@ov.names,
                                DataOvnamesx  = lavdata@ov.names.x,
+                               DataWT        = lavdata@weights,
                                missing       = lavoptions$missing,
                                rescale       = (lavoptions$estimator == "ML" &&
                                                 lavoptions$likelihood == "normal"),
@@ -297,7 +294,10 @@ bootstrap.internal <- function(object          = NULL,
                                missing.h1    = TRUE,
                                verbose       = FALSE), silent=TRUE) 
         if(inherits(bootSampleStats, "try-error")) {
-            if(verbose) cat("     FAILED: creating sample statistics\n")
+            if(verbose) {
+                cat("     FAILED: creating sample statistics\n")
+                cat(bootSampleStats[1])
+            }
             options(old_options)
             return(NULL)
         }
@@ -356,6 +356,9 @@ bootstrap.internal <- function(object          = NULL,
 
     # this is from the boot function in package boot
     RR <- R
+    if(verbose) {
+        cat("\n")
+    }
     res <- if (ncpus > 1L && (have_mc || have_snow)) {
         if (have_mc) {
             parallel::mclapply(seq_len(RR), fn, mc.cores = ncpus)
