@@ -3,7 +3,7 @@
 #
 # YR 17 Sept 2013
 # 
-# - YR 26 Nov 2013: big change - make it wrapper around lavaan()
+# - YR 26 Nov 2013: big change - make it a wrapper around lavaan()
 #                   estimator = "none" means two.step (starting values)
 
 lavCor <- function(object, 
@@ -19,6 +19,12 @@ lavCor <- function(object,
                    ...,
                    output = "cor") {
 
+    # shortcut if object = lavaan object
+    if(inherits(object, "lavaan")) {
+        out <- lav_cor_output(object, output = output)
+        return(out)
+    }
+
     # check estimator
     estimator <- tolower(estimator)
     if(estimator %in% c("two.step", "two.stage")) {
@@ -35,14 +41,7 @@ lavCor <- function(object,
     }
 
     # check object class
-    if(inherits(object, "lavaan")) {
-        lav.data <- object@Data
-        if(missing(missing)) {
-            missing <- object@Options$missing
-        } else {
-            missing <- "default"
-        }
-    } else if(inherits(object, "lavData")) {
+    if(inherits(object, "lavData")) {
         lav.data <- object
     } else if(inherits(object, "data.frame")) {
         NAMES <- names(object)
@@ -52,7 +51,7 @@ lavCor <- function(object,
         lav.data <- lavData(data = object, group = group, 
                             ov.names = NAMES, ordered = ordered,
                             ov.names.x = ov.names.x,
-                            missing = missing)
+                            lavoptions = list(missing = missing))
     } else {
         stop("lavaan ERROR: lavCor can not handle objects of class ",
              paste(class(object), collapse= " "))
@@ -70,9 +69,15 @@ lavCor <- function(object,
 
     # extract partable options from dots
     dots <- list(...)
-    meanstructure <- FALSE; fixed.x <- FALSE; mimic <- "lavaan"
+    meanstructure <- FALSE 
+    fixed.x       <- FALSE
+    mimic         <- "lavaan"
+    conditional.x <- FALSE
     if(!is.null(dots$meanstructure)) {
         meanstructure <- dots$meanstructure
+    } 
+    if(categorical) {
+        meanstructure <- TRUE
     }
     if(!is.null(dots$fixed.x)) {
         fixed.x <- dots$fixed.x
@@ -80,52 +85,80 @@ lavCor <- function(object,
     if(!is.null(dots$mimic)) {
         mimic <- dots$mimic
     }
+    if(!is.null(dots$conditional.x)) {
+        conditional.x <- dots$conditional.x
+    }
 
     # generate partable for unrestricted model
     PT.un <- 
         lav_partable_unrestricted(lavobject     = NULL,
                                   lavdata       = lav.data,
                                   lavoptions    = list(meanstructure = meanstructure,
-                                                       fixed.x = fixed.x, 
+                                                       fixed.x = fixed.x,
+                                                       conditional.x = conditional.x,
                                                        mimic = mimic),
                                   sample.cov    = NULL,
                                   sample.mean   = NULL,
                                   sample.th     = NULL)
 
     
-    fit <- lavaan(slotParTable = PT.un, slotData = lav.data,
+    FIT <- lavaan(slotParTable = PT.un, slotData = lav.data,
                   model.type = "unrestricted",
                   missing = missing,
                   se = se, estimator = estimator, ...)
 
+    out <- lav_cor_output(FIT, output = output)
+
+    out
+}
+
+lav_cor_output <- function(object, output = "cor") {
+
     # check output
     if(output %in% c("cor","cov")) {
-        out <- inspect(fit, "sampstat")
-        if(fit@Data@ngroups == 1L) {
-            out <- out$cov
+        out <- inspect(object, "sampstat")
+        if(object@Data@ngroups == 1L) {
+            if(object@Model@conditional.x) {
+                out <- out$res.cov
+            } else {
+                out <- out$cov
+            }
             if(output == "cor") {
                 out <- cov2cor(out)
             }
         } else {
-            out <- lapply(out, "[[", "cov")
+            if(object@Model@conditional.x) {
+                out <- lapply(out, "[[", "res.cov")
+            } else {
+                out <- lapply(out, "[[", "cov")
+            }
             if(output == "cor") {
                 out <- lapply(out, cov2cor)
             }
         }
     } else if(output %in% c("th","thresholds")) {
-        out <- inspect(fit, "sampstat")
-        if(fit@Data@ngroups == 1L) {
-            out <- out$th
+        out <- inspect(object, "sampstat")
+        if(object@Data@ngroups == 1L) {
+            if(object@Model@conditional.x) {
+                out <- out$res.th
+            } else {
+                out <- out$th
+            }
         } else {
-            out <- lapply(out, "[[", "th")
+            if(object@Model@conditional.x) {
+                out <- lapply(out, "[[", "res.th")
+            } else {
+                out <- lapply(out, "[[", "th")
+            }
         }
     } else if(output %in% c("sampstat")) {
-        out <- inspect(fit, "sampstat")
+        out <- inspect(object, "sampstat")
     } else if(output %in% c("parameterEstimates", "pe", 
               "parameterestimates", "est")) {
-        out <- parameterEstimates(fit)
+        #out <- parameterEstimates(object)
+        out <- standardizedSolution(object)
     } else {
-        out <- fit
+        out <- object
     }
 
     out
