@@ -9,18 +9,37 @@
 # construct MATRIX lavoptions$representation of the model
 lav_model <- function(lavpartable      = NULL,
                       lavoptions       = NULL,
-                      th.idx           = list()) {
+                      th.idx           = list(),   
+                      cov.x            = list(),
+                      mean.x           = list()) { # for conditional.x only
+                                                   # (not really needed,
+                                                   #  but as a failsafe)
 
     # handle bare-minimum partables
     lavpartable <- lav_partable_complete(lavpartable)
  
     # global info from user model
     nblocks <- lav_partable_nblocks(lavpartable)
+    ngroups <- lav_partable_ngroups(lavpartable)
     meanstructure <- any(lavpartable$op == "~1")
     categorical <- any(lavpartable$op == "|")
-    if(categorical) meanstructure <- TRUE
-    group.w.free <- any(lavpartable$lhs == "group" & lavpartable$op == "%")
+    if(categorical) { 
+        meanstructure <- TRUE
 
+        # handle th.idx if length(th.idx) != nblocks
+        if(nblocks != length(th.idx)) {
+            th.idx <- rep(th.idx, each = nblocks)
+        }
+        
+    }
+    group.w.free <- any(lavpartable$lhs == "group" & lavpartable$op == "%")
+    multilevel <- FALSE
+    if(!is.null(lavpartable$level)) {
+        nlevels <- lav_partable_nlevels(lavpartable)
+        if(nlevels > 1L) {
+            multilevel <- TRUE
+        }
+    }
 
     # handle variable definitions and (in)equality constraints
     CON <- lav_constraints_parse(partable = lavpartable,
@@ -40,7 +59,7 @@ lav_model <- function(lavpartable      = NULL,
 
     # select model matrices
     if(lavoptions$representation == "LISREL") {
-        REP <- representation.LISREL(lavpartable, target=NULL, extra=TRUE)
+        REP <- representation.LISREL(lavpartable, target = NULL, extra = TRUE)
     } else {
         stop("lavaan ERROR: only representation \"LISREL\" has been implemented.")
     }
@@ -90,10 +109,11 @@ lav_model <- function(lavpartable      = NULL,
         ov.num <-       lav_partable_vnames(lavpartable, "ov.num", block = g)
         if(lavoptions$conditional.x) {
             nvar[g] <- length(ov.names.nox)
+            num.idx[[g]] <- which(ov.names.nox %in% ov.num)
         } else {
             nvar[g] <- length(ov.names)
+            num.idx[[g]] <- which(ov.names %in% ov.num)
         }
-        num.idx[[g]] <- match(ov.num, ov.names.nox)
 
         # model matrices for this block
         mmNumber    <- attr(REP, "mmNumber")[[g]]
@@ -172,6 +192,19 @@ lav_model <- function(lavpartable      = NULL,
                 T <- t(tmp); tmp[lower.tri(tmp)] <- T[lower.tri(T)]
             }
 
+            # 4b. override with cov.x (if conditional.x = TRUE)
+            # new in 0.6-1
+            # shouldn't be needed, if lavpartable$start contains cov.x values
+            if(mmNames[mm] == "cov.x") {
+                tmp <- cov.x[[g]]
+            }
+            # 4c. override with mean.x (if conditional.x = TRUE)
+            # new in 0.6-1
+            # shouldn't be needed, if lavpartable$start contains mean.x values
+            if(mmNames[mm] == "mean.x") {
+                tmp <- as.matrix(mean.x[[g]])
+            }
+
             # representation specific stuff
             if(lavoptions$representation == "LISREL" && mmNames[mm] == "lambda") { 
                 ov.dummy.names.nox <- attr(REP, "ov.dummy.names.nox")[[g]]
@@ -200,7 +233,7 @@ lav_model <- function(lavpartable      = NULL,
                 idx <- which(tmp[,1L] == 0.0)
                 tmp[idx,1L] <- 1.0    
             }
-            
+
             # assign matrix to GLIST
             GLIST[[offset]] <- tmp
         } # mm
@@ -212,6 +245,21 @@ lav_model <- function(lavpartable      = NULL,
     #    fixed.x <- TRUE
     #}
 
+    # dirty hack to mimic MUML
+    if(!is.null(lavoptions$tech.muml.scale)) {
+        warning("lavaan WARNING: using muml scale in group 2")
+
+        # find matrix
+        lambda.idx <- which(names(GLIST) == "lambda")[2L]
+
+        # find rows/cols
+        B.names <- paste0("b", ov.names) ## ad-hoc assumption!!!
+        COLS <-  match(B.names, LV.names)
+        ROWS <- seq_len(nvar[2])
+        stopifnot(length(COLS) == length(ROWS))
+        GLIST[[ lambda.idx ]][ cbind(ROWS, COLS) ] <- lavoptions$tech.muml.scale
+    }
+
 
     Model <- new("lavModel",
                  GLIST=GLIST,
@@ -221,9 +269,10 @@ lav_model <- function(lavpartable      = NULL,
                  representation=lavoptions$representation,
                  meanstructure=meanstructure,
                  categorical=categorical,
+                 multilevel=multilevel,
                  link=lavoptions$link,
                  nblocks=nblocks,
-                 ngroups=nblocks, # for rsem!!!
+                 ngroups=ngroups, # breaks rsem????
                  group.w.free=group.w.free,
                  nmat=nmat,
                  nvar=nvar,
