@@ -1,4 +1,3 @@
-## FIXME: allow for GAMMA in continuous case
 computeSigmaHat <- function(lavmodel = NULL, GLIST = NULL, extra = FALSE, 
                             delta = TRUE, debug = FALSE) {
 
@@ -7,14 +6,13 @@ computeSigmaHat <- function(lavmodel = NULL, GLIST = NULL, extra = FALSE,
 
     nmat           <- lavmodel@nmat
     nvar           <- lavmodel@nvar
-    ngroups        <- lavmodel@ngroups
+    nblocks        <- lavmodel@nblocks
     representation <- lavmodel@representation
-    categorical    <- lavmodel@categorical
 
     # return a list
-    Sigma.hat <- vector("list", length=ngroups)
+    Sigma.hat <- vector("list", length=nblocks)
 
-    for(g in 1:ngroups) {
+    for(g in 1:nblocks) {
 
         # which mm belong to group g?
         mm.in.group <- 1:nmat[g] + cumsum(c(0,nmat))[g]
@@ -48,26 +46,97 @@ computeSigmaHat <- function(lavmodel = NULL, GLIST = NULL, extra = FALSE,
                 attr(Sigma.hat[[g]], "log.det") <- Sigma.hat.log.det
             }
         }
-    } # ngroups
+    } # nblocks
+
+    Sigma.hat
+}
+
+## only if conditional.x = TRUE
+## compute the (larger) unconditional 'joint' covariance matrix (y,x)
+##
+## Sigma (Joint ) = [ (S11, S12), 
+##                    (S21, S22) ] where
+##     S11 = Sigma.res + PI %*% cov.x %*% t(PI) 
+##     S12 = PI %*% cov.x
+##     S21 = cov.x %*% t(PI)
+##     S22 = cov.x
+computeSigmaHatJoint <- function(lavmodel = NULL, GLIST = NULL, extra = FALSE,
+                                 delta = TRUE, debug = FALSE) {
+
+    stopifnot(lavmodel@conditional.x)
+
+    # state or final?
+    if(is.null(GLIST)) GLIST <- lavmodel@GLIST
+
+    nmat           <- lavmodel@nmat
+    nvar           <- lavmodel@nvar
+    nblocks        <- lavmodel@nblocks
+    representation <- lavmodel@representation
+
+    # return a list
+    Sigma.hat <- vector("list", length=nblocks)
+
+    for(g in 1:nblocks) {
+
+        # which mm belong to group g?
+        mm.in.group <- 1:nmat[g] + cumsum(c(0,nmat))[g]
+        MLIST <- GLIST[mm.in.group]
+
+        if(representation == "LISREL") {
+            res.Sigma  <- computeSigmaHat.LISREL(MLIST = MLIST, delta = delta)
+            res.int    <- computeMuHat.LISREL(MLIST = MLIST)
+            res.slopes <- computePI.LISREL(MLIST = MLIST)
+            S.xx       <- MLIST$cov.x
+
+            S.yy <- res.Sigma + res.slopes %*% S.xx %*% t(res.slopes)
+            S.yx <- res.slopes %*% S.xx
+            S.xy <- S.xx %*% t(res.slopes)
+
+            Sigma.hat[[g]] <- rbind( cbind(S.yy, S.yx), cbind(S.xy, S.xx) )
+        } else {
+            stop("only representation LISREL has been implemented for now")
+        }
+        if(debug) print(Sigma.hat[[g]])
+
+        if(extra) {
+            # check if matrix is positive definite
+            ev <- eigen(Sigma.hat[[g]], symmetric=TRUE, only.values=TRUE)$values
+            if(any(ev < .Machine$double.eps) || sum(ev) == 0) {
+                Sigma.hat.inv <-  MASS::ginv(Sigma.hat[[g]])
+                Sigma.hat.log.det <- log(.Machine$double.eps)
+                attr(Sigma.hat[[g]], "po") <- FALSE
+                attr(Sigma.hat[[g]], "inv") <- Sigma.hat.inv
+                attr(Sigma.hat[[g]], "log.det") <- Sigma.hat.log.det
+            } else {
+                ## FIXME
+                ## since we already do an 'eigen' decomposition, we should
+                ## 'reuse' that information, instead of doing a new cholesky
+                Sigma.hat.inv <-  inv.chol(Sigma.hat[[g]], logdet=TRUE)
+                Sigma.hat.log.det <- attr(Sigma.hat.inv, "logdet")
+                attr(Sigma.hat[[g]], "po") <- TRUE
+                attr(Sigma.hat[[g]], "inv") <- Sigma.hat.inv
+                attr(Sigma.hat[[g]], "log.det") <- Sigma.hat.log.det
+            }
+        }
+    } # nblocks
 
     Sigma.hat
 }
 
-### FIXME:: replace by computeEY (to allow for GAMMA in continuous case)?
 computeMuHat <- function(lavmodel = NULL, GLIST = NULL) {
 
     # state or final?
     if(is.null(GLIST)) GLIST <- lavmodel@GLIST
 
     nmat           <- lavmodel@nmat
-    ngroups        <- lavmodel@ngroups
+    nblocks        <- lavmodel@nblocks
     representation <- lavmodel@representation
     meanstructure  <- lavmodel@meanstructure
 
     # return a list
-    Mu.hat <- vector("list", length=ngroups)
+    Mu.hat <- vector("list", length=nblocks)
 
-    for(g in 1:ngroups) {
+    for(g in 1:nblocks) {
 
         # which mm belong to group g?
         mm.in.group <- 1:nmat[g] + cumsum(c(0,nmat))[g]
@@ -80,7 +149,50 @@ computeMuHat <- function(lavmodel = NULL, GLIST = NULL) {
         } else {
             stop("only representation LISREL has been implemented for now")
         }
-    } # ngroups
+    } # nblocks
+
+    Mu.hat
+}
+
+## only if conditional.x = TRUE
+## compute the (larger) unconditional 'joint' mean vector (y,x)
+##
+## Mu (Joint ) = [ Mu.y, Mu.x ] where
+##     Mu.y = res.int + PI %*% M.x
+##     Mu.x = M.x
+computeMuHatJoint <- function(lavmodel = NULL, GLIST = NULL) {
+
+    # state or final?
+    if(is.null(GLIST)) GLIST <- lavmodel@GLIST
+
+    nmat           <- lavmodel@nmat
+    nblocks        <- lavmodel@nblocks
+    representation <- lavmodel@representation
+    meanstructure  <- lavmodel@meanstructure
+
+    # return a list
+    Mu.hat <- vector("list", length=nblocks)
+
+    for(g in 1:nblocks) {
+
+        # which mm belong to group g?
+        mm.in.group <- 1:nmat[g] + cumsum(c(0,nmat))[g]
+
+        if(!meanstructure) {
+            Mu.hat[[g]] <- numeric( lavmodel@nvar[g] )
+        } else if(representation == "LISREL") {
+            MLIST <- GLIST[ mm.in.group ]
+            res.int    <- computeMuHat.LISREL(MLIST = MLIST)
+            res.slopes <- computePI.LISREL(MLIST = MLIST)
+            M.x        <- MLIST$mean.x
+
+            Mu.y <- res.int + res.slopes %*% M.x
+            Mu.x <- M.x
+            Mu.hat[[g]] <- c(Mu.y, Mu.x)
+        } else {
+            stop("only representation LISREL has been implemented for now")
+        }
+    } # nblocks
 
     Mu.hat
 }
@@ -92,16 +204,16 @@ computeTH <- function(lavmodel = NULL, GLIST = NULL) {
     # state or final?
     if(is.null(GLIST)) GLIST <- lavmodel@GLIST
 
-    ngroups <- lavmodel@ngroups
+    nblocks <- lavmodel@nblocks
     nmat    <- lavmodel@nmat
     representation <- lavmodel@representation
     th.idx  <- lavmodel@th.idx
 
     # return a list
-    TH <- vector("list", length=ngroups)
+    TH <- vector("list", length=nblocks)
 
     # compute TH for each group
-    for(g in 1:ngroups) {
+    for(g in 1:nblocks) {
 
         if(length(th.idx[[g]]) == 0) {
             TH[[g]] <- numeric(0L)
@@ -129,21 +241,21 @@ computePI <- function(lavmodel = NULL, GLIST = NULL) {
     # state or final?
     if(is.null(GLIST)) GLIST <- lavmodel@GLIST
 
-    ngroups        <- lavmodel@ngroups
+    nblocks        <- lavmodel@nblocks
     nmat           <- lavmodel@nmat
     representation <- lavmodel@representation
-    fixed.x        <- lavmodel@fixed.x
+    conditional.x  <- lavmodel@conditional.x
 
     # return a list
-    PI <- vector("list", length=ngroups)
+    PI <- vector("list", length=nblocks)
 
     # compute TH for each group
-    for(g in 1:ngroups) {
+    for(g in 1:nblocks) {
         # which mm belong to group g?
         mm.in.group <- 1:nmat[g] + cumsum(c(0,nmat))[g]
         MLIST <- GLIST[ mm.in.group ]
 
-        if(!fixed.x) {
+        if(!conditional.x) {
             PI.g <- numeric( lavmodel@nvar[g] )
         } else
         if(representation == "LISREL") {
@@ -165,16 +277,16 @@ computeGW <- function(lavmodel = NULL, GLIST=NULL) {
     # state or final?
     if(is.null(GLIST)) GLIST <- lavmodel@GLIST
 
-    ngroups        <- lavmodel@ngroups
+    nblocks        <- lavmodel@nblocks
     nmat           <- lavmodel@nmat
     representation <- lavmodel@representation
     group.w.free   <- lavmodel@group.w.free
 
     # return a list
-    GW <- vector("list", length=ngroups)
+    GW <- vector("list", length=nblocks)
 
     # compute GW for each group
-    for(g in 1:ngroups) {
+    for(g in 1:nblocks) {
         # which mm belong to group g?
         mm.in.group <- 1:nmat[g] + cumsum(c(0,nmat))[g]
         MLIST <- GLIST[ mm.in.group ]
@@ -194,78 +306,70 @@ computeGW <- function(lavmodel = NULL, GLIST=NULL) {
     # transform to proportions
     #gw <- unlist(GW)
     #gw <- exp(gw) / sum(exp(gw))
-    #for(g in 1:ngroups) {
+    #for(g in 1:nblocks) {
     #    GW[[g]] <- gw[g]
     #}
 
     GW
 }
 
-# unconditional variances of Y
-#  - same as diag(Sigma.hat) if all Y are continuous)
-#  - 1.0 (or delta^2) if categorical
+# *unconditional* variance/covariance matrix of Y
+#  - same as Sigma.hat if all Y are continuous)
 #  - if also Gamma, cov.x is used (only if categorical)
-computeVY <- function(lavmodel = NULL, GLIST = NULL, lavsamplestats = NULL,
-                      samplestats = NULL) {
+computeVY <- function(lavmodel = NULL, GLIST = NULL, diagonal.only = FALSE) {
     # state or final?
     if(is.null(GLIST)) GLIST <- lavmodel@GLIST
 
-    # backwards compatibility (semTools!)
-    if(!is.null(samplestats)) {
-        lavsamplestats <- samplestats
-    }
-
-    ngroups        <- lavmodel@ngroups
+    nblocks        <- lavmodel@nblocks
     nmat           <- lavmodel@nmat
     representation <- lavmodel@representation
 
     # return a list
-    VY <- vector("list", length=ngroups)
+    VY <- vector("list", length=nblocks)
 
     # compute TH for each group
-    for(g in 1:ngroups) {
+    for(g in 1:nblocks) {
         # which mm belong to group g?
         mm.in.group <- 1:nmat[g] + cumsum(c(0,nmat))[g]
         MLIST <- GLIST[ mm.in.group ]
 
-        cov.x <- lavsamplestats@cov.x[[g]]
-
         if(representation == "LISREL") {
-            VY.g <- computeVY.LISREL(MLIST = MLIST, cov.x = cov.x)
+            VY.g <- computeVY.LISREL(MLIST = MLIST)
         } else {
             stop("only representation LISREL has been implemented for now")
         }
 
-        VY[[g]] <- VY.g
+        if(diagonal.only) {
+            VY[[g]] <- diag(VY.g)
+        } else {
+            VY[[g]] <- VY.g
+        }
     }
 
     VY
 }
 
 # V(ETA): latent variances variances/covariances
-# - if conditional=TRUE, ignore all ov.dummy. * and GAMMA
 computeVETA <- function(lavmodel = NULL, GLIST = NULL, 
-                        remove.dummy.lv = FALSE, lavsamplestats = NULL) {
+                        remove.dummy.lv = FALSE) {
     # state or final?
     if(is.null(GLIST)) GLIST <- lavmodel@GLIST
 
-    ngroups        <- lavmodel@ngroups
+    nblocks        <- lavmodel@nblocks
     nmat           <- lavmodel@nmat
     representation <- lavmodel@representation
 
     # return a list
-    ETA <- vector("list", length=ngroups)
+    ETA <- vector("list", length=nblocks)
 
     # compute ETA for each group
-    for(g in 1:ngroups) {
+    for(g in 1:nblocks) {
         # which mm belong to group g?
         mm.in.group <- 1:nmat[g] + cumsum(c(0,nmat))[g]
         MLIST <- GLIST[ mm.in.group ]
 
-        cov.x <- lavsamplestats@cov.x[[g]]
-       
         if(representation == "LISREL") {
-            ETA.g <- computeVETA.LISREL(MLIST = MLIST, cov.x = cov.x)
+            ETA.g <- computeVETA.LISREL(MLIST = MLIST)
 
             if(remove.dummy.lv) {
                 # remove all dummy latent variables
@@ -291,15 +395,15 @@ computeVETAx <- function(lavmodel = NULL, GLIST = NULL) {
     # state or final?
     if(is.null(GLIST)) GLIST <- lavmodel@GLIST
 
-    ngroups        <- lavmodel@ngroups
+    nblocks        <- lavmodel@nblocks
     nmat           <- lavmodel@nmat
     representation <- lavmodel@representation
 
     # return a list
-    ETA <- vector("list", length=ngroups)
+    ETA <- vector("list", length=nblocks)
 
     # compute ETA for each group
-    for(g in 1:ngroups) {
+    for(g in 1:nblocks) {
         # which mm belong to group g?
         mm.in.group <- 1:nmat[g] + cumsum(c(0,nmat))[g]
         MLIST <- GLIST[ mm.in.group ]
@@ -320,29 +424,27 @@ computeVETAx <- function(lavmodel = NULL, GLIST = NULL) {
 }
 
 # COV: observed+latent variances variances/covariances
-computeCOV <- function(lavmodel = NULL, GLIST = NULL, lavsamplestats = NULL,
-    remove.dummy.lv = FALSE) {
+computeCOV <- function(lavmodel = NULL, GLIST = NULL, 
+                       remove.dummy.lv = FALSE) {
 
     # state or final?
     if(is.null(GLIST)) GLIST <- lavmodel@GLIST
 
-    ngroups        <- lavmodel@ngroups
+    nblocks        <- lavmodel@nblocks
     nmat           <- lavmodel@nmat
     representation <- lavmodel@representation
 
     # return a list
-    COV <- vector("list", length=ngroups)
+    COV <- vector("list", length=nblocks)
 
     # compute COV for each group
-    for(g in 1:ngroups) {
+    for(g in 1:nblocks) {
         # which mm belong to group g?
         mm.in.group <- 1:nmat[g] + cumsum(c(0,nmat))[g]
         MLIST <- GLIST[ mm.in.group ]
 
-        cov.x <- lavsamplestats@cov.x[[g]]
-
         if(representation == "LISREL") {
-            COV.g <- computeCOV.LISREL(MLIST = MLIST, cov.x = cov.x)
+            COV.g <- computeCOV.LISREL(MLIST = MLIST)
 
             if(remove.dummy.lv) {
                 # remove all dummy latent variables
@@ -374,15 +476,15 @@ computeEETA <- function(lavmodel = NULL, GLIST = NULL, lavsamplestats = NULL,
     # state or final?
     if(is.null(GLIST)) GLIST <- lavmodel@GLIST
 
-    ngroups        <- lavmodel@ngroups
+    nblocks        <- lavmodel@nblocks
     nmat           <- lavmodel@nmat
     representation <- lavmodel@representation
 
     # return a list
-    EETA <- vector("list", length=ngroups)
+    EETA <- vector("list", length=nblocks)
 
     # compute E(ETA) for each group
-    for(g in 1:ngroups) {
+    for(g in 1:nblocks) {
         # which mm belong to group g?
         mm.in.group <- 1:nmat[g] + cumsum(c(0,nmat))[g]
         MLIST <- GLIST[ mm.in.group ]
@@ -416,20 +518,20 @@ computeEETA <- function(lavmodel = NULL, GLIST = NULL, lavsamplestats = NULL,
 # E(ETA|x_i): conditional expectation (means) of latent variables
 # for a given value of x_i (instead of E(x_i))
 computeEETAx <- function(lavmodel = NULL, GLIST = NULL, lavsamplestats = NULL, 
-                         eXo = NULL, remove.dummy.lv = FALSE) {
+                         eXo = NULL, nobs = NULL, remove.dummy.lv = FALSE) {
   
     # state or final?
     if(is.null(GLIST)) GLIST <- lavmodel@GLIST
 
-    ngroups        <- lavmodel@ngroups
+    nblocks        <- lavmodel@nblocks
     nmat           <- lavmodel@nmat
     representation <- lavmodel@representation
 
     # return a list
-    EETAx <- vector("list", length=ngroups)
+    EETAx <- vector("list", length=nblocks)
 
     # compute E(ETA) for each group
-    for(g in 1:ngroups) {
+    for(g in 1:nblocks) {
         # which mm belong to group g?
         mm.in.group <- 1:nmat[g] + cumsum(c(0,nmat))[g]
         MLIST <- GLIST[ mm.in.group ]
@@ -437,12 +539,12 @@ computeEETAx <- function(lavmodel = NULL, GLIST = NULL, lavsamplestats = NULL,
         EXO <- eXo[[g]]
         if(is.null(EXO)) {
             # create empty matrix
-            EXO <- matrix(0, lavsamplestats@nobs[[g]], 0L)
+            EXO <- matrix(0, nobs[[g]], 0L)
         }
 
         if(representation == "LISREL") {
             EETAx.g <- computeEETAx.LISREL(MLIST, 
-                eXo=EXO, N=lavsamplestats@nobs[[g]],
+                eXo=EXO, N=nobs[[g]],
                 sample.mean=lavsamplestats@mean[[g]],
                 ov.y.dummy.lv.idx=lavmodel@ov.y.dummy.lv.idx[[g]],
                 ov.x.dummy.lv.idx=lavmodel@ov.x.dummy.lv.idx[[g]],
@@ -474,15 +576,15 @@ computeLAMBDA <- function(lavmodel = NULL, GLIST = NULL,
     # state or final?
     if(is.null(GLIST)) GLIST <- lavmodel@GLIST
 
-    ngroups        <- lavmodel@ngroups
+    nblocks        <- lavmodel@nblocks
     nmat           <- lavmodel@nmat
     representation <- lavmodel@representation
 
     # return a list
-    LAMBDA <- vector("list", length=ngroups)
+    LAMBDA <- vector("list", length=nblocks)
 
     # compute LAMBDA for each group
-    for(g in 1:ngroups) {
+    for(g in 1:nblocks) {
         # which mm belong to group g?
         mm.in.group <- 1:nmat[g] + cumsum(c(0,nmat))[g]
         MLIST <- GLIST[ mm.in.group ]
@@ -509,15 +611,15 @@ computeTHETA <- function(lavmodel = NULL, GLIST = NULL) {
     # state or final?
     if(is.null(GLIST)) GLIST <- lavmodel@GLIST
 
-    ngroups        <- lavmodel@ngroups
+    nblocks        <- lavmodel@nblocks
     nmat           <- lavmodel@nmat
     representation <- lavmodel@representation
 
     # return a list
-    THETA <- vector("list", length=ngroups)
+    THETA <- vector("list", length=nblocks)
 
     # compute THETA for each group
-    for(g in 1:ngroups) {
+    for(g in 1:nblocks) {
         # which mm belong to group g?
         mm.in.group <- 1:nmat[g] + cumsum(c(0,nmat))[g]
         MLIST <- GLIST[ mm.in.group ]
@@ -545,15 +647,15 @@ computeEY <- function(lavmodel = NULL, GLIST = NULL, lavsamplestats = NULL) {
     # state or final?
     if(is.null(GLIST)) GLIST <- lavmodel@GLIST
 
-    ngroups        <- lavmodel@ngroups
+    nblocks        <- lavmodel@nblocks
     nmat           <- lavmodel@nmat
     representation <- lavmodel@representation
 
     # return a list
-    EY <- vector("list", length=ngroups)
+    EY <- vector("list", length=nblocks)
 
     # compute E(Y) for each group
-    for(g in 1:ngroups) {
+    for(g in 1:nblocks) {
         # which mm belong to group g?
         mm.in.group <- 1:nmat[g] + cumsum(c(0,nmat))[g]
         MLIST <- GLIST[ mm.in.group ]
@@ -580,28 +682,32 @@ computeEY <- function(lavmodel = NULL, GLIST = NULL, lavsamplestats = NULL) {
 # E(Y | ETA, x_i): conditional expectation (means) of observed variables
 # for a given value of x_i AND eta_i
 computeYHAT <- function(lavmodel = NULL, GLIST = NULL, lavsamplestats = NULL, 
-                        eXo = NULL, ETA = NULL, duplicate = FALSE) {
+                        eXo = NULL, nobs = NULL, ETA = NULL, duplicate = FALSE) {
   
     # state or final?
     if(is.null(GLIST)) GLIST <- lavmodel@GLIST
 
+    # ngroups, not nblocks!
+    ngroups <- lavsamplestats@ngroups
+
     # return a list
-    YHAT <- vector("list", length=lavsamplestats@ngroups)
+    YHAT <- vector("list", length=ngroups)
 
     # compute YHAT for each group
-    for(g in seq_len(lavsamplestats@ngroups)) {
+    for(g in seq_len(ngroups)) {
         # which mm belong to group g?
+        # FIXME: what if more than g blocks???
         mm.in.group <- 1:lavmodel@nmat[g] + cumsum(c(0L,lavmodel@nmat))[g]
         MLIST <- GLIST[ mm.in.group ]
 
         if(is.null(eXo[[g]]) && duplicate) {
-            Nobs <- lavsamplestats@nobs[[g]]
+            Nobs <- nobs[[g]]
         } else {
             Nobs <- 1L
         }
 
         if(lavmodel@representation == "LISREL") {
-            if(lavmodel@categorical) {
+            if(lavmodel@conditional.x) {
                 YHAT[[g]] <- computeEYetax.LISREL(MLIST = MLIST,
                           eXo = eXo[[g]], ETA = ETA[[g]], N = Nobs,
                           sample.mean = lavsamplestats@mean[[g]],
