@@ -4,9 +4,8 @@ lav_model_gradient <- function(lavmodel       = NULL,
                                GLIST          = NULL, 
                                lavsamplestats = NULL, 
                                lavdata        = NULL,
-                               lavcache       = NULL, 
+                               lavcache       = NULL,
                                type           = "free", 
-                               estimator      = "ML", 
                                verbose        = FALSE, 
                                forcePD        = TRUE, 
                                group.weight   = TRUE,
@@ -15,6 +14,7 @@ lav_model_gradient <- function(lavmodel       = NULL,
                                x.el.idx       = NULL) {
 
     nmat           <- lavmodel@nmat
+    estimator      <- lavmodel@estimator
     representation <- lavmodel@representation
     meanstructure  <- lavmodel@meanstructure
     categorical    <- lavmodel@categorical
@@ -31,6 +31,7 @@ lav_model_gradient <- function(lavmodel       = NULL,
     if(estimator == "REML") warning("analytical gradient not implement; use numerical approximation")
 
     # group.weight
+    # FIXME --> block.weight
     if(group.weight) {
         if(estimator %in% c("ML","PML","FML","MML","REML","NTRLS")) {
             group.w <- (unlist(lavsamplestats@nobs)/lavsamplestats@ntotal)
@@ -39,7 +40,7 @@ lav_model_gradient <- function(lavmodel       = NULL,
             group.w <- ((unlist(lavsamplestats@nobs)-1)/lavsamplestats@ntotal)
         }
     } else {
-        group.w <- rep(1.0, lavsamplestats@ngroups)
+        group.w <- rep(1.0, lavmodel@nblocks)
     }
 
     # do we need WLS.est?
@@ -54,7 +55,7 @@ lav_model_gradient <- function(lavmodel       = NULL,
         # compute moments for all groups
         #if(conditional.x) {
         #    Sigma.hat <- computeSigmaHatJoint(lavmodel = lavmodel,
-        #                     GLIST = GLIST, lavsamplestats = lavsamplestats,
+        #                     GLIST = GLIST,
         #                     extra = (estimator %in% c("ML", "REML","NTRLS")))
         #} else {
             Sigma.hat <- computeSigmaHat(lavmodel = lavmodel, GLIST = GLIST,
@@ -62,26 +63,24 @@ lav_model_gradient <- function(lavmodel       = NULL,
         #}
 
         # ridge here?
-        if(meanstructure && !categorical) {
+        if(meanstructure) {
             #if(conditional.x) {
-            #    Mu.hat <- computeMuHat(lavmodel = lavmodel, GLIST = GLIST,
-            #                           lavsamplestats = lavsamplestats)
+            #    Mu.hat <- computeMuHat(lavmodel = lavmodel, GLIST = GLIST)
             #} else { 
                 Mu.hat <- computeMuHat(lavmodel = lavmodel, GLIST = GLIST)
             #}
-        } else if(categorical) {
+        } 
+
+        if(categorical) {
             TH <- computeTH(lavmodel = lavmodel, GLIST = GLIST)
         }
+
         if(conditional.x) {
             PI <- computePI(lavmodel = lavmodel, GLIST = GLIST)
+        } else if(estimator == "PML") {
+            PI <- vector("list", length = lavmodel@nblocks)
         }
-        if(estimator == "PML") {
-            if(lavmodel@nexo > 0L) {
-                PI <- computePI(lavmodel = lavmodel)
-            } else {
-                PI <- vector("list", length = lavsamplestats@ngroups)
-            }
-        }
+
         if(group.w.free) {
             GW <- computeGW(lavmodel = lavmodel, GLIST = GLIST)
         }
@@ -101,6 +100,7 @@ lav_model_gradient <- function(lavmodel       = NULL,
 
     # 1. ML approach
     if( (estimator == "ML" || estimator == "REML") && 
+        lavdata@nlevels == 1L &&
         !lavmodel@conditional.x ) {
 
         if(meanstructure) {
@@ -116,13 +116,13 @@ lav_model_gradient <- function(lavmodel       = NULL,
                                   estimator=estimator,
                                   meanstructure=FALSE,
                                   conditional.x = conditional.x)
-            Omega.mu <- vector("list", length=lavsamplestats@ngroups)
+            Omega.mu <- vector("list", length = lavmodel@nblocks)
         }
 
         # compute DX (for all elements in every model matrix)
         DX <- vector("list", length=length(GLIST))
       
-        for(g in 1:lavsamplestats@ngroups) {
+        for(g in 1:lavmodel@nblocks) {
             # which mm belong to group g?
             mm.in.group <- 1:nmat[g] + cumsum(c(0,nmat))[g]
             mm.names <- names( GLIST[mm.in.group] )
@@ -145,7 +145,7 @@ lav_model_gradient <- function(lavmodel       = NULL,
             }
 
             # weight by group
-            if(lavsamplestats@ngroups > 1L) {
+            if(lavmodel@nblocks > 1L) {
                 for(mm in mm.in.group) {
                     DX[[mm]] <- group.w[g] * DX[[mm]]
                 }
@@ -155,7 +155,7 @@ lav_model_gradient <- function(lavmodel       = NULL,
         # extract free parameters
         if(type == "free") {
             dx <- numeric( nx.free )
-            for(g in 1:lavsamplestats@ngroups) {
+            for(g in 1:lavmodel@nblocks) {
                 mm.in.group <- 1:nmat[g] + cumsum(c(0,nmat))[g]
                 for(mm in mm.in.group) {
                       m.free.idx  <- lavmodel@m.free.idx[[mm]]
@@ -187,7 +187,7 @@ lav_model_gradient <- function(lavmodel       = NULL,
             Delta <- computeDelta(lavmodel = lavmodel, GLIST. = GLIST)
         }
 
-        for(g in 1:lavsamplestats@ngroups) {
+        for(g in 1:lavmodel@nblocks) {
             #diff <- as.matrix(lavsamplestats@WLS.obs[[g]]  - WLS.est[[g]])
             #group.dx <- -1 * ( t(Delta[[g]]) %*% lavsamplestats@WLS.V[[g]] %*% diff)
             # 0.5-17: use crossprod twice; treat DWLS/ULS special
@@ -275,7 +275,7 @@ lav_model_gradient <- function(lavmodel       = NULL,
             Delta <- computeDelta(lavmodel = lavmodel, GLIST. = GLIST)
         }
 
-        for(g in 1:lavsamplestats@ngroups) {
+        for(g in 1:lavmodel@nblocks) {
 
             # augmented mean.x + cov.x matrix
             mean.x <- lavsamplestats@mean.x[[g]]
@@ -313,7 +313,9 @@ lav_model_gradient <- function(lavmodel       = NULL,
             # NOTE: the vecr here, unlike lav_mvreg_dlogl_beta
             #       this is because DELTA has used vec(t(BETA)), 
             #       instead of vec(BETA)
-            POST.beta <- 2 * lav_matrix_vecr(d.BETA)
+            #POST.beta <- 2 * lav_matrix_vecr(d.BETA)
+            # NOT any longer, since 0.6-1!!! 
+            POST.beta <- 2 * lav_matrix_vec(d.BETA)
 
             #POST.sigma1 <- lav_matrix_duplication_pre(
             #        (Sigma.inv %x% Sigma.inv)  %*% t(AB)  %*% (t(a) %x% b) )
@@ -333,6 +335,9 @@ lav_model_gradient <- function(lavmodel       = NULL,
 
             group.dx <- as.numeric( -1 * crossprod(Delta[[g]], POST) )
 
+            # because we still use obj/2, we need to divide by 2!
+            group.dx <- group.dx / 2 # fixed in 0.6-1
+
             group.dx <- group.w[g] * group.dx
             if(g == 1) {
                 dx <- group.dx
@@ -352,6 +357,43 @@ lav_model_gradient <- function(lavmodel       = NULL,
         }
     } # ML + conditional.x
 
+    else if(estimator == "ML" && lavdata@nlevels > 1L) {
+        if(type != "free") {
+            stop("FIXME: type != free in lav_model_gradient for estimator ML for nlevels > 1")
+        } else {
+            Delta <- computeDelta(lavmodel = lavmodel, GLIST. = GLIST)
+        }
+
+        # for each upper-level group....
+        for(g in 1:lavmodel@ngroups) {
+            DX <- lav_mvnorm_cluster_dlogl_2l_samplestats(
+                       YLp = lavsamplestats@YLp[[g]],
+                       Lp  = lavdata@Lp[[g]],
+                       Mu.W    = Mu.hat[[(g-1)*2 + 1]],
+                       Sigma.W = Sigma.hat[[(g-1)*2 + 1]],
+                       Mu.B    = Mu.hat[[(g-1)*2 + 2]],
+                       Sigma.B = Sigma.hat[[(g-1)*2 + 2]],
+                       Sinv.method  = "eigen")
+
+            group.dx <- as.numeric( DX %*% Delta[[g]] )
+
+            # group weights (if any)
+            group.dx <- group.w[g] * group.dx
+            if(g == 1) {
+                dx <- group.dx
+            } else {
+                dx <- dx + group.dx
+            }
+        } # g
+
+        # divide by 2 * N
+        dx <- dx / (2 * lavsamplestats@ntotal)
+
+        #cat("dx1 (numerical) = \n"); print( zapsmall(dx1) )
+        #cat("dx  (analytic)  = \n"); print( zapsmall(dx ) )
+            
+    } # ML + two-level
+
     else if(estimator == "PML" || estimator == "FML" ||
             estimator == "MML") {
 
@@ -361,7 +403,7 @@ lav_model_gradient <- function(lavmodel       = NULL,
             Delta <- computeDelta(lavmodel = lavmodel, GLIST. = GLIST)
         }
 
-        for(g in 1:lavsamplestats@ngroups) {
+        for(g in 1:lavmodel@nblocks) {
 
             #print(GLIST)
             #print(lav_model_get_parameters(lavmodel = lavmodel, GLIST = GLIST))
@@ -372,20 +414,40 @@ lav_model_gradient <- function(lavmodel       = NULL,
             # compute partial derivative of logLik with respect to 
             # thresholds/means, slopes, variances, correlations
             if(estimator == "PML") {
-                d1 <- pml_deriv1(Sigma.hat  = Sigma.hat[[g]],
-                                 TH         = TH[[g]],
-                                 th.idx     = th.idx[[g]],
-                                 num.idx    = num.idx[[g]],
-                                 X          = lavdata@X[[g]],
-                                 lavcache   = lavcache[[g]],
-                                 eXo        = lavdata@eXo[[g]],
-                                 PI         = PI[[g]],
-                                 missing    = lavdata@missing)
+
+                if(lavdata@nlevels  > 1L) {
+                    stop("lavaan ERROR: PL gradient + multilevel not implemented; try optim.gradient = \"numerical\"")
+                } else if(conditional.x) {
+                    d1 <- pml_deriv1(Sigma.hat  = Sigma.hat[[g]],
+                                     Mu.hat     = Mu.hat[[g]],
+                                     TH         = TH[[g]],
+                                     th.idx     = th.idx[[g]],
+                                     num.idx    = num.idx[[g]],
+                                     X          = lavdata@X[[g]],
+                                     lavcache   = lavcache[[g]],
+                                     eXo        = lavdata@eXo[[g]],
+                                     PI         = PI[[g]],
+                                     missing    = lavdata@missing)
+                } else {
+                    d1 <- pml_deriv1(Sigma.hat  = Sigma.hat[[g]],
+                                     Mu.hat     = Mu.hat[[g]],
+                                     TH         = TH[[g]],
+                                     th.idx     = th.idx[[g]],
+                                     num.idx    = num.idx[[g]],
+                                     X          = lavdata@X[[g]],
+                                     lavcache   = lavcache[[g]],
+                                     eXo        = NULL,
+                                     PI         = NULL,
+                                     missing    = lavdata@missing)
+                } # not conditional.x
+
                 # chain rule (fmin)
                 group.dx <- 
                     as.numeric(t(d1) %*% Delta[[g]])
 
-            } else if(estimator == "FML") {
+           } # PML
+
+           else if(estimator == "FML") {
                 d1 <- fml_deriv1(Sigma.hat = Sigma.hat[[g]],
                                  TH        = TH[[g]],
                                  th.idx    = th.idx[[g]],
@@ -406,6 +468,7 @@ lav_model_gradient <- function(lavmodel       = NULL,
                                     group          = g,
                                     lavdata        = lavdata,
                                     sample.mean    = lavsamplestats@mean[[g]],
+                                    sample.mean.x  = lavsamplestats@mean.x[[g]],
                                     lavcache       = lavcache)
             }
 
@@ -492,7 +555,7 @@ computeDelta <- function(lavmodel = NULL, GLIST. = NULL,
     conditional.x    <- lavmodel@conditional.x
     group.w.free     <- lavmodel@group.w.free
     nmat             <- lavmodel@nmat
-    ngroups          <- lavmodel@ngroups
+    nblocks          <- lavmodel@nblocks
     nvar             <- lavmodel@nvar
     num.idx          <- lavmodel@num.idx
     th.idx           <- lavmodel@th.idx
@@ -515,8 +578,8 @@ computeDelta <- function(lavmodel = NULL, GLIST. = NULL,
          type <- "free"
 
     # number of rows in DELTA.group
-    pstar <- integer(ngroups)
-    for(g in 1:ngroups) {
+    pstar <- integer(nblocks)
+    for(g in 1:nblocks) {
         pstar[g] <- as.integer(nvar[g] * (nvar[g] + 1) / 2)
         if(lavmodel@meanstructure) {
             pstar[g] <- nvar[g] + pstar[g]  # first the means, then sigma
@@ -572,8 +635,8 @@ computeDelta <- function(lavmodel = NULL, GLIST. = NULL,
 
 
     # compute Delta
-    Delta <- vector("list", length=ngroups)
-    for(g in 1:ngroups) {
+    Delta <- vector("list", length=nblocks)
+    for(g in 1:nblocks) {
         Delta.group <- matrix(0, nrow=pstar[g], ncol=NCOL)
 
         # which mm belong to group g?
@@ -632,27 +695,38 @@ computeDelta <- function(lavmodel = NULL, GLIST. = NULL,
                     DELTA <- rbind(DELTA[var.idx,,drop=FALSE], 
                                    DELTA[cor.idx,,drop=FALSE])
                 }
+
                 if(!categorical) {
-                    if(lavmodel@meanstructure) {
+                    if(conditional.x && lavmodel@nexo[g] > 0L) {
+                        # ATTENTION: we need to change the order here
+                        # lav_mvreg_scores_* uses 'Beta' where the
+                        # the intercepts are just the first row
+                        # using the col-major approach, we need to
+                        # interweave the intercepts with the slopes!
                         DELTA.mu <- derivative.mu.LISREL(m=mname,
                                idx=m.el.idx[[mm]], MLIST=GLIST[ mm.in.group ])
-                        if(conditional.x && lavmodel@nexo[g] > 0L) {
-                            DELTA.pi <- derivative.pi.LISREL(m=mname,
+                        DELTA.pi <- derivative.pi.LISREL(m=mname,
                              idx=m.el.idx[[mm]], MLIST=GLIST[ mm.in.group ])
-                            DELTA <- rbind(DELTA.mu, DELTA.pi, DELTA)
-                        } else {
+
+                        nEls <- NROW(DELTA.mu) + NROW(DELTA.pi)
+                        # = (nexo + 1 int) * nvar
+
+                        # intercepts on top
+                        tmp <- rbind(DELTA.mu, DELTA.pi)
+                        # change row index
+                        row.idx <- lav_matrix_vec(matrix(seq.int(nEls),
+                            nrow = lavmodel@nexo[g] + 1L,
+                            ncol = lavmodel@nvar[g], byrow = TRUE))
+                        DELTA.beta <- tmp[row.idx,,drop = FALSE]
+                        DELTA <- rbind(DELTA.beta, DELTA)
+                    } else if(!conditional.x && lavmodel@meanstructure) {
+                            DELTA.mu <- derivative.mu.LISREL(m=mname,
+                               idx=m.el.idx[[mm]], MLIST=GLIST[ mm.in.group ])
                             DELTA <- rbind(DELTA.mu, DELTA)
-                        }
-                    } else {
-                        if(conditional.x && lavmodel@nexo[g] > 0L) {
-                            DELTA.pi <- derivative.pi.LISREL(m=mname,
-                             idx=m.el.idx[[mm]], MLIST=GLIST[ mm.in.group ])
-                            DELTA <- rbind(DELTA.pi, DELTA)
-                        } else {
-                            # nothing to do
-                        }
                     }
-                } else if(categorical) {
+                } 
+
+                else if(categorical) {
                     DELTA.th <- derivative.th.LISREL(m=mname,
                                                      idx=m.el.idx[[mm]],
                                                      th.idx=th.idx[[g]],
@@ -674,7 +748,7 @@ computeDelta <- function(lavmodel = NULL, GLIST. = NULL,
                             DELTA.th[no.num.idx,,drop=FALSE] +
                             (dth.dDelta %*% dDelta.dx)[no.num.idx,,drop=FALSE]
                     }
-                    if(lavmodel@nexo[g] > 0L) {
+                    if(conditional.x && lavmodel@nexo[g] > 0L) {
                         DELTA.pi <- 
                             derivative.pi.LISREL(m=mname,
                                                  idx=m.el.idx[[mm]],
@@ -725,6 +799,16 @@ computeDelta <- function(lavmodel = NULL, GLIST. = NULL,
 
     } # g
 
+    # if multilevel, rbind levels within group
+    if(.hasSlot(lavmodel, "multilevel") && lavmodel@multilevel) {
+        DELTA <- vector("list", length = lavmodel@ngroups)
+        for(g in 1:lavmodel@ngroups) {
+            DELTA[[g]] <- rbind( Delta[[(g-1)*2 + 1]],
+                                 Delta[[(g-1)*2 + 2]] )
+        }
+        Delta <- DELTA
+    }
+
     Delta
 }
 
@@ -735,7 +819,7 @@ computeDeltaDx <- function(lavmodel = NULL, GLIST = NULL, target = "lambda") {
 
     representation   <- lavmodel@representation
     nmat             <- lavmodel@nmat
-    ngroups          <- lavmodel@ngroups
+    nblocks          <- lavmodel@nblocks
     num.idx          <- lavmodel@num.idx
     th.idx           <- lavmodel@th.idx
 
@@ -763,8 +847,8 @@ computeDeltaDx <- function(lavmodel = NULL, GLIST = NULL, target = "lambda") {
     #}
 
     # compute Delta per group
-    Delta <- vector("list", length=ngroups)
-    for(g in 1:ngroups) {
+    Delta <- vector("list", length=nblocks)
+    for(g in 1:nblocks) {
         mm.in.group <- 1:nmat[g] + cumsum(c(0,nmat))[g]
         Delta.group <- NULL
         for(mm in mm.in.group) {
@@ -836,10 +920,13 @@ computeOmega <- function(Sigma.hat=NULL, Mu.hat=NULL,
                          lavsamplestats=NULL, estimator="ML", 
                          meanstructure=FALSE, conditional.x = FALSE) {
 
-    Omega    <- vector("list", length=lavsamplestats@ngroups)
-    Omega.mu <- vector("list", length=lavsamplestats@ngroups)
+    # nblocks
+    nblocks <- length(Sigma.hat)
 
-    for(g in 1:lavsamplestats@ngroups) {
+    Omega    <- vector("list", length = nblocks)
+    Omega.mu <- vector("list", length = nblocks)
+
+    for(g in 1:nblocks) {
 
         # ML
         if(estimator == "ML" || estimator == "REML") {
