@@ -6,9 +6,29 @@
 # 3) casewise scores with respect to mu, vech(Sigma), mu + vech(Sigma)
 # 4) hessian of mu + vech(Sigma)
 # 5) (unit) information of mu + vech(Sigma)
+#    5a: (unit)    expected information
+#    5b: (unit)    observed information
+#    5c: (unit) first.order information
+#    5d: lav_mvnorm_missing_information_both (both observed + first.order)
+
+# 6) inverted information h0 mu + vech(Sigma)
+#    6a: /
+#    6b: /
+#    6c: /
+# 7) ACOV h0 mu + vech(Sigma)
+#    7a: 1/N * inverted expected    information
+#    7b: 1/N * inverted observed    information
+#    7c: 1/N * inverted first-order information
+#    7d: sandwich acov
+
+# 10) additional functions
+#      - lav_mvnorm_missing_impute_pattern
+#      - lav_mvnorm_missing_estep
 
 
 # YR 09 Feb 2016: first version
+# YR 19 Mar 2017: 10)
+# YR 03 Okt 2018: a few functions gain a wt= argument
 
 
 # 1) likelihood
@@ -20,17 +40,27 @@
 #    (much) faster than the other
 lav_mvnorm_missing_loglik_data <- function(Y           = NULL,
                                            Mu          = NULL,
+                                           wt          = NULL,
                                            Sigma       = NULL,
+                                           x.idx       = NULL,
                                            casewise    = FALSE,
                                            pattern     = TRUE,
-                                           Sinv.method = "eigen") {
+                                           Sinv.method = "eigen",
+                                           log2pi      = TRUE,
+                                           minus.two   = FALSE) {
+
+    if(!is.null(x.idx) && length(x.idx) > 0L) {
+        warning("lavaan WARNING: x.idx not supported yet (ignored)")
+    }
 
     if(pattern) {
-        llik <- lav_mvnorm_missing_llik_pattern(Y = Y, Mu = Mu, 
-                    Sigma = Sigma, Sinv.method = Sinv.method)
+        llik <- lav_mvnorm_missing_llik_pattern(Y = Y, wt = wt, Mu = Mu, 
+                    Sigma = Sigma, Sinv.method = Sinv.method,
+                    log2pi = log2pi, minus.two = minus.two)
     } else {
-        llik <- lav_mvnorm_missing_llik_casewise(Y = Y, Mu = Mu, 
-                    Sigma = Sigma, Sinv.method = Sinv.method)
+        llik <- lav_mvnorm_missing_llik_casewise(Y = Y, wt = wt, Mu = Mu, 
+                    Sigma = Sigma, Sinv.method = Sinv.method,
+                    log2pi = log2pi, minus.two = minus.two)
     }
 
     if(casewise) {
@@ -46,7 +76,17 @@ lav_mvnorm_missing_loglik_data <- function(Y           = NULL,
 lav_mvnorm_missing_loglik_samplestats <- function(Yp          = NULL,
                                                   Mu          = NULL,
                                                   Sigma       = NULL,
-                                                  Sinv.method = "eigen") {
+                                                  x.idx       = NULL,
+                                                  x.mean      = NULL,
+                                                  x.cov       = NULL,
+                                                  Sinv.method = "eigen",
+                                                  log2pi      = TRUE,
+                                                  minus.two   = FALSE) {
+
+    if(!is.null(x.idx) && length(x.idx) > 0L) {
+        #warning("lavaan WARNING: x.idx not supported yet (ignored)")
+        
+    }
 
     LOG.2PI <- log(2*pi); pat.N <- length(Yp);  P <- length(Yp[[1]]$var.idx)
 
@@ -85,8 +125,38 @@ lav_mvnorm_missing_loglik_samplestats <- function(Yp          = NULL,
         DIST[p] <- sum(sigma.inv * TT) * Yp[[p]]$freq
     }
 
-    # compute loglikelihoods per pattern
-    loglik <- sum(-(P.LOG.2PI + logdet + DIST)/2)
+    # loglikelihood all data
+    if(log2pi) {
+        loglik <- sum(-(P.LOG.2PI + logdet + DIST)/2)
+    } else {
+        loglik <- sum(-(logdet + DIST)/2)
+    }
+
+    if(minus.two) {
+        loglik <- -2 * loglik
+    }
+
+    # x.idx
+    if(!is.null(x.idx) && length(x.idx) > 0L) {
+        #warning("lavaan WARNING: x.idx not supported yet (ignored)")
+        P.x <- length(x.idx); N <- sum(sapply(Yp, "[[", "freq"))
+        Sigma.x <- Sigma[x.idx, x.idx, drop = FALSE]
+        Sigma.inv.x <- lav_matrix_symmetric_inverse(S = Sigma.x, logdet = TRUE,
+                                              Sinv.method = Sinv.method)
+        logdet.x <- attr(Sigma.inv.x, "logdet")
+
+        if(log2pi) {
+            loglik.x <- -N/2 * (P.x * LOG.2PI + logdet.x + P.x)
+        } else {
+            loglik.x <- -N/2 * (logdet.x + P.x)
+        }
+
+        if(minus.two) {
+            loglik.x <- -2 * loglik.x
+        }
+
+        loglik <- loglik - loglik.x
+    }
 
     loglik
 }
@@ -95,9 +165,12 @@ lav_mvnorm_missing_loglik_samplestats <- function(Yp          = NULL,
 
 # casewise Sinv.method
 lav_mvnorm_missing_llik_casewise <- function(Y           = NULL,
+                                             wt          = NULL,
                                              Mu          = NULL,
                                              Sigma       = NULL,
-                                             Sinv.method = "eigen") {
+                                             Sinv.method = "eigen",
+                                             log2pi      = TRUE,
+                                             minus.two   = FALSE) {
 
     P <- NCOL(Y); N <- NROW(Y); LOG.2PI <- log(2*pi); Mu <- as.numeric(Mu)
 
@@ -138,20 +211,30 @@ lav_mvnorm_missing_llik_casewise <- function(Y           = NULL,
         if(length(na.idx) == P) next
 
         # invert Sigma for this pattern
-        if(length(na.idx) > 0L) {
-            sigma.inv <- lav_matrix_symmetric_inverse_update(S.inv = Sigma.inv,
+        sigma.inv <- lav_matrix_symmetric_inverse_update(S.inv = Sigma.inv,
                 rm.idx = na.idx, logdet = TRUE, S.logdet = Sigma.logdet)
-            logdet[i] <- attr(sigma.inv, "logdet")
-        } else {
-            sigma.inv <- Sigma.inv
-            logdet[i] <- Sigma.logdet
-        }
+        logdet[i] <- attr(sigma.inv, "logdet")
 
+        # distance for this case
         DIST[i] <- sum(sigma.inv * crossprod(Yc[i, OBS[i,], drop = FALSE]))
     }
 
     # compute casewise loglikelihoods
-    llik <- -(P.LOG.2PI + logdet + DIST)/2
+    if(log2pi) {
+        llik <- -(P.LOG.2PI + logdet + DIST)/2
+    } else {
+        llik <- -(logdet + DIST)/2
+    }
+
+    # minus.two
+    if(minus.two) {
+        llik <- -2 * llik
+    }
+
+    # weights?
+    if(!is.null(wt)) {
+        llik <- llik * wt
+    }
 
     llik
 }
@@ -159,9 +242,12 @@ lav_mvnorm_missing_llik_casewise <- function(Y           = NULL,
 # pattern-based, but casewise loglikelihoods
 lav_mvnorm_missing_llik_pattern <- function(Y           = NULL,
                                             Mp          = NULL,
+                                            wt          = NULL,
                                             Mu          = NULL,
                                             Sigma       = NULL,
-                                            Sinv.method = "eigen") {
+                                            Sinv.method = "eigen",
+                                            log2pi      = TRUE,
+                                            minus.two   = FALSE) {
 
     P <- NCOL(Y); N <- NROW(Y); LOG.2PI <- log(2*pi); Mu <- as.numeric(Mu)
 
@@ -218,7 +304,21 @@ lav_mvnorm_missing_llik_pattern <- function(Y           = NULL,
     }
 
     # compute casewise loglikelihoods
-    llik <- -(P.LOG.2PI + logdet + DIST)/2
+    if(log2pi) {
+        llik <- -(P.LOG.2PI + logdet + DIST)/2
+    } else {
+        llik <- -(logdet + DIST)/2
+    }
+
+    # minus.two
+    if(minus.two) {
+        llik <- -2 * llik
+    }
+
+    # weights?
+    if(!is.null(wt)) {
+        llik <- llik * wt
+    }
 
     llik
 }
@@ -230,12 +330,13 @@ lav_mvnorm_missing_llik_pattern <- function(Y           = NULL,
 
 # 2a: derivative logl with respect to mu
 lav_mvnorm_missing_dlogl_dmu <- function(Y           = NULL,
+                                         wt          = NULL,
                                          Mu          = NULL,
                                          Sigma       = NULL,
                                          Sigma.inv   = NULL,
                                          Sinv.method = "eigen") {
 
-    SC <- lav_mvnorm_missing_scores_mu(Y = Y, Mu = Mu, Sigma = Sigma,
+    SC <- lav_mvnorm_missing_scores_mu(Y = Y, wt = wt, Mu = Mu, Sigma = Sigma,
               Sigma.inv = Sigma.inv, Sinv.method = Sinv.method)
 
     colSums(SC, na.rm = TRUE)
@@ -289,6 +390,7 @@ lav_mvnorm_missing_dlogl_dmu_samplestats <- function(Yp          = NULL,
 # 2b: derivative logl with respect to Sigma (full matrix, ignoring symmetry)
 lav_mvnorm_missing_dlogl_dSigma <- function(Y           = NULL,
                                             Mp          = NULL,
+                                            wt          = NULL, 
                                             Mu          = NULL,
                                             Sigma       = NULL,
                                             Sigma.inv   = NULL,
@@ -312,7 +414,7 @@ lav_mvnorm_missing_dlogl_dSigma <- function(Y           = NULL,
         Mp <- lav_data_missing_patterns(Y)
     }
 
-    # for each pattern, compute Yc %*% sigma.inv
+    # for each pattern
     for(p in seq_len(Mp$npatterns)) {
 
         # observed values for this pattern
@@ -332,8 +434,22 @@ lav_mvnorm_missing_dlogl_dSigma <- function(Y           = NULL,
             sigma.inv <- Sigma.inv
         }
 
+        if(!is.null(wt)) {
+            FREQ <- sum( wt[case.idx] )
+        } else {
+            FREQ <- Mp$freq[p]
+        }
+
         if(length(case.idx) > 1L) {
-            W.tilde <- crossprod(Yc[case.idx, var.idx, drop = FALSE])/Mp$freq[p]
+            if(!is.null(wt)) {
+                out <- stats::cov.wt(Y[case.idx, var.idx, drop = FALSE], 
+                                     wt = wt[Mp$case.idx[[p]]], method = "ML")
+                SY <- out$cov
+                MY <- out$center
+                W.tilde <- SY + tcrossprod(MY - Mu[var.idx])
+            } else {
+                W.tilde <- crossprod(Yc[case.idx, var.idx, drop = FALSE])/FREQ
+            }
         } else {
             W.tilde <- tcrossprod(Yc[case.idx, var.idx])
         }
@@ -344,7 +460,7 @@ lav_mvnorm_missing_dlogl_dSigma <- function(Y           = NULL,
                         (sigma.inv %*% W.tilde %*% sigma.inv))
 
         # update dSigma
-        dSigma <- dSigma + (dSigma.pattern * Mp$freq[p])
+        dSigma <- dSigma + (dSigma.pattern * FREQ)
     }
 
     dSigma
@@ -367,7 +483,7 @@ lav_mvnorm_missing_dlogl_dSigma_samplestats <- function(Yp          = NULL,
     # dvechSigma
     dSigma <- matrix(0, P, P)
 
-    # for each pattern, compute Yc %*% sigma.inv
+    # for each pattern
     for(p in seq_len(pat.N)) {
 
         # observed variables for this pattern
@@ -401,13 +517,14 @@ lav_mvnorm_missing_dlogl_dSigma_samplestats <- function(Yp          = NULL,
 
 # 2c: derivative logl with respect to vech(Sigma)
 lav_mvnorm_missing_dlogl_dvechSigma <- function(Y           = NULL,
+                                                wt          = NULL,
                                                 Mu          = NULL,
                                                 Sigma       = NULL,
                                                 Sigma.inv   = NULL,
                                                 Sinv.method = "eigen") {
 
-    FULL <- lav_mvnorm_missing_dlogl_dSigma(Y = Y, Mu = Mu, Sigma = Sigma,
-              Sigma.inv = Sigma.inv, Sinv.method = Sinv.method)
+    FULL <- lav_mvnorm_missing_dlogl_dSigma(Y = Y, wt = wt, Mu = Mu, 
+            Sigma = Sigma, Sigma.inv = Sigma.inv, Sinv.method = Sinv.method)
     as.numeric( lav_matrix_duplication_pre( as.matrix(lav_matrix_vec(FULL)) ) )
 }
 
@@ -429,7 +546,7 @@ lav_mvnorm_missing_dlogl_dvechSigma_samplestats <-
     # dvechSigma
     dvechSigma <- numeric(P*(P+1)/2)
 
-    # for each pattern, compute Yc %*% sigma.inv
+    # for each pattern
     for(p in seq_len(pat.N)) {
 
         # observed variables for this pattern
@@ -471,6 +588,7 @@ lav_mvnorm_missing_dlogl_dvechSigma_samplestats <-
 
 # 3a: casewise scores with respect to mu
 lav_mvnorm_missing_scores_mu <- function(Y           = NULL,
+                                         wt          = NULL,
                                          Mp          = NULL,
                                          Mu          = NULL,
                                          Sigma       = NULL,
@@ -499,6 +617,9 @@ lav_mvnorm_missing_scores_mu <- function(Y           = NULL,
     # for each pattern, compute sigma.inv
     for(p in seq_len(Mp$npatterns)) {
 
+        # observed values for this pattern
+        var.idx <- Mp$pat[p,]
+
         # missing values for this pattern
         na.idx <- which(!var.idx)
 
@@ -517,11 +638,16 @@ lav_mvnorm_missing_scores_mu <- function(Y           = NULL,
             Yc[case.idx, var.idx, drop = FALSE] %*% sigma.inv
     }
 
+    if(!is.null(wt)) {
+        dmu <- dmu * wt
+    }
+
     dmu
 }
 
 # 3b: casewise scores with respect to vech(Sigma)
 lav_mvnorm_missing_scores_vech_sigma <- function(Y           = NULL,
+                                                 wt          = NULL,
                                                  Mp          = NULL,
                                                  Mu          = NULL,
                                                  Sigma       = NULL,
@@ -553,7 +679,7 @@ lav_mvnorm_missing_scores_vech_sigma <- function(Y           = NULL,
     # SC
     SC <- matrix(as.numeric(NA), nrow = N, ncol = length(iSigma))
 
-    # for each pattern, compute Yc %*% sigma.inv
+    # for each pattern
     for(p in seq_len(Mp$npatterns)) {
 
         # observed values for this pattern
@@ -590,12 +716,17 @@ lav_mvnorm_missing_scores_vech_sigma <- function(Y           = NULL,
     # adjust for vech
     SC[,lav_matrix_diagh_idx(P)] <- SC[,lav_matrix_diagh_idx(P)] / 2
 
+    if(!is.null(wt)) {
+        SC <- SC * wt
+    }
+
     SC
 }
 
 # 3c: casewise scores with respect to mu + vech(Sigma)
 lav_mvnorm_missing_scores_mu_vech_sigma <- function(Y           = NULL,
                                                     Mp          = NULL,
+                                                    wt          = NULL,
                                                     Mu          = NULL,
                                                     Sigma       = NULL,
                                                     Sigma.inv   = NULL,
@@ -670,19 +801,26 @@ lav_mvnorm_missing_scores_mu_vech_sigma <- function(Y           = NULL,
     # adjust for vech
     SC[,lav_matrix_diagh_idx(P)] <- SC[,lav_matrix_diagh_idx(P)] / 2
 
-    cbind(dmu, SC)
+    out <- cbind(dmu, SC)
+
+    if(!is.null(wt)) {
+        out <- out * wt
+    }
+
+    out
 }
 
 
 # 4) Hessian of logl
 lav_mvnorm_missing_logl_hessian_data <- function(Y           = NULL,
                                                  Mp          = NULL,
+                                                 wt          = NULL,
                                                  Mu          = NULL,
                                                  Sigma       = NULL,
                                                  Sinv.method = "eigen",
                                                  Sigma.inv   = NULL) {
     # missing patterns
-    Yp <- lav_samplestats_missing_patterns(Y = Y, Mp = Mp)
+    Yp <- lav_samplestats_missing_patterns(Y = Y, Mp = Mp, wt = wt)
 
     lav_mvnorm_missing_logl_hessian_samplestats(Yp = Yp, Mu = Mu,
         Sigma = Sigma, Sinv.method = Sinv.method, Sigma.inv = Sigma.inv)
@@ -762,6 +900,7 @@ lav_mvnorm_missing_logl_hessian_samplestats <-
 # (old term: Abeta, expected)
 lav_mvnorm_missing_information_expected <- function(Y           = NULL,
                                                     Mp          = NULL,
+                                                    wt          = NULL,
                                                     Mu          = NULL,# unused
                                                     Sigma       = NULL,
                                                     Sigma.inv   = NULL,
@@ -780,7 +919,11 @@ lav_mvnorm_missing_information_expected <- function(Y           = NULL,
     }
 
     # N
-    N <- sum(Mp$freq) # removed empty cases!
+    if(!is.null(wt)) {
+        N <- sum(wt)
+    } else {
+        N <- sum(Mp$freq) # removed empty cases!
+    }
 
     I11 <- matrix(0, P, P)
     I22 <- matrix(0, P*(P+1)/2, P*(P+1)/2)
@@ -807,8 +950,14 @@ lav_mvnorm_missing_information_expected <- function(Y           = NULL,
 
         S2.inv <- 0.5 * lav_matrix_duplication_pre_post(S.inv %x% S.inv)
 
-        I11 <- I11 + Mp$freq[p] * S.inv
-        I22 <- I22 + Mp$freq[p] * S2.inv
+        if(!is.null(wt)) { 
+            FREQ <- sum( wt[ Mp$case.idx[[p]] ] )
+        } else {
+            FREQ <- Mp$freq[p]
+        }
+
+        I11 <- I11 + FREQ * S.inv
+        I22 <- I22 + FREQ * S2.inv
     }
 
     lav_matrix_bdiag(I11, I22)/N
@@ -818,6 +967,7 @@ lav_mvnorm_missing_information_expected <- function(Y           = NULL,
 # (old term: Abeta, observed)
 lav_mvnorm_missing_information_observed_data <- function(Y           = NULL,
                                                          Mp          = NULL,
+                                                         wt          = NULL,
                                                          Mu          = NULL,
                                                          Sigma       = NULL,
                                                          Sinv.method = "eigen",
@@ -828,11 +978,15 @@ lav_mvnorm_missing_information_observed_data <- function(Y           = NULL,
     }
 
     # N
-    N <- sum(Mp$freq)
+    if(!is.null(wt)) {
+        N <- sum(wt)
+    } else {
+        N <- sum(Mp$freq)
+    }
 
     # observed information
-    observed <- lav_mvnorm_missing_logl_hessian_data(Y = Y, Mp = Mp, Mu = Mu,
-                    Sigma = Sigma, Sinv.method = Sinv.method, 
+    observed <- lav_mvnorm_missing_logl_hessian_data(Y = Y, Mp = Mp, wt = wt,
+                    Mu = Mu, Sigma = Sigma, Sinv.method = Sinv.method, 
                     Sigma.inv = Sigma.inv)
 
     -observed/N
@@ -860,6 +1014,7 @@ lav_mvnorm_missing_information_observed_samplestats <-
 # (old term: Bbeta)
 lav_mvnorm_missing_information_firstorder <- function(Y           = NULL,
                                                       Mp          = NULL,
+                                                      wt          = NULL,
                                                       Mu          = NULL,
                                                       Sigma       = NULL,
                                                       Sinv.method = "eigen",
@@ -870,10 +1025,15 @@ lav_mvnorm_missing_information_firstorder <- function(Y           = NULL,
     }
 
     # N
-    N <- sum(Mp$freq)
+    if(!is.null(wt)) {
+        N <- sum(wt)
+    } else {
+        N <- sum(Mp$freq)
+    }
 
-    SC <- lav_mvnorm_missing_scores_mu_vech_sigma(Y = Y, Mp = Mp, Mu = Mu, 
-              Sigma = Sigma, Sinv.method = Sinv.method, Sigma.inv = Sigma.inv)
+    SC <- lav_mvnorm_missing_scores_mu_vech_sigma(Y = Y, Mp = Mp, wt = wt, 
+              Mu = Mu, Sigma = Sigma, Sinv.method = Sinv.method, 
+              Sigma.inv = Sigma.inv)
 
     lav_matrix_crossprod(SC)/N
 }
@@ -882,6 +1042,7 @@ lav_mvnorm_missing_information_firstorder <- function(Y           = NULL,
 #     from raw data, in one go for efficiency
 lav_mvnorm_missing_information_both <- function(Y           = NULL,
                                                 Mp          = NULL,
+                                                wt          = NULL,
                                                 Mu          = NULL,
                                                 Sigma       = NULL,
                                                 Sinv.method = "eigen",
@@ -907,11 +1068,15 @@ lav_mvnorm_missing_information_both <- function(Y           = NULL,
     }
 
     if(information == "observed") {
-        Yp <- lav_samplestats_missing_patterns(Y = Y, Mp = Mp)
+        Yp <- lav_samplestats_missing_patterns(Y = Y, Mp = Mp, wt = wt)
     }
 
     # N
-    N <- sum(Mp$freq)
+    if(!is.null(wt)) {
+        N <- sum(wt)
+    } else {
+        N <- sum(Mp$freq)
+    }
 
     # subtract Mu
     Yc <- t( t(Y) - Mu )
@@ -957,11 +1122,17 @@ lav_mvnorm_missing_information_both <- function(Y           = NULL,
         S.inv <- matrix(0, P, P)
         S.inv[var.idx, var.idx] <- sigma.inv
 
+        if(!is.null(wt)) {
+            FREQ <- sum( wt[case.idx] )
+        } else {
+            FREQ <- Mp$freq[p]
+        }
+
         if(information == "expected") {
             S2.inv <- 0.5 * lav_matrix_duplication_pre_post(S.inv %x% S.inv)
 
-            I11 <- I11 + Mp$freq[p] * S.inv
-            I22 <- I22 + Mp$freq[p] * S2.inv
+            I11 <- I11 + FREQ * S.inv
+            I22 <- I22 + FREQ * S2.inv
         } else {
             pat.freq <- Yp[[p]]$freq
         
@@ -1004,6 +1175,10 @@ lav_mvnorm_missing_information_both <- function(Y           = NULL,
     # add dmu
     SC <- cbind(dmu, SC)
 
+    if(!is.null(wt)) {
+        SC <- SC * wt
+    }
+
     # first order information
     Bbeta <- lav_matrix_crossprod(SC)/N
 
@@ -1018,3 +1193,235 @@ lav_mvnorm_missing_information_both <- function(Y           = NULL,
     list(Abeta = Abeta, Bbeta = Bbeta)
 }
 
+
+# 6) inverted information h0 mu + vech(Sigma)
+
+#    6a: (unit) inverted expected information
+#    NOT USED: is not equal to solve(expected)
+#    (although it does converge to the same solution eventually)
+# lav_mvnorm_missing_inverted_information_expected <- function(Y  = NULL,
+#                                                    Mp          = NULL,
+#                                                    Mu          = NULL,# unused
+#                                                    Sigma       = NULL) {
+#    P <- NCOL(Y)
+#
+#    # missing patterns
+#    if(is.null(Mp)) {
+#        Mp <- lav_data_missing_patterns(Y)
+#    }
+#
+#    # N
+#    N <- sum(Mp$freq) # removed empty cases!
+#
+#    I11 <- matrix(0, P, P)
+#    I22 <- matrix(0, P*(P+1)/2, P*(P+1)/2)
+#
+#    # for each pattern
+#    for(p in seq_len(Mp$npatterns)) {
+#
+#        # observed variables
+#        var.idx <- Mp$pat[p,]
+#
+#        sigma <- matrix(0, P, P)
+#        sigma[var.idx, var.idx] <- Sigma[var.idx, var.idx]
+#        sigma2 <- 2 * lav_matrix_duplication_ginv_pre_post(sigma %x% sigma)
+#
+#        I11 <- I11 + Mp$freq[p] * sigma
+#        I22 <- I22 + Mp$freq[p] * sigma2
+#    }
+#
+#    lav_matrix_bdiag(I11, I22)/N
+#}
+
+#    6b: /
+
+#    6c: /
+
+
+# 7) ACOV h0 mu + vech(Sigma)
+
+#    7a: 1/N * inverted expected    information
+
+#    7b: 1/N * inverted observed    information
+
+#    7c: 1/N * inverted first-order information
+
+#    7d: sandwich acov
+
+
+
+# 10) other stuff
+
+# single imputation missing cells, under the normal model, pattern-based
+lav_mvnorm_missing_impute_pattern <- function(Y           = NULL,
+                                              Mp          = NULL,
+                                              Mu          = NULL,
+                                              Sigma       = NULL,
+                                              Sigma.inv   = NULL,
+                                              Sinv.method = "eigen") {
+
+    Mu <- as.numeric(Mu)
+
+    # complete data
+    Y.complete <- Y
+
+    # missing patterns
+    if(is.null(Mp)) {
+        Mp <- lav_data_missing_patterns(Y)
+    }
+
+    if(is.null(Sigma.inv)) {
+        # invert Sigma
+        Sigma.inv <- lav_matrix_symmetric_inverse(S = Sigma, logdet = FALSE,
+                                                  Sinv.method = Sinv.method)
+    }
+
+    # subtract Mu
+    Yc <- t( t(Y) - Mu )
+
+    # fill in data per pattern
+    for(p in seq_len(Mp$npatterns)) {
+
+        # observed values for this pattern
+        var.idx <- Mp$pat[p,]
+
+        # if complete, nothing to do
+        if(all(var.idx)) {
+            next
+        }
+
+        # missing values for this pattern
+        na.idx <- which(!var.idx)
+
+        # extract observed data for these (centered) cases
+        Oc <- Yc[Mp$case.idx[[p]], Mp$pat[p, ], drop = FALSE]
+
+        # invert Sigma (Sigma_22, observed part only) for this pattern
+        Sigma_22.inv <- try(lav_matrix_symmetric_inverse_update(S.inv =
+                            Sigma.inv, rm.idx = na.idx, logdet = FALSE),
+                            silent = TRUE)
+        if(inherits(Sigma_22.inv, "try-error")) {
+            stop("lavaan ERROR: Sigma_22.inv cannot be inverted")
+        }
+
+        # estimate missing values in this pattern
+        Sigma_12 <- Sigma[!var.idx,  var.idx, drop=FALSE]
+        Y.missing <- t( Sigma_12 %*% Sigma_22.inv %*% t(Oc) + Mu[!var.idx] )
+
+        # complete data for this pattern
+        Y.complete[Mp$case.idx[[p]], !var.idx] <- Y.missing
+    }
+
+    Y.complete
+}
+
+
+# E-step: expectations of sum, sum of squares, sum of crossproducts
+# plus correction
+lav_mvnorm_missing_estep <- function(Y           = NULL,
+                                     Mp          = NULL,
+                                     wt          = NULL,
+                                     Mu          = NULL,
+                                     Sigma       = NULL,
+                                     Sigma.inv   = NULL,
+                                     Sinv.method = "eigen") {
+
+    P <- NCOL(Y); Mu <- as.numeric(Mu)
+
+    # missing patterns
+    if(is.null(Mp)) {
+        Mp <- lav_data_missing_patterns(Y)
+    }
+
+    if(is.null(Sigma.inv)) {
+        # invert Sigma
+        Sigma.inv <- lav_matrix_symmetric_inverse(S = Sigma, logdet = FALSE,
+                                                  Sinv.method = Sinv.method)
+    }
+
+    # T1, T2
+    T1 <- numeric(P)
+    T2 <- matrix(0, P, P)
+
+    # update T1 and T2 per pattern
+    for(p in seq_len(Mp$npatterns)) {
+
+        # observed values for this pattern
+        var.idx <- Mp$pat[p,]
+
+        # missing values for this pattern
+        na.idx <- which(!var.idx)
+
+        # extract observed data
+        O <- Y[Mp$case.idx[[p]], Mp$pat[p, ], drop = FALSE]
+
+        # if complete, just compute first and second moments
+        if(all(var.idx)) {
+
+            if(!is.null(wt)) {
+                WT <- wt[Mp$case.idx[[p]]]
+                T1 <- T1 + colSums(WT * O) 
+                T2 <- T2 + crossprod(sqrt(WT) * O)
+            } else {
+                # complete pattern
+                T1 <- T1 + colSums(O)
+                T2 <- T2 + crossprod(O)
+            }
+            next
+        }
+
+        # missing values for this pattern
+        na.idx <- which(!var.idx)
+
+        # partition Sigma (1=missing, 2=complete)
+        Sigma_11 <- Sigma[!var.idx, !var.idx, drop=FALSE]
+        Sigma_12 <- Sigma[!var.idx,  var.idx, drop=FALSE]
+        Sigma_21 <- Sigma[ var.idx, !var.idx, drop=FALSE]
+
+        # invert Sigma (Sigma_22, observed part only) for this pattern
+        Sigma_22.inv <- try(lav_matrix_symmetric_inverse_update(S.inv =
+                            Sigma.inv, rm.idx = na.idx, logdet = FALSE),
+                            silent = TRUE)
+        if(inherits(Sigma_22.inv, "try-error")) {
+            stop("lavaan ERROR: Sigma_22.inv cannot be inverted")
+        }
+
+        # estimate missing values in this pattern
+        Oc <- t( t(O) - Mu[var.idx])
+        Y.missing <- t( Sigma_12 %*% Sigma_22.inv %*% t(Oc) + Mu[!var.idx] )
+
+        # complete data for this pattern
+        Y.complete <- matrix(0, Mp$freq[[p]], P)
+        Y.complete[, var.idx] <- O
+        Y.complete[,!var.idx] <- Y.missing
+
+        if(!is.null(wt)) {
+            WT <- wt[Mp$case.idx[[p]]]
+            T1.pat <- colSums(WT * Y.complete)
+            T2.pat <- crossprod(sqrt(WT) * Y.complete)
+        } else {
+            # 1. SUM `completed' pattern
+            T1.pat <- colSums(Y.complete)
+
+            # 2. CROSSPROD `completed' pattern
+            T2.pat <- crossprod(Y.complete)
+        }
+
+        # correction for missing cells: conditional covariances
+        T2.p11 <- Sigma_11 - (Sigma_12 %*% Sigma_22.inv %*% Sigma_21)
+        if(!is.null(wt)) {
+            T2.pat[!var.idx, !var.idx] <-
+                T2.pat[!var.idx, !var.idx] + (T2.p11 * sum(WT))
+        } else {
+            T2.pat[!var.idx, !var.idx] <- 
+                T2.pat[!var.idx, !var.idx] + (T2.p11 * Mp$freq[[p]])
+        }
+
+        # accumulate
+        T1 <- T1 + T1.pat
+        T2 <- T2 + T2.pat
+    }
+
+    list(T1 = T1, T2 = T2)
+}
+
