@@ -26,33 +26,49 @@ lav_object_independence <- function(object, se = FALSE, verbose = FALSE,
         ## FIXME: if test = scaled, we need it anyway?
         lavoptions$se <- "none" 
     }
+
+    # set baseline/h1 to FALSE
+    lavoptions$h1 <- FALSE
+    lavoptions$baseline <- FALSE
+    lavoptions$loglik <- TRUE # eg for multilevel
+    lavoptions$implied <- TRUE #, needed for loglik
+    lavoptions$check <- character(0L)
    
     # ALWAYS do.fit
     lavoptions$do.fit  <- TRUE
 
     # verbose?
-    if(verbose) {
-        lavoptions$verbose <- TRUE
-    } else {
-        lavoptions$verbose <- FALSE
-    }
+    lavoptions$verbose <- verbose
 
     # warn?
-    if(warn) {
-        lavoptions$warn <- TRUE
-    } else {
-        lavoptions$warn <- FALSE
-    }
-
+    lavoptions$warn <- warn
 
     # needed?
     if(any(lavpartable$op == "~1")) lavoptions$meanstructure <- TRUE
 
+    # FIXME: it is crucial that the order of the ov's, as returned by
+    # lavNames() remains the same
+    # so lavNames(object) should equal lavNames(lavpartable)
+    # otherwise, we will use the wrong sample statistics!!!
+    #
+    # this seems ok now, because we first generate the covariances in
+    # lavpartable, and they should be in the right order (unlike the
+    # intercepts)
+
+    if(.hasSlot(object, "h1"))  {
+        lavh1 <- object@h1
+    } else {
+        lavh1 <- lav_h1_logl(lavdata = object@Data,
+                             lavsamplestats = object@SampleStats,
+                             lavoptions = object@Options)
+    }
+ 
     FIT <- lavaan(lavpartable,  
                   slotOptions     = lavoptions,
                   slotSampleStats = object@SampleStats,
                   slotData        = object@Data,
-                  slotCache       = object@Cache)
+                  slotCache       = object@Cache,
+                  sloth1          = lavh1)
 
     FIT
 }
@@ -98,11 +114,20 @@ lav_object_unrestricted <- function(object, se = FALSE, verbose = FALSE,
     # needed?
     if(any(lavpartable$op == "~1")) lavoptions$meanstructure <- TRUE
 
+    if(.hasSlot(object, "h1"))  {
+        lavh1 <- object@h1
+    } else {
+        lavh1 <- lav_h1_logl(lavdata = object@Data,
+                             lavsamplestats = object@SampleStats,
+                             lavoptions = object@Options)
+    }
+
     FIT <- lavaan(lavpartable,
                   slotOptions     = lavoptions,
                   slotSampleStats = object@SampleStats,
                   slotData        = object@Data,
-                  slotCache       = object@Cache)
+                  slotCache       = object@Cache,
+                  sloth1          = lavh1)
 
     FIT
 }
@@ -116,8 +141,38 @@ lav_object_extended <- function(object, add = NULL,
                                 do.fit = FALSE) {
 
     # partable original model
-    partable <- object@ParTable[c("lhs","op","rhs","group","free",
-                                  "exo","label","plabel")] # do we need 'exo'?
+    partable <- object@ParTable[c("lhs", "op", "rhs", "free", "exo", "label",
+                                 "plabel")]
+
+    # always add block/group/level
+    if(!is.null(object@ParTable$group)) {
+        partable$group <- object@ParTable$group
+    } else {
+        partable$group <- rep(1L, length(partable$lhs))
+    }
+    if(!is.null(object@ParTable$level)) {
+        partable$level <- object@ParTable$level
+    } else {
+        partable$level <- rep(1L, length(partable$lhs))
+    }
+    if(!is.null(object@ParTable$block)) {
+        partable$block <- object@ParTable$block
+    } else {
+        partable$block <- rep(1L, length(partable$lhs))
+    }
+    
+    # TDJ: Added to prevent error when lav_partable_merge() is called below.
+    #      Problematic if object@ParTable is missing one of the requested slots,
+    #      which returns a NULL slot with a missing <NA> name.  For example: 
+    #        example(cfa)
+    #        lav_partable_independence(lavdata = fit@Data, lavpta = fit@pta,
+    #                                  lavoptions = lavInspect(fit, "options"))
+    #     Has no "label" or "plabel" elements.
+    empties <- which(sapply(partable, is.null))
+    if(length(empties)) {
+        partable[empties] <- NULL
+    }
+    
     if(all.free) {
         partable$user <- rep(1L, length(partable$lhs))
         non.free.idx <- which(partable$free == 0L & partable$op != "==" &
@@ -129,8 +184,9 @@ lav_object_extended <- function(object, add = NULL,
  
     # replace 'start' column, since lav_model will fill these in in GLIST
     partable$start <- parameterEstimates(object, remove.system.eq = FALSE,
-                          remove.def = FALSE,
-                          remove.eq = FALSE, remove.ineq = FALSE)$est
+                                         remove.def = FALSE,
+                                         remove.eq = FALSE, 
+                                         remove.ineq = FALSE)$est
 
     # add new parameters, extend model
     if(is.list(add)) {
@@ -139,8 +195,27 @@ lav_object_extended <- function(object, add = NULL,
                   !is.null(add$rhs))
         ADD <- add
     } else if(is.character(add)) {
-        ADD <- lavaanify(add, ngroups = object@Data@ngroups)
-        ADD <- ADD[,c("lhs","op","rhs","group","user","label")]
+        ngroups <- lav_partable_ngroups(partable)
+        ADD.orig <- lavaanify(add, ngroups = ngroups)
+        ADD <- ADD.orig[,c("lhs","op","rhs","user","label")] # minimum
+
+        # always add block/group/level
+        if(!is.null(ADD.orig$group)) {
+            ADD$group <- ADD.orig$group
+        } else {
+            ADD$group <- rep(1L, length(ADD$lhs))
+        }
+        if(!is.null(ADD.orig$level)) {
+            ADD$level <- ADD.orig$level
+        } else {
+            ADD$level <- rep(1L, length(ADD$lhs))
+        }
+        if(!is.null(ADD.orig$block)) {
+            ADD$block <- ADD.orig$block
+        } else {
+            ADD$block <- rep(1L, length(ADD$lhs))
+        }
+        
         remove.idx <- which(ADD$user == 0)
         if(length(remove.idx) > 0L) {
             ADD <- ADD[-remove.idx,]
@@ -171,28 +246,43 @@ lav_object_extended <- function(object, add = NULL,
     lavoptions <- object@Options
 
     # verbose?
-    if(verbose) {
-        lavoptions$verbose <- TRUE
-    } else {
-        lavoptions$verbose <- FALSE
-    }
+    lavoptions$verbose <- verbose
 
     # warn?
-    if(warn) {
-        lavoptions$warn <- TRUE
-    } else {
-        lavoptions$warn <- FALSE
-    }
+    lavoptions$warn <- warn
+
+    # do.fit?
+    lavoptions$do.fit <- do.fit
 
     # needed?
     if(any(LIST$op == "~1")) lavoptions$meanstructure <- TRUE
 
+    if(.hasSlot(object, "h1"))  {
+        lavh1 <- object@h1
+    } else {
+        # old object -- for example 'usemmodelfit' in package 'pompom'
+
+        # add a few fields 
+        lavoptions$h1 <- FALSE
+        lavoptions$implied <- FALSE
+        lavoptions$baseline <- FALSE
+        lavoptions$loglik <- FALSE
+ 
+        # add a few slots
+        object@Data@weights <- vector("list", object@Data@ngroups)
+        object@Model@estimator <- object@Options$estimator
+
+        lavh1 <- lav_h1_logl(lavdata = object@Data,
+                             lavsamplestats = object@SampleStats,
+                             lavoptions = object@Options)
+    }
+
     FIT <- lavaan(LIST,
-                  do.fit          = do.fit,
                   slotOptions     = lavoptions,
                   slotSampleStats = object@SampleStats,
                   slotData        = object@Data,
-                  slotCache       = object@Cache)
+                  slotCache       = object@Cache,
+                  sloth1          = lavh1)
 
     FIT
 }
