@@ -1,12 +1,165 @@
 # initial version YR 02/08/2010
 
+# YR 28 Jan 2017: add lavOptions(), lav_options_default()
+
+# public function
+lavOptions <- function(x = NULL, default = NULL, mimic = "lavaan") {
+
+    lavoptions <- lav_options_default(mimic = mimic)
+
+    # selection only
+    if(!is.null(x)) {
+        if(is.character(x)) {
+            # lower case only
+            x <- tolower(x)
+
+            # check if x is in names(lavoptions)
+            not.ok <- which(!x %in% names(lavoptions))
+            if(length(not.ok) > 0L) {
+                # only warn if multiple options were requested
+                if(length(x) > 1L) {
+                    warning("lavaan WARNING: option `", x[not.ok], 
+                            "' not available")
+                }
+                x <- x[ -not.ok ]
+            }
+
+            # return requested option(s)
+            if(length(x) == 0L) {
+                return(default)
+            } else {
+                lavoptions[x]
+            }
+        } else {
+            stop("lavaan ERROR: `x' must be a character string")
+        }
+    } else {
+        lavoptions
+    }
+}
+
+# set the default options (including unspecified values "default")
+lav_options_default <- function(mimic = "lavaan") {
+
+    opt <- list(model.type         = "sem",
+
+                # global
+                mimic              = "lavaan",
+
+                # model modifiers
+                meanstructure      = "default",
+                int.ov.free        = FALSE,
+                int.lv.free        = FALSE,
+                conditional.x      = "default", # or FALSE?
+                fixed.x            = "default", # or FALSE?
+                orthogonal         = FALSE,
+                std.lv             = FALSE,
+                parameterization   = "default",
+
+                auto.fix.first     = FALSE,
+                auto.fix.single    = FALSE,
+                auto.var           = FALSE,
+                auto.cov.lv.x      = FALSE,
+                auto.cov.y         = FALSE,
+                auto.th            = FALSE,
+                auto.delta         = FALSE,
+
+                # full data
+                std.ov             = FALSE,
+                missing            = "default",
+
+                # summary data
+                sample.cov.rescale = "default",
+                ridge              = FALSE,
+                ridge.x            = FALSE,
+                ridge.constant     = "default",
+                ridge.constant.x   = 1e-5,
+
+                # multiple groups
+                group              = NULL,
+                group.label        = NULL,
+                group.equal        = '',
+                group.partial      = '',
+                group.w.free       = FALSE,
+
+                # clusters
+                cluster            = NULL,
+                level.label        = NULL,
+
+                # sampling weights 
+                sampling.weights   = NULL,
+
+                # estimation
+                estimator              = "default",
+                likelihood             = "default",
+                link                   = "default",
+                representation         = "default",
+                do.fit                 = TRUE,
+
+                # inference
+                information            = "default",
+                h1.information         = "structured",
+                #h1.information.se      = "structured",
+                #h1.information.test    = "structured",
+                se                     = "default",
+                test                   = "default",
+                bootstrap              = 1000L,
+                observed.information   = "hessian",
+                gamma.n.minus.one      = FALSE,
+                #gamma.unbiased         = FALSE,
+
+                # optimization
+                control                = list(),
+                optim.method           = "nlminb",
+                optim.method.cor       = "nlminb",
+                optim.force.converged  = FALSE,
+                optim.gradient         = "analytic",
+                optim.init_nelder_mead = FALSE,
+                em.iter.max            = 10000L,
+                em.fx.tol              = 1e-08,
+                em.dx.tol              = 1e-04,
+                em.zerovar.offset      = 0.0001,
+
+                # numerical integration
+                integration.ngh        = 21L,
+
+                # parallel
+                parallel               = "no",
+                ncpus                  = 1L,
+                cl                     = NULL,
+                iseed                  = NULL,
+
+                # zero values
+                zero.add               = "default",
+                zero.keep.margins      = "default",
+                zero.cell.warn         = FALSE, # since 0.6-1
+
+                # starting values
+                start                  = "default",
+
+                # sanity checks
+                check                  = c("start", "post"),
+
+                # more models/info
+                h1                     = TRUE,
+                baseline               = TRUE,
+                implied                = TRUE,
+                loglik                 = TRUE,
+
+                # verbosity
+                verbose                = FALSE,
+                warn                   = TRUE,
+                debug                  = FALSE)
+
+    opt
+}
+
 # this function collects and checks the user-provided options/arguments, 
 # and fills in the "default" values, or changes them in an attempt to
 # produce a consistent set of values...
 #
 # returns a list with the named options 
-
-lav_options_set <- function(opt = formals(lavaan)) {
+lav_options_set <- function(opt = NULL) {
 
     if(opt$debug) { cat("lavaan DEBUG: lavaanOptions IN\n"); str(opt) }
 
@@ -21,7 +174,10 @@ lav_options_set <- function(opt = formals(lavaan)) {
     opt <- lapply(opt, function(x) { if(is.character(x)) tolower(x) else x})
     # except group,group.partial, which may contain capital letters
     opt$group <- opt.old$group
+    opt$group.label <- opt.old$group.label
     opt$group.partial <- opt.old$group.partial
+    opt$cluster <- opt.old$cluster
+    
 
     # do.fit implies se="none and test="none" (unless not default)
     if(!opt$do.fit) {
@@ -44,7 +200,7 @@ lav_options_set <- function(opt = formals(lavaan)) {
         cat("Warning: mimic=\"LISREL\" is not ready yet. Using EQS instead.\n")
         opt$mimic <- "EQS"
     } else {
-        stop("mimic must be \"lavaan\", \"Mplus\" or \"EQS\" \n")
+        stop("lavaan ERROR: mimic must be \"lavaan\", \"Mplus\" or \"EQS\" \n")
     }
 
     # group.equal and group.partial
@@ -68,7 +224,7 @@ lav_options_set <- function(opt = formals(lavaan)) {
                                          "lv.variances", "lv.covariances"))) {
         # nothing to do 
     } else {
-        stop("unknown value for `group.equal' argument: ",
+        stop("lavaan ERROR: unknown value for `group.equal' argument: ",
              opt$group.equal, "\n")
     }
     if(is.null(opt$group.partial) || nchar(opt$group.partial) == 0L) {
@@ -99,7 +255,49 @@ lav_options_set <- function(opt = formals(lavaan)) {
               opt$representation == "bentler-weeks") {
         opt$representation <- "EQS"
     } else {
-        stop("representation must be \"LISREL\" or \"EQS\" \n")
+        stop("lavaan ERROR: representation must be \"LISREL\" or \"EQS\" \n")
+    }
+
+    # multilevel
+    # brute-force override (for now)
+    if(opt$multilevel) {
+        opt$meanstructure <- TRUE
+        opt$missing <- "listwise"
+
+        # test
+        if(opt$test == "default") {
+            opt$test <- "standard"
+        } else if(opt$test %in% c("none", "standard","yuan.bentler")) {
+            # nothing to do
+        } else if(opt$est == "robust") {
+            opt$test <- "yuan.bentler"
+        } else {
+            stop("lavaan ERROR: `test' argument must one of \"none\", \"standard\" or \"yuan.bentler\" in the multilevel case")
+        }
+
+        # se
+        if(opt$se == "default") {
+            opt$se <- "standard"
+        } else if(opt$se %in% c("none", "standard", "robust.huber.white")) {
+            # nothing to do
+        } else if(opt$se == "robust.sem") {
+            opt$se <- "robust.huber.white"
+        } else {
+            stop("lavaan ERROR: `se' argument must one of \"none\", \"standard\" or \"robust.huber.white\" in the multilevel case")
+        }
+
+        # information
+        if(opt$information == "default") {
+            opt$information <- "observed"
+        }
+        #} else if(opt$information %in% c("observed", "first.order")) {
+        #    # nothing to do
+        #} else {
+        #    stop("lavaan ERROR: `information' argument must be \"observed\" in the multilevel case (for now)")
+        #}
+
+        #opt$fixed.x = FALSE
+        #opt$control <- list(gradient = "numerical")
     }
 
 
@@ -119,7 +317,29 @@ lav_options_set <- function(opt = formals(lavaan)) {
                                 "uls", "ulsm", "ulsmv", "pml")) {
             stop("lavaan ERROR: missing=\"ml\" is not allowed for estimator MLM, MLMV, GLS, ULS, ULSM, ULSMV, DWLS, WLS, WLSM, WLSMV, PML")
         }
-    } else if(opt$missing %in% c("two.stage", "listwise")) {
+    } else if(opt$missing %in% c("two.stage", "twostage", "two-stage",
+                                 "two.step",  "twostep",  "two-step")) {
+        opt$missing <- "two.stage"
+        if(opt$categorical) {
+            stop("lavaan ERROR: missing=\"two.stage\" not available in the categorical setting")
+        }
+        if(opt$estimator %in% c("mlm", "mlmv", "gls", "wls", "wlsm", "wlsmv",
+                                "uls", "ulsm", "ulsmv", "pml", "mml")) {
+            stop("lavaan ERROR: missing=\"two.stage\" is not allowed for estimator MLM, MLMV, GLS, ULS, ULSM, ULSMV, DWLS, WLS, WLSM, WLSMV, PML, MML")
+        }
+    } else if(opt$missing %in% c("robust.two.stage", "robust.twostage", 
+                                 "robust.two-stage", "robust-two-stage",
+                                 "robust.two.step",  "robust.twostep",  
+                                 "robust-two-step")) {
+        opt$missing <- "robust.two.stage"
+        if(opt$categorical) {
+            stop("lavaan ERROR: missing=\"robust.two.stage\" not available in the categorical setting")
+        }
+        if(opt$estimator %in% c("mlm", "mlmv", "gls", "wls", "wlsm", "wlsmv",
+                                "uls", "ulsm", "ulsmv", "pml", "mml")) {
+            stop("lavaan ERROR: missing=\"robust.two.stage\" is not allowed for estimator MLM, MLMV, GLS, ULS, ULSM, ULSMV, DWLS, WLS, WLSM, WLSMV, PML, MML")
+        }
+    } else if(opt$missing == "listwise") {
         # nothing to do
     } else if(opt$missing == "pairwise") {
         # nothing to do
@@ -137,12 +357,17 @@ lav_options_set <- function(opt = formals(lavaan)) {
     } else if(opt$missing == "available_cases") {
         opt$missing <- "available.cases"
     } else {
-        stop("unknown value for `missing' argument: ", opt$missing, "\n")
+        stop("lavaan ERROR: unknown value for `missing' argument: ", opt$missing, "\n")
     }
 
     # default test statistic
     if(opt$test == "default") {
-        opt$test <- "standard"
+        if(opt$missing == "two.stage" ||
+           opt$missing == "robust.two.stage") {
+            opt$test <- "satorra.bentler"
+        } else {
+            opt$test <- "standard"
+        }
     } else if(opt$test %in% c("none", "standard")) {
         # nothing to do
     } else if(opt$test == "satorra" || 
@@ -157,6 +382,10 @@ lav_options_set <- function(opt = formals(lavaan)) {
               opt$test == "yuan.bentler" ||
               opt$test == "yuan-bentler") {
         opt$test <- "yuan.bentler"
+    } else if(opt$test == "yuan.bentler.mplus" ||
+              opt$test == "yuan-bentler.mplus" ||
+              opt$test == "yuan-bentler-mplus") {
+        opt$test <- "yuan.bentler.mplus"
     } else if(opt$test == "m.adjusted" ||
               opt$test == "m" ||
               opt$test == "mean.adjusted" ||
@@ -180,8 +409,8 @@ lav_options_set <- function(opt = formals(lavaan)) {
               opt$test == "bollen-stine") {
         opt$test <- "bollen.stine"
     } else {
-        stop("`test' argument must one of \"none\", \"standard\",
-            \"satorra.bentler\", \"yuan.bentler\",
+        stop("lavaan ERROR: `test' argument must one of \"none\", \"standard\",
+            \"satorra.bentler\", \"yuan.bentler\", \"yuan.bentler.mplus\",
             \"mean.var.adjusted\", \"scaled.shifted\",
             \"bollen.stine\", or \"bootstrap\"")
     }
@@ -202,23 +431,78 @@ lav_options_set <- function(opt = formals(lavaan)) {
         opt$missing <- "listwise"
     }
 
+    # missing = "two.stage"
+    if(opt$missing == "two.stage" ||
+       opt$missing == "robust.two.stage") {
+        opt$meanstructure <- TRUE
+        # se
+        if(opt$se == "default") {
+            if(opt$missing == "two.stage") {
+                opt$se <- "two.stage"
+            } else {
+                opt$se <- "robust.two.stage"
+            }
+        } else if(opt$missing == "two.stage" && 
+                  opt$se      == "two.stage") {
+            # nothing to do
+        } else if(opt$missing == "robust.two.stage" &&
+                  opt$se      == "robust.two.stage") {
+            # nothing to do
+        } else {
+            warning("lavaan WARNING: se will be set to ",
+                     dQuote(opt$missing), " if missing = ",
+                     dQuote(opt$missing) )
+            opt$se <- opt$missing
+        }
+        # information
+        if(opt$information == "default") {
+            # for both two.stage and robust.two.stage
+            opt$information <- "observed"
+        } else if(opt$information == "first.order") {
+            warning("lavaan WARNING: information will be set to ",
+                     dQuote("observed"), " if missing = ",
+                     dQuote(opt$missing) )
+            opt$information <- "observed"
+        }
+        # observed.information (ALWAYS "h1" for now)
+        opt$observed.information <- "h1"
+        # test
+        if(opt$test == "default" ||
+           opt$test == "satorra.bentler") {
+            opt$test <- "satorra.bentler"
+        } else {
+            warning("lavaan WARNING: test will be set to ",
+                     dQuote("satorra.bentler"), " if missing = ",
+                     dQuote(opt$missing) )
+            opt$test <- "satorra.bentler"
+        }
+    }
+
+
+
     # meanstructure
     if(is.logical(opt$meanstructure)) {
         if(opt$meanstructure == FALSE) {
             # user explicitly wants meanstructure == FALSE
             # check for conflicting arguments
-            if(opt$estimator %in% c("mlm", "mlmv", "mlr", "mlf", "ulsm", "ulsmv", "wlsm", "wlsmv", "pml"))
-                warning("lavaan WARNING: estimator forces meanstructure = TRUE")
-            if(opt$missing == "ml")
+            #if(opt$estimator %in% c("mlm", "mlmv", "mlr", "mlf", "ulsm", "ulsmv", "wlsm", "wlsmv", "pml")) {
+            #    warning("lavaan WARNING: estimator forces meanstructure = TRUE")
+            #}
+            if(opt$missing %in% c("ml", "two.stage")) {
                 warning("lavaan WARNING: missing argument forces meanstructure = TRUE")
+            }
         }
     } else if(opt$meanstructure == "default") {
         # by default: no meanstructure!
-        opt$meanstructure <- FALSE
+        if(opt$estimator == "pml") {
+            opt$meanstructure <- TRUE
+        } else {
+            opt$meanstructure <- FALSE
+        }
         # unless there is a group argument? (added since 0.4-10)
-        if(!is.null(opt$group)) opt$meanstructure <- TRUE
+        # if(!is.null(opt$group)) opt$meanstructure <- TRUE
     } else {
-        stop("meanstructure must be TRUE, FALSE or \"default\"\n")
+        stop("lavaan ERROR: meanstructure must be TRUE, FALSE or \"default\"\n")
     }
 
     # estimator and se
@@ -231,10 +515,11 @@ lav_options_set <- function(opt = formals(lavaan)) {
 
     # default estimator
     if(opt$estimator == "default") {
-        if(opt$categorical) 
+        if(opt$categorical) {
             opt$estimator <- "wlsmv"
-        else
+        } else {
             opt$estimator <- "ml"
+        }
     }
 
     # backwards compatibility (0.4 -> 0.5)
@@ -245,22 +530,30 @@ lav_options_set <- function(opt = formals(lavaan)) {
         opt$estimator <- "ML"
         if(opt$se == "default") {
             opt$se <- "standard"
-        } else if(opt$se == "first.order" || 
-                  opt$se == "bootstrap"   ||
-                  opt$se == "none"        || 
-                  opt$se == "external"    ||
-                  opt$se == "standard"    ||
-                  opt$se == "robust.huber.white"  || 
-                  opt$se == "robust.sem") {
+        } else if(opt$se %in% c("bootstrap", "none", 
+                  "external", "standard", "robust.huber.white",
+                  "two.stage", "robust.two.stage", "robust.sem")) {
             # nothing to do
+        } else if(opt$se == "first.order") {
+            # backwards compatibility
+            opt$se <- "standard"
+            opt$information <- "first.order"
+        } else if(opt$se == "observed") {
+            opt$se <- "standard"
+            opt$information <- "observed"
+        } else if(opt$se == "expected") {
+            opt$se <- "standard"
+            opt$information <- "expected"
         } else if(opt$se == "robust") {
             if(opt$missing == "ml") {
                 opt$se <- "robust.huber.white"
+            } else if(opt$missing == "two.stage") {
+                opt$se <- "robust.two.stage"
             } else {
                 opt$se <- "robust.sem"
             }
         } else {
-            stop("unknown value for `se' argument when estimator is ML: ", 
+            stop("lavaan ERROR: unknown value for `se' argument when estimator is ML: ", 
                  opt$se, "\n")
         }
 
@@ -278,27 +571,43 @@ lav_options_set <- function(opt = formals(lavaan)) {
             }
         }
         opt$estimator <- "ML"
-        opt$meanstructure <- TRUE
-        if(opt$se == "bootstrap") stop("use ML estimator for bootstrap")
-        if(opt$se != "none" && opt$se != "external") opt$se <- "robust.sem"
-        if(!(opt$information %in% c("expected", "default"))) {
-            warning("lavaan WARNING: information will be set to ",
-                    dQuote("expected"), " for estimator = ", 
-                    dQuote(toupper(est.orig)) )
+        #opt$meanstructure <- TRUE
+        if(opt$se == "bootstrap") {
+            stop("lavaan ERROR: use ML estimator for bootstrap")
         }
-        opt$information <- "expected"
+        if(opt$se != "none" && opt$se != "external") opt$se <- "robust.sem"
+        #if(!(opt$information %in% c("expected", "default"))) {
+        #    warning("lavaan WARNING: information will be set to ",
+        #            dQuote("expected"), " for estimator = ", 
+        #            dQuote(toupper(est.orig)) )
+        #}
+        #opt$information <- "expected"
+        # in 0.6, we allow for information = "observed" as well
         opt$missing <- "listwise"
     } else if(opt$estimator == "mlf") {
         opt$estimator <- "ML"
-        opt$meanstructure <- TRUE
-        if(opt$se == "bootstrap") stop("use ML estimator for bootstrap")
-        if(opt$se != "none" && opt$se != "external") opt$se <- "first.order"
+        #opt$meanstructure <- TRUE
+        if(opt$se == "bootstrap") {
+            stop("lavaan ERROR: use ML estimator for bootstrap")
+        }
+        if(opt$se != "none" && opt$se != "external") {
+            opt$se <- "standard"
+            opt$information <- "first.order"    
+        }
     } else if(opt$estimator == "mlr") {
         opt$estimator <- "ML"
-        if(opt$se == "bootstrap") stop("use ML estimator for bootstrap")
+        if(opt$se == "bootstrap") {
+            stop("lavaan ERROR: use ML estimator for bootstrap")
+        }
         if(opt$se != "none" && opt$se != "external") opt$se <- "robust.huber.white"
-        if(opt$test != "none" && opt$se != "external") opt$test <- "yuan.bentler"
-        opt$meanstructure <- TRUE
+        if(opt$test != "none" && opt$se != "external") {
+            if(opt$mimic == "Mplus" || opt$test == "yuan.bentler.mplus") {
+                opt$test <- "yuan.bentler.mplus"
+            } else {
+                opt$test <- "yuan.bentler.mplus" # for now
+            }
+        }
+        #opt$meanstructure <- TRUE
     } else if(opt$estimator == "gls") {
         opt$estimator <- "GLS"
         if(opt$se == "default" || opt$se == "standard") {
@@ -308,11 +617,11 @@ lav_options_set <- function(opt = formals(lavaan)) {
                   opt$se == "external") {
             # nothing to do
         } else {
-            stop("invalid value for `se' argument when estimator is GLS: ", 
+            stop("lavaan ERROR: invalid value for `se' argument when estimator is GLS: ", 
                  opt$se, "\n")
         }
         if(!opt$test %in% c("standard","none")) {
-            stop("invalid value for `test' argument when estimator is GLS: ", 
+            stop("lavaan ERROR: invalid value for `test' argument when estimator is GLS: ", 
                  opt$test, "\n")
         }
         opt$missing <- "listwise"       
@@ -325,11 +634,11 @@ lav_options_set <- function(opt = formals(lavaan)) {
                   opt$se == "external") {
             # nothing to do
         } else {
-            stop("invalid value for `se' argument when estimator is NTRLS: ",
+            stop("lavaan ERROR: invalid value for `se' argument when estimator is NTRLS: ",
                  opt$se, "\n")
         }
         if(!opt$test %in% c("standard","none")) {
-            stop("invalid value for `test' argument when estimator is NTRLS: ",
+            stop("lavaan ERROR: invalid value for `test' argument when estimator is NTRLS: ",
                  opt$test, "\n")
         }
         opt$missing <- "listwise"
@@ -346,11 +655,11 @@ lav_options_set <- function(opt = formals(lavaan)) {
         } else if(opt$se == "robust") {
             opt$se <- "robust.sem"
         } else {
-            stop("invalid value for `se' argument when estimator is WLS: ", 
+            stop("lavaan ERROR: invalid value for `se' argument when estimator is WLS: ", 
                  opt$se, "\n")
         }
         if(!opt$test %in% c("standard","none")) {
-            stop("invalid value for `test' argument when estimator is WLS: ", 
+            stop("lavaan ERROR: invalid value for `test' argument when estimator is WLS: ", 
                  opt$test, "\n")
         }
         #opt$missing <- "listwise"
@@ -367,33 +676,54 @@ lav_options_set <- function(opt = formals(lavaan)) {
         } else if(opt$se == "robust") {
             opt$se <- "robust.sem"
         } else {
-            stop("invalid value for `se' argument when estimator is DWLS: ",
+            stop("lavaan ERROR: invalid value for `se' argument when estimator is DWLS: ",
                  opt$se, "\n")
         }
         if(!opt$test %in% c("standard","none","satorra.bentler", 
                             "mean.adjusted",
                             "mean.var.adjusted","scaled.shifted")) {
-            stop("invalid value for `test' argument when estimator is DWLS: ",
+            stop("lavaan ERROR: invalid value for `test' argument when estimator is DWLS: ",
                  opt$test, "\n")
         }
         #opt$missing <- "listwise"
     } else if(opt$estimator == "wlsm") {
         opt$estimator <- "DWLS"
-        if(opt$se == "bootstrap") stop("use (D)WLS estimator for bootstrap")
+        if(opt$se == "bootstrap") { 
+            stop("lavaan ERROR: use (D)WLS estimator for bootstrap")
+        }
         if(opt$se != "none" && opt$se != "external") opt$se <- "robust.sem"
-        if(opt$test != "none") opt$test <- "satorra.bentler"
+        if(opt$test %in% c("mean.var.adjusted", "satorra.bentler",
+                           "scaled.shifted")) {
+            # nothing to do
+        } else if(opt$test != "none") {
+            opt$test <- "satorra.bentler"
+        }
         #opt$missing <- "listwise"
     } else if(opt$estimator == "wlsmv") {
         opt$estimator <- "DWLS"
-        if(opt$se == "bootstrap") stop("use (D)WLS estimator for bootstrap")
+        if(opt$se == "bootstrap") {
+            stop("lavaan ERROR: use (D)WLS estimator for bootstrap")
+        }
         if(opt$se != "none" && opt$se != "external") opt$se <- "robust.sem"
-        if(opt$test != "none") opt$test <- "scaled.shifted"
+        if(opt$test %in% c("mean.var.adjusted", "satorra.bentler",
+                           "scaled.shifted")) {
+            # nothing to do
+        } else if(opt$test != "none") {
+            opt$test <- "scaled.shifted"
+        }
         #opt$missing <- "listwise"
     } else if(opt$estimator == "wlsmvs") {
         opt$estimator <- "DWLS"
-        if(opt$se == "bootstrap") stop("use (D)WLS estimator for bootstrap")
+        if(opt$se == "bootstrap") {
+            stop("lavaan ERROR: use (D)WLS estimator for bootstrap")
+        }
         if(opt$se != "none" && opt$se != "external") opt$se <- "robust.sem"
-        if(opt$test != "none") opt$test <- "mean.var.adjusted"
+        if(opt$test %in% c("mean.var.adjusted", "satorra.bentler",
+                           "scaled.shifted")) {
+            # nothing to do
+        } else if(opt$test != "none") {
+            opt$test <- "mean.var.adjusted"
+        }
         #opt$missing <- "listwise"
     } else if(opt$estimator == "uls") {
         opt$estimator <- "ULS"
@@ -408,33 +738,54 @@ lav_options_set <- function(opt = formals(lavaan)) {
         } else if(opt$se == "robust") {
             opt$se <- "robust.sem"
         } else {
-            stop("invalid value for `se' argument when estimator is ULS: ", 
+            stop("lavaan ERROR: invalid value for `se' argument when estimator is ULS: ", 
                  opt$se, "\n")
         }
         if(!opt$test %in% c("standard","none", "satorra.bentler",
                             "mean.adjusted",
                             "mean.var.adjusted","scaled.shifted")) {
-            stop("invalid value for `test' argument when estimator is ULS: ",
+            stop("lavaan ERROR: invalid value for `test' argument when estimator is ULS: ",
                  opt$test, "\n")
         }
         #opt$missing <- "listwise"
     } else if(opt$estimator == "ulsm") {
         opt$estimator <- "ULS"
-        if(opt$se == "bootstrap") stop("use ULS estimator for bootstrap")
+        if(opt$se == "bootstrap") {
+            stop("lavaan ERROR: use ULS estimator for bootstrap")
+        }
         if(opt$se != "none" && opt$se != "external") opt$se <- "robust.sem"
-        if(opt$test != "none") opt$test <- "satorra.bentler"
+        if(opt$test %in% c("mean.var.adjusted", "satorra.bentler",
+                           "scaled.shifted")) {
+            # nothing to do
+        } else if(opt$test != "none") {
+            opt$test <- "satorra.bentler"
+        }
         #opt$missing <- "listwise"
     } else if(opt$estimator == "ulsmv") {
         opt$estimator <- "ULS"
-        if(opt$se == "bootstrap") stop("use ULS estimator for bootstrap")
+        if(opt$se == "bootstrap") {
+            stop("lavaan ERROR: use ULS estimator for bootstrap")
+        }
         if(opt$se != "none" && opt$se != "external") opt$se <- "robust.sem"
-        if(opt$test != "none") opt$test <- "scaled.shifted"
+        if(opt$test %in% c("mean.var.adjusted", "satorra.bentler",
+                           "scaled.shifted")) {
+            # nothing to do
+        } else if(opt$test != "none") {
+            opt$test <- "scaled.shifted"
+        }
         #opt$missing <- "listwise"
     } else if(opt$estimator == "ulsmvs") {
         opt$estimator <- "ULS"
-        if(opt$se == "bootstrap") stop("use ULS estimator for bootstrap")
+        if(opt$se == "bootstrap") {
+            stop("lavaan ERROR: use ULS estimator for bootstrap")
+        }
         if(opt$se != "none" && opt$se != "external") opt$se <- "robust.sem"
-        if(opt$test != "none") opt$test <- "mean.var.adjusted"
+        if(opt$test %in% c("mean.var.adjusted", "satorra.bentler",
+                           "scaled.shifted")) {
+            # nothing to do
+        } else if(opt$test != "none") {
+            opt$test <- "mean.var.adjusted"
+        }
         #opt$missing <- "listwise"
     } else if(opt$estimator == "pml") {
         opt$estimator <- "PML"
@@ -487,7 +838,7 @@ lav_options_set <- function(opt = formals(lavaan)) {
             opt$test <- "none"
         }
     } else {
-        stop("unknown value for `estimator' argument: ", opt$estimator, "\n")
+        stop("lavaan ERROR: unknown value for `estimator' argument: ", opt$estimator, "\n")
     }
 
 
@@ -511,7 +862,7 @@ lav_options_set <- function(opt = formals(lavaan)) {
     # likelihood approach (wishart or normal) + sample.cov.rescale
     if(!opt$estimator %in% c("ML", "REML", "PML", "FML","NTRLS")) {
         if(opt$likelihood != "default") {
-            stop("likelihood argument is only relevant if estimator = ML")
+            stop("lavaan ERROR: likelihood argument is only relevant if estimator = ML")
         }
         if(opt$sample.cov.rescale == "default") {
             opt$sample.cov.rescale <- FALSE
@@ -531,7 +882,7 @@ lav_options_set <- function(opt = formals(lavaan)) {
         } else if(opt$likelihood == "wishart" || opt$likelihood == "normal") {
             # nothing to do
         } else {
-            stop("invalid value for `likelihood' argument: ", 
+            stop("lavaan ERROR: invalid value for `likelihood' argument: ", 
                  opt$likelihood, "\n")
         }
 
@@ -541,7 +892,7 @@ lav_options_set <- function(opt = formals(lavaan)) {
                 opt$sample.cov.rescale <- TRUE
             }
         } else if(!is.logical(opt$sample.cov.rescale)) {
-            stop("sample.cov.rescale must be either \"default\", TRUE, or FALSE")
+            stop("lavaan ERROR: sample.cov.rescale must be either \"default\", TRUE, or FALSE")
         } else {
             # nothing to do
         }
@@ -551,23 +902,42 @@ lav_options_set <- function(opt = formals(lavaan)) {
     if(opt$information == "default") {
         if(opt$missing == "ml"     || 
            opt$se == "robust.huber.white"  || 
-           opt$se == "first.order" ||
-           nchar(opt$constraints) > 0L) {
+           opt$se == "first.order") {
+           #nchar(opt$constraints) > 0L) {
             opt$information <- "observed"
         } else {
             opt$information <- "expected"
         }
-    } else if(opt$information %in% c("observed", "expected")) {
+    } else if(opt$information %in% c("observed", "expected", "first.order")) {
         # nothing to do
     } else {
-        stop("information must be either \"expected\" or \"observed\"\n")
+        stop("lavaan ERROR: information must be either \"expected\", \"observed\", or \"first.order\"\n")
     }
 
-    # check information if se == "robust.sem"
-    if(opt$se == "robust.sem" && opt$information == "observed") {
-        warning("lavaan WARNING: information will be set to ",
-                dQuote("expected"), " for se = ", dQuote(opt$se))
-        opt$information <- "expected"
+    if(opt$h1.information == "structured" ||
+       opt$h1.information == "unstructured") {
+        # nothing to do
+    } else {
+        stop("lavaan ERROR: h1.information must be either \"structured\" or \"unstructured\"\n")
+    }
+    #if(opt$h1.information.test == "structured" ||
+    #   opt$h1.information.test == "unstructured") {
+    #    # nothing to do
+    #} else {
+    #    stop("lavaan ERROR: h1.information.se must be either \"structured\" or \"unstructured\"\n")
+    #}
+
+    # check information if estimator is uls/wls and friends
+    if(opt$estimator %in% c("ULS", "WLS", "DWLS")) {
+        if(opt$information != "expected") {
+            warning("lavaan WARNING: information will be set to ",
+                    dQuote("expected"), " for estimator = ", dQuote(opt$estimator))
+            opt$information <- "expected"
+        }
+        #if(opt$h1.information != "structured") { # default value
+        #    warning("lavaan WARNING: h1.information is not used if estimator = ", dQuote(opt$estimator))
+        opt$h1.information <- "unstructured"
+        #}
     }
 
     # conditional.x
@@ -617,14 +987,14 @@ lav_options_set <- function(opt = formals(lavaan)) {
        "means" %in% opt$group.equal) {
         opt$meanstructure <- TRUE
     }
-    if(opt$se == "robust.huber.white" || 
-       opt$se == "robust.sem" ||
-       opt$test == "satorra.bentler" ||
-       opt$test == "mean.var.adjusted" ||
-       opt$test == "scaled.shifted" ||
-       opt$test == "yuan.bentler") {
-        opt$meanstructure <- TRUE
-    }
+    #if(opt$se == "robust.huber.white" || 
+    #   opt$se == "robust.sem" ||
+    #   opt$test == "satorra.bentler" ||
+    #   opt$test == "mean.var.adjusted" ||
+    #   opt$test == "scaled.shifted" ||
+    #   opt$test == "yuan.bentler") {
+    #    opt$meanstructure <- TRUE
+    #}
     stopifnot(is.logical(opt$meanstructure))
     stopifnot(is.logical(opt$verbose))
     stopifnot(is.logical(opt$warn))
