@@ -15,18 +15,19 @@ setMethod("predict", "lavaan",
 function(object, newdata = NULL) {
     lavPredict(object = object, newdata = newdata, type="lv", method="EBM",
                fsm = FALSE,
-               optim.method = "nlminb")
+               optim.method = "bfgs")
 })
 
 # main function
 lavPredict <- function(object, type = "lv", newdata = NULL, method = "EBM",
-                       se.fit = FALSE, label = TRUE, fsm = FALSE,
-                       optim.method = "nlminb") {
+                       se = "none", label = TRUE, fsm = FALSE, level = 1L,
+                       optim.method = "bfgs", ETA = NULL) {
 
     stopifnot(inherits(object, "lavaan"))
     lavmodel       <- object@Model
     lavdata        <- object@Data
     lavsamplestats <- object@SampleStats
+    lavimplied     <- object@implied
     lavpta         <- object@pta
 
     # type
@@ -36,6 +37,16 @@ lavPredict <- function(object, type = "lv", newdata = NULL, method = "EBM",
     if(type %in% c("ov","yhat")) 
         type <- "yhat"
 
+    # se?
+    if(se != "none") {
+        if(is.logical(se) && se) {
+            se <- "standard"
+        }
+        if(type != "lv") {
+            stop("lavaan ERROR: standard errors only available if type = \"lv\"")
+        }
+    }
+
     # need full data set supplied
     if(is.null(newdata)) {
         # use internal copy:
@@ -51,20 +62,23 @@ lavPredict <- function(object, type = "lv", newdata = NULL, method = "EBM",
         OV <- lavdata@ov
         newData <- lavData(data        = newdata,
                            group       = lavdata@group,
-                           group.label = lavdata@group.label,
                            ov.names    = lavdata@ov.names,
-                           ordered     = OV$name[ OV$type == "ordered" ],
                            ov.names.x  = lavdata@ov.names.x,
-                           std.ov      = lavdata@std.ov,
-                           missing     = lavdata@missing,
-                           # warn      = FALSE,
+                           ordered     = OV$name[ OV$type == "ordered" ],
+                           lavoptions  = list(std.ov = lavdata@std.ov,
+                                              group.label = lavdata@group.label,
+                                              missing = lavdata@missing,
+                                              warn = FALSE),
                            allow.single.case = TRUE)
         data.obs <- newData@X
         eXo <- newData@eXo
     }
 
     if(type == "lv") {
-
+        if(!is.null(ETA)) {
+            warning("lavaan WARNING: lvs will be predicted here; supplying ETA has no effect")
+        }
+            
         # post fit check (lv pd?)
         ok <- lav_object_post_check(object)
         #if(!ok) {
@@ -73,13 +87,21 @@ lavPredict <- function(object, type = "lv", newdata = NULL, method = "EBM",
 
         out <- lav_predict_eta(lavobject = NULL, lavmodel = lavmodel,
                    lavdata = lavdata, lavsamplestats = lavsamplestats,
+                   lavimplied = lavimplied, se = se, level = level,
                    data.obs = data.obs, eXo = eXo, method = method,
                    fsm = fsm, optim.method = optim.method)
 
-        # remove dummy lv?
+        # extract fsm here
         if(fsm) {
             FSM <- attr(out, "fsm")
         }
+
+        # extract se here
+        if(se != "none") {
+            SE <- attr(out, "se")
+        }
+
+        # remove dummy lv? (removes attr!)
         out <- lapply(seq_len(lavdata@ngroups), function(g) {
                    lv.idx <- c(lavmodel@ov.y.dummy.lv.idx[[g]],
                                lavmodel@ov.x.dummy.lv.idx[[g]])
@@ -87,13 +109,53 @@ lavPredict <- function(object, type = "lv", newdata = NULL, method = "EBM",
                    if(length(lv.idx) > 0L) {
                        ret <- out[[g]][, -lv.idx, drop=FALSE]
                    }
-                   ret 
+                   ret
                })
 
+        if(fsm) {
+            #FSM <- attr(out, "fsm")
+            FSM <- lapply(seq_len(lavdata@ngroups), function(g) {
+                       lv.idx <- c(lavmodel@ov.y.dummy.lv.idx[[g]],
+                                   lavmodel@ov.x.dummy.lv.idx[[g]])
+                       ov.idx <- c(lavmodel@ov.y.dummy.ov.idx[[g]],
+                                   lavmodel@ov.x.dummy.ov.idx[[g]])
+                       ret <- FSM[[g]]
+                       if(length(lv.idx) > 0L) {
+                           ret <- FSM[[g]][-lv.idx, -ov.idx, drop=FALSE]
+                       }
+                       ret
+                   })
+        }
+
+        if(se != "none") {
+            SE <- lapply(seq_len(lavdata@ngroups), function(g) {
+                       lv.idx <- c(lavmodel@ov.y.dummy.lv.idx[[g]],
+                                   lavmodel@ov.x.dummy.lv.idx[[g]])
+                       ret <- SE[[g]]
+                       if(length(lv.idx) > 0L) {
+                           ret <- SE[[g]][, -lv.idx, drop=FALSE]
+                       }
+                       ret
+                   })
+        }
+
         # label?
         if(label) {
             for(g in seq_len(lavdata@ngroups)) {
-                colnames(out[[g]]) <- lavpta$vnames$lv[[g]]
+                if(lavdata@nlevels > 1L && level == 2L) {
+                    gg <- (g - 1)*lavdata@nlevels + 2L
+                } else {
+                    gg <- g
+                }
+                colnames(out[[g]]) <- lavpta$vnames$lv[[gg]]
+            }
+            if(se != "none") {
+                if(lavdata@nlevels > 1L && level == 2L) {
+                    gg <- (g - 1)*lavdata@nlevels + 2L
+                } else {
+                    gg <- g
+                }
+                colnames(SE[[g]]) <- lavpta$vnames$lv[[gg]]
             }
         }
 
@@ -102,8 +164,9 @@ lavPredict <- function(object, type = "lv", newdata = NULL, method = "EBM",
     } else if(type == "yhat") {
         out <- lav_predict_yhat(lavobject = NULL, lavmodel = lavmodel,
                    lavdata = lavdata, lavsamplestats = lavsamplestats,
+                   lavimplied = lavimplied,
                    data.obs = data.obs, eXo = eXo,
-                   ETA = NULL, method = method, optim.method = optim.method)
+                   ETA = ETA, method = method, optim.method = optim.method)
 
         # label?
         if(label) {
@@ -116,8 +179,9 @@ lavPredict <- function(object, type = "lv", newdata = NULL, method = "EBM",
     } else if(type == "fy") {
         out <- lav_predict_fy(lavobject = NULL, lavmodel = lavmodel,
                    lavdata = lavdata, lavsamplestats = lavsamplestats,
+                   lavimplied = lavimplied, 
                    data.obs = data.obs, eXo = eXo,
-                   ETA = NULL, method = method, optim.method = optim.method)
+                   ETA = ETA, method = method, optim.method = optim.method)
 
         # label?
         if(label) {
@@ -142,6 +206,10 @@ lavPredict <- function(object, type = "lv", newdata = NULL, method = "EBM",
     if(fsm) {
         attr(out, "fsm") <- FSM
     }
+   
+    if(se != "none") {
+        attr(out, "se") <- SE
+    }
 
     out
 }
@@ -151,12 +219,15 @@ lav_predict_eta <- function(lavobject = NULL,  # for convenience
                             # sub objects
                             lavmodel = NULL, lavdata = NULL,
                             lavsamplestats = NULL,
+                            lavimplied = NULL,
                             # new data
                             data.obs = NULL, eXo = NULL,
                             # options
                             method = "EBM",
                             fsm = FALSE,
-                            optim.method = "nlminb") {
+                            se = "none",
+                            level = 1L,
+                            optim.method = "bfgs") {
 
     # full object?
     if(inherits(lavobject, "lavaan")) {
@@ -171,29 +242,43 @@ lav_predict_eta <- function(lavobject = NULL,  # for convenience
     # alias
     if(method == "regression") {
         method <- "ebm"
+    } else if(method == "bartlett" || method == "bartlet") {
+        method <- "ml"
     }
 
     # normal case?
     if(all(lavdata@ov$type == "numeric")) {
         if(method == "ebm") {
             out <- lav_predict_eta_normal(lavobject = lavobject,
-                       lavmodel = lavmodel, lavdata = lavdata, 
+                       lavmodel = lavmodel, lavdata = lavdata,
+                       lavimplied = lavimplied, se = se, level = level,
                        lavsamplestats = lavsamplestats,
                        data.obs = data.obs, eXo = eXo, fsm = fsm)
-        } else if(method == "bartlett" || method == "bartlet") {
+        } else if(method == "ml") {
             out <- lav_predict_eta_bartlett(lavobject = lavobject,
                        lavmodel = lavmodel, lavdata = lavdata,
+                       lavimplied = lavimplied, se = se, level = level,
                        lavsamplestats = lavsamplestats,
                        data.obs = data.obs, eXo = eXo, fsm = fsm)
         } else {
             stop("lavaan ERROR: unkown method: ", method)
         }
     } else {
-        out <- lav_predict_eta_ebm(lavobject = lavobject,
-                   lavmodel = lavmodel, lavdata = lavdata,
-                   lavsamplestats = lavsamplestats,
-                   data.obs = data.obs, eXo = eXo,
-                   optim.method = optim.method)
+        if(method == "ebm") {
+            out <- lav_predict_eta_ebm_ml(lavobject = lavobject,
+                       lavmodel = lavmodel, lavdata = lavdata,
+                       lavsamplestats = lavsamplestats, se = se, 
+                       level = level, data.obs = data.obs, eXo = eXo, 
+                       ML = FALSE, optim.method = optim.method)
+        } else if(method == "ml") {
+            out <- lav_predict_eta_ebm_ml(lavobject = lavobject,
+                       lavmodel = lavmodel, lavdata = lavdata,
+                       lavsamplestats = lavsamplestats, se = se, 
+                       level = level, data.obs = data.obs, eXo = eXo, 
+                       ML = TRUE, optim.method = optim.method)
+        } else {
+            stop("lavaan ERROR: unkown method: ", method)
+        }
     }
 
     out
@@ -207,8 +292,10 @@ lav_predict_eta_normal <- function(lavobject = NULL,  # for convenience
                                    # sub objects
                                    lavmodel = NULL, lavdata = NULL, 
                                    lavsamplestats = NULL,
+                                   lavimplied = NULL,
                                    # optional new data
                                    data.obs = NULL, eXo = NULL,
+                                   se = "none", level = 1L,
                                    fsm = FALSE) { 
 
     # full object?
@@ -216,28 +303,107 @@ lav_predict_eta_normal <- function(lavobject = NULL,  # for convenience
         lavmodel       <- lavobject@Model
         lavdata        <- lavobject@Data
         lavsamplestats <- lavobject@SampleStats
+        lavimplied     <- lavobject@implied
     } else {
         stopifnot(!is.null(lavmodel), !is.null(lavdata),
-                  !is.null(lavsamplestats))
+                  !is.null(lavsamplestats), !is.null(lavimplied))
     }
 
     if(is.null(data.obs)) {
         data.obs <- lavdata@X
+        newdata.flag <- FALSE
+    } else {
+        newdata.flag <- TRUE
     }
     # eXo not needed
 
-    Sigma.hat <- computeSigmaHat(lavmodel = lavmodel)
-    Sigma.hat.inv <- lapply(Sigma.hat, solve)
-    VETA   <- computeVETA(lavmodel = lavmodel, lavsamplestats = lavsamplestats)
+    # missings? and missing = "ml"?
+    # impute values under the normal
+    if(lavdata@missing == "ml") {
+        for(g in seq_len(lavdata@ngroups)) {
+            if(newdata.flag) {
+                DATA <- data.obs[[g]]
+                MP   <- lav_data_missing_patterns(data.obs[[g]])
+            } else {
+                DATA <- lavdata@X[[g]]
+                MP   <- lavdata@Mp[[g]]
+            }
+            data.obs[[g]] <- 
+                lav_mvnorm_missing_impute_pattern(Y = DATA,
+                                                  Mp = MP,
+                                                  Mu = lavimplied$mean[[g]],
+                                                  Sigma = lavimplied$cov[[g]])
+        }
+    }
+
+    LAMBDA <- computeLAMBDA(lavmodel = lavmodel, remove.dummy.lv = FALSE)
+    Sigma.hat <- lavimplied$cov
+    Sigma.hat.inv <- lapply(Sigma.hat, MASS::ginv)
+    VETA   <- computeVETA(lavmodel = lavmodel)
     EETA   <- computeEETA(lavmodel = lavmodel, lavsamplestats = lavsamplestats)
     EY     <- computeEY(  lavmodel = lavmodel, lavsamplestats = lavsamplestats)
-    LAMBDA <- computeLAMBDA(lavmodel = lavmodel, remove.dummy.lv = FALSE)
      
     FS <- vector("list", length = lavdata@ngroups)
     if(fsm) {
         FSM <- vector("list", length = lavdata@ngroups)
     }
+    if(se != "none") {
+        SE <- vector("list", length = lavdata@ngroups)
+    }
+
     for(g in 1:lavdata@ngroups) {
+
+        if(lavdata@nlevels > 1L) {
+            Lp <- lavdata@Lp[[g]]
+            YLp <- lavsamplestats@YLp[[g]]
+  
+            # implied for this group
+            group.idx <- (g - 1)*lavdata@nlevels + seq_len(lavdata@nlevels)
+            implied.group <- lapply(lavimplied, function(x) x[group.idx])
+
+            # random effects (=random intercepts or cluster means)
+            out <- lav_mvnorm_cluster_implied22l(Lp = Lp, 
+                                                 implied = implied.group)
+            MB.j <- lav_mvnorm_cluster_em_estep_ranef(YLp = YLp, Lp = Lp,
+                        sigma.w = out$sigma.w, sigma.b = out$sigma.b,
+                        sigma.zz = out$sigma.zz, sigma.yz = out$sigma.yz,
+                        mu.z = out$mu.z, mu.w = out$mu.w, mu.b = out$mu.b,
+                        se = FALSE)
+   
+            ov.idx <- Lp$ov.idx
+
+            if(level == 1L) {
+                data.W <- data.obs[[g]][, ov.idx[[1]] ]
+                data.B <- MB.j[ Lp$cluster.idx[[2]], , drop = FALSE]
+
+                # center
+                data.obs.g <- data.W - data.B
+            } else if(level == 2L) {
+                Data.B <- matrix(0, nrow = nrow(MB.j), 
+                                    ncol = ncol(data.obs[[g]]))
+                Data.B[, ov.idx[[1]] ] <- MB.j
+                data.obs.g <- Data.B[, ov.idx[[2]] ]
+            } else {
+                stop("lavaan ERROR: only 2 levels are supported")
+            }
+
+            gg <- (g-1)*lavdata@nlevels + level
+            VETA.g     <- VETA[[gg]]
+            EETA.g     <- EETA[[gg]]
+            LAMBDA.g   <- LAMBDA[[gg]]
+            EY.g       <- EY[[gg]]
+            Sigma.hat.inv.g <- Sigma.hat.inv[[gg]]
+
+        } else {
+            data.obs.g <- data.obs[[g]]
+            VETA.g     <- VETA[[g]]
+            EETA.g     <- EETA[[g]]
+            LAMBDA.g   <- LAMBDA[[g]]
+            EY.g       <- EY[[g]]
+            Sigma.hat.inv.g <- Sigma.hat.inv[[g]]
+        }
+
+
         nfac <- ncol(VETA[[g]])
         if(nfac == 0L) {
             FS[[g]] <- matrix(0, lavdata@nobs[[g]], nfac)
@@ -245,29 +411,31 @@ lav_predict_eta_normal <- function(lavobject = NULL,  # for convenience
         }
 
         # factor score coefficient matrix 'C'
-        FSC <- VETA[[g]] %*% t(LAMBDA[[g]]) %*% Sigma.hat.inv[[g]]
+        FSC <- VETA.g %*% t(LAMBDA.g) %*% Sigma.hat.inv.g
         if(fsm) {
             FSM[[g]] <- FSC
         }
 
-        RES  <- sweep(data.obs[[g]],  MARGIN = 2L, STATS = EY[[g]],   FUN = "-")
-        FS.g <- sweep(RES %*% t(FSC), MARGIN = 2L, STATS = EETA[[g]], FUN = "+")
+        # standard error
+        if(se == "standard") {
+            tmp <- (VETA.g - 
+             VETA.g %*% t(LAMBDA.g) %*% Sigma.hat.inv.g %*% LAMBDA.g %*% VETA.g)
+            tmp.d <- diag(tmp)
+            tmp.d[ tmp.d < 1e-05 ] <- as.numeric(NA)
+            SE[[g]] <- matrix(sqrt(tmp.d), nrow = 1L)
+        }
 
-        # remove dummy lv's
-        #lv.dummy.idx <- c(lavmodel@ov.y.dummy.lv.idx[[g]],
-        #                  lavmodel@ov.x.dummy.lv.idx[[g]])
-        #if(length(lv.dummy.idx) > 0L) {
-        #    FS.g <- FS.g[,-lv.dummy.idx,drop=FALSE]
-        #}
+        RES  <- sweep(data.obs.g,  MARGIN = 2L, STATS = EY.g,   FUN = "-")
+        FS.g <- sweep(RES %*% t(FSC), MARGIN = 2L, STATS = EETA.g, FUN = "+")
 
         # replace values in dummy lv's by their observed counterpart
-        if(length(lavmodel@ov.y.dummy.lv.idx[[g]]) > 0L) {
+        if(length(lavmodel@ov.y.dummy.lv.idx[[g]]) > 0L && level == 1L) {
             FS.g[, lavmodel@ov.y.dummy.lv.idx[[g]] ] <-
-                data.obs[[g]][, lavmodel@ov.y.dummy.ov.idx[[g]], drop = FALSE]
+                data.obs.g[, lavmodel@ov.y.dummy.ov.idx[[g]], drop = FALSE]
         }
-        if(length(lavmodel@ov.x.dummy.lv.idx[[g]]) > 0L) {
+        if(length(lavmodel@ov.x.dummy.lv.idx[[g]]) > 0L && level == 1L) {
             FS.g[, lavmodel@ov.x.dummy.lv.idx[[g]] ] <-
-                data.obs[[g]][, lavmodel@ov.x.dummy.ov.idx[[g]], drop = FALSE]
+                data.obs.g[, lavmodel@ov.x.dummy.ov.idx[[g]], drop = FALSE]
         }
 
         FS[[g]] <- FS.g
@@ -276,19 +444,35 @@ lav_predict_eta_normal <- function(lavobject = NULL,  # for convenience
     if(fsm) {
         attr(FS, "fsm") <- FSM
     }
+    if(se != "none") {
+        attr(FS, "se") <- SE
+    }
 
     FS
 }
 
 # factor scores - normal case - Bartlett method
-# NOTE: this is the classic 'Bartlett' method; for the linear/continuous 
-#       case, this is equivalent to 'ML'
+# NOTES: 1) this is the classic 'Bartlett' method; for the linear/continuous 
+#           case, this is equivalent to 'ML'
+#        2) the usual formula is: 
+#               FSC = solve(lambda' theta.inv lambda) (lambda' theta.inv)
+#           BUT to deal with zero or negative variances, we use the 
+#           'GLS' version instead:
+#               FSC = solve(lambda' sigma.inv lambda) (lambda' sigma.inv)
+#           Reference: Bentler & Yuan (1997) 'Optimal Conditionally Unbiased 
+#                      Equivariant Factor Score Estimators' 
+#                      in Berkane (Ed) 'Latent variable modeling with 
+#                      applications to causality' (Springer-Verlag)
+#        3) instead of solve(), we use MASS::ginv, for special settings where
+#           -by construction- (lambda' sigma.inv lambda) is singular
 lav_predict_eta_bartlett <- function(lavobject = NULL, # for convenience
                                      # sub objects
                                      lavmodel = NULL, lavdata = NULL, 
                                      lavsamplestats = NULL,
+                                     lavimplied = NULL,
                                      # optional new data
                                      data.obs = NULL, eXo = NULL,
+                                     se = "none", level = 1L,
                                      fsm = FALSE) { 
 
     # full object?
@@ -296,20 +480,46 @@ lav_predict_eta_bartlett <- function(lavobject = NULL, # for convenience
         lavmodel       <- lavobject@Model
         lavdata        <- lavobject@Data
         lavsamplestats <- lavobject@SampleStats
+        lavimplied     <- lavobject@implied
     } else {
         stopifnot(!is.null(lavmodel), !is.null(lavdata),
-                  !is.null(lavsamplestats))
+                  !is.null(lavsamplestats), !is.null(lavimplied))
     }
 
     if(is.null(data.obs)) {
         data.obs <- lavdata@X
+        newdata.flag <- FALSE
+    } else {
+        newdata.flag <- TRUE
     }
     # eXo not needed
 
-    LAMBDA <- computeLAMBDA(lavmodel = lavmodel, remove.dummy.lv = FALSE)
-    THETA  <- computeTHETA(lavmodel = lavmodel)
-    THETA.inv <- lapply(THETA, solve)
+    # missings? and missing = "ml"?
+    # impute values under the normal
+    #
+    # FIXME: THIS IS NOT CORRECT; we should use FIML!!!
+    #
+    if(lavdata@missing == "ml") {
+        for(g in seq_len(lavdata@ngroups)) {
+            if(newdata.flag) {
+                DATA <- data.obs[[g]]
+                MP   <- lav_data_missing_patterns(data.obs[[g]])
+            } else {
+                DATA <- lavdata@X[[g]]
+                MP   <- lavdata@Mp[[g]]
+            }
+            data.obs[[g]] <-
+                lav_mvnorm_missing_impute_pattern(Y = DATA,
+                                                  Mp = MP,
+                                                  Mu = lavimplied$mean[[g]],
+                                                  Sigma = lavimplied$cov[[g]])
+        }
+    }
 
+
+    LAMBDA <- computeLAMBDA(lavmodel = lavmodel, remove.dummy.lv = FALSE)
+    Sigma.hat.inv <- lapply(lavimplied$cov, MASS::ginv)
+    VETA   <- computeVETA(lavmodel = lavmodel) # for se only
     EETA   <- computeEETA(lavmodel = lavmodel, lavsamplestats = lavsamplestats)
     EY     <- computeEY(  lavmodel = lavmodel, lavsamplestats = lavsamplestats)
      
@@ -317,7 +527,62 @@ lav_predict_eta_bartlett <- function(lavobject = NULL, # for convenience
     if(fsm) {
         FSM <- vector("list", length = lavdata@ngroups)
     }
+    if(se != "none") {
+        SE <- vector("list", length = lavdata@ngroups)
+    }
+
     for(g in 1:lavdata@ngroups) {
+
+        if(lavdata@nlevels > 1L) {
+            Lp <- lavdata@Lp[[g]]
+            YLp <- lavsamplestats@YLp[[g]]
+  
+            # implied for this group
+            group.idx <- (g - 1)*lavdata@nlevels + seq_len(lavdata@nlevels)
+            implied.group <- lapply(lavimplied, function(x) x[group.idx])
+
+            # random effects (=random intercepts or cluster means)
+            out <- lav_mvnorm_cluster_implied22l(Lp = Lp, 
+                                                 implied = implied.group)
+            MB.j <- lav_mvnorm_cluster_em_estep_ranef(YLp = YLp, Lp = Lp,
+                        sigma.w = out$sigma.w, sigma.b = out$sigma.b,
+                        sigma.zz = out$sigma.zz, sigma.yz = out$sigma.yz,
+                        mu.z = out$mu.z, mu.w = out$mu.w, mu.b = out$mu.b,
+                        se = FALSE)
+   
+            ov.idx <- Lp$ov.idx
+
+            if(level == 1L) {
+                data.W <- data.obs[[g]][, ov.idx[[1]] ]
+                data.B <- MB.j[ Lp$cluster.idx[[2]], , drop = FALSE]
+
+                # center
+                data.obs.g <- data.W - data.B
+            } else if(level == 2L) {
+                Data.B <- matrix(0, nrow = nrow(MB.j), 
+                                    ncol = ncol(data.obs[[g]]))
+                Data.B[, ov.idx[[1]] ] <- MB.j
+                data.obs.g <- Data.B[, ov.idx[[2]] ]
+            } else {
+                stop("lavaan ERROR: only 2 levels are supported")
+            }
+
+            gg <- (g-1)*lavdata@nlevels + level
+            VETA.g     <- VETA[[gg]]
+            EETA.g     <- EETA[[gg]]
+            LAMBDA.g   <- LAMBDA[[gg]]
+            EY.g       <- EY[[gg]]
+            Sigma.hat.inv.g <- Sigma.hat.inv[[gg]]
+
+        } else {
+            data.obs.g <- data.obs[[g]]
+            VETA.g     <- VETA[[g]]
+            EETA.g     <- EETA[[g]]
+            LAMBDA.g   <- LAMBDA[[g]]
+            EY.g       <- EY[[g]]
+            Sigma.hat.inv.g <- Sigma.hat.inv[[g]]
+        }
+
         nfac <- length(EETA[[g]])
         if(nfac == 0L) {
             FS[[g]] <- matrix(0, lavdata@nobs[[g]], nfac)
@@ -325,15 +590,31 @@ lav_predict_eta_bartlett <- function(lavobject = NULL, # for convenience
         }
 
         # factor score coefficient matrix 'C'
-        FSC = ( solve(t(LAMBDA[[g]]) %*% THETA.inv[[g]] %*% LAMBDA[[g]]) %*%
-                t(LAMBDA[[g]]) %*% THETA.inv[[g]] )
+        FSC <- ( MASS::ginv(t(LAMBDA.g) %*% Sigma.hat.inv.g %*% LAMBDA.g)
+                %*% t(LAMBDA.g) %*% Sigma.hat.inv.g )
 
         if(fsm) {
             FSM[[g]] <- FSC
         }
 
-        RES  <- sweep(data.obs[[g]],  MARGIN = 2L, STATS = EY[[g]],   FUN = "-")
-        FS.g <- sweep(RES %*% t(FSC), MARGIN = 2L, STATS = EETA[[g]], FUN = "+")
+        # standard error
+        if(se == "standard") {
+            # the traditional formula is: 
+            #     solve(t(lambda) %*% solve(theta) %*% lambda)
+            # but we replace it by 
+            #     solve( t(lambda) %*% solve(sigma) %*% lambda ) - psi
+            # to handle negative variances
+            # in addition, we use ginv 
+            tmp <- ( MASS::ginv(t(LAMBDA.g) %*% 
+                                  Sigma.hat.inv.g %*% LAMBDA.g) 
+                     - VETA.g )
+            tmp.d <- diag(tmp)
+            tmp.d[ tmp.d < 1e-05 ] <- as.numeric(NA)
+            SE[[g]] <- matrix(sqrt(tmp.d), nrow = 1L)
+        }
+
+        RES  <- sweep(data.obs.g,  MARGIN = 2L, STATS = EY.g,   FUN = "-")
+        FS.g <- sweep(RES %*% t(FSC), MARGIN = 2L, STATS = EETA.g, FUN = "+")
 
         # remove dummy lv's
         #lv.dummy.idx <- c(lavmodel@ov.y.dummy.lv.idx[[g]],
@@ -343,11 +624,11 @@ lav_predict_eta_bartlett <- function(lavobject = NULL, # for convenience
         #}
 
         # replace values in dummy lv's by their observed counterpart
-        if(length(lavmodel@ov.y.dummy.lv.idx[[g]]) > 0L) {
+        if(length(lavmodel@ov.y.dummy.lv.idx[[g]]) > 0L && level == 1L) {
             FS.g[, lavmodel@ov.y.dummy.lv.idx[[g]] ] <-
                 data.obs[[g]][, lavmodel@ov.y.dummy.ov.idx[[g]], drop = FALSE]
         }
-        if(length(lavmodel@ov.x.dummy.lv.idx[[g]]) > 0L) {
+        if(length(lavmodel@ov.x.dummy.lv.idx[[g]]) > 0L && level == 1L) {
             FS.g[, lavmodel@ov.x.dummy.lv.idx[[g]] ] <-
                 data.obs[[g]][, lavmodel@ov.x.dummy.ov.idx[[g]], drop = FALSE]
         }
@@ -358,20 +639,27 @@ lav_predict_eta_bartlett <- function(lavobject = NULL, # for convenience
     if(fsm) {
         attr(FS, "fsm") <- FSM
     }
+    if(se != "none") {
+        attr(FS, "se") <- SE
+    }
 
     FS
 }
 
-# factor scores - EBM
-lav_predict_eta_ebm <- function(lavobject = NULL,  # for convenience
-                                # sub objects
-                                lavmodel = NULL, lavdata = NULL,
-                                lavsamplestats = NULL,
-                                # optional new data
-                                data.obs = NULL, eXo = NULL,
-                                optim.method = "nlminb") {
+# factor scores - EBM or ML
+lav_predict_eta_ebm_ml <- function(lavobject = NULL,  # for convenience
+                                   # sub objects
+                                   lavmodel = NULL, lavdata = NULL,
+                                   lavsamplestats = NULL,
+                                   # optional new data
+                                   data.obs = NULL, eXo = NULL,
+                                   se = "none", level = 1L,
+                                   ML = FALSE,
+                                   optim.method = "bfgs") {
 
-    stopifnot(optim.method %in% c("nlminb", "BFGS"))
+    optim.method <- tolower(optim.method)
+
+    stopifnot(optim.method %in% c("nlminb", "bfgs"))
 
     ### FIXME: if all indicators of a factor are normal, can we not
     ###        just use the `classic' regression method??
@@ -395,6 +683,11 @@ lav_predict_eta_ebm <- function(lavobject = NULL,  # for convenience
         eXo <- lavdata@eXo
     }
 
+    # se?
+    if(se != "none") {
+        warning("lavaan WARNING: standard errors are not available (yet) for the non-normal case")
+    }
+
     VETAx <- computeVETAx(lavmodel = lavmodel)
     VETAx.inv <- VETAx
     for(g in seq_len(lavdata@ngroups)) {
@@ -403,7 +696,8 @@ lav_predict_eta_ebm <- function(lavobject = NULL,  # for convenience
         }
     }
     EETAx <- computeEETAx(lavmodel = lavmodel, lavsamplestats = lavsamplestats,
-                          eXo = eXo, remove.dummy.lv = TRUE) ## FIXME? 
+                          eXo = eXo, nobs = lavdata@norig, 
+                          remove.dummy.lv = TRUE) ## FIXME? 
     TH    <- computeTH(   lavmodel = lavmodel)
     THETA <- computeTHETA(lavmodel = lavmodel)
 
@@ -430,12 +724,23 @@ lav_predict_eta_ebm <- function(lavobject = NULL,  # for convenience
                                        th.idx         = th.idx,
                                        log            = TRUE)
 
-        diff <- t(x) - mu.i
-        V <- VETAx.inv[[g]]
-        # handle missing values: we skip them
-        tmp <- as.numeric(0.5 * diff %*% V %*% t(diff))
-        # handle missing values: we skip them: FIXME!!!
-        out <- tmp - sum(log.fy, na.rm=TRUE)
+        if(ML) {
+            # NOTE: 'true' ML is simply  -1*sum(log.fy)
+            #  - but there is no upper/lower bound for the extrema:
+            #    a pattern of all (in)correct drives the 'theta' parameter
+            #    towards +/- Inf
+            # - therefore, we add a vague prior, just to stabilize
+            #    
+            diff <- t(x) - mu.i
+            V <- diag( length(x) ) * 1e-05
+            tmp <- as.numeric(0.5 * diff %*% V %*% t(diff))
+            out <- 1 + tmp - sum(log.fy, na.rm=TRUE)
+        } else {
+            diff <- t(x) - mu.i
+            V <- VETAx.inv[[g]]
+            tmp <- as.numeric(0.5 * diff %*% V %*% t(diff))
+            out <- tmp - sum(log.fy, na.rm=TRUE)
+        }
         out
     }
 
@@ -488,23 +793,27 @@ lav_predict_eta_ebm <- function(lavobject = NULL,  # for convenience
 
             START <- numeric(nfac) # initial values for eta
 
-            # find best values for eta.i
-            if(optim.method == "nlminb") {
-                out <- nlminb(start=START, objective=f.eta.i,
-                              gradient=NULL, # for now
-                              control=list(rel.tol=1e-8),
-                              y.i=y.i, x.i=x.i, mu.i=mu.i)
-            } else if(optim.method == "BFGS") {
-                out <- optim(par = START, fn = f.eta.i,
-                             gr = NULL,
-                             control = list(reltol = 1e-8),
-                             method = "BFGS",
-                             y.i = y.i, x.i = x.i, mu.i = mu.i)
-            }
-            if(out$convergence == 0L) {
-                eta.i <- out$par
+            if(!all(is.na(y.i))) {
+                # find best values for eta.i
+                if(optim.method == "nlminb") {
+                    out <- nlminb(start=START, objective=f.eta.i,
+                                  gradient=NULL, # for now
+                                  control=list(rel.tol=1e-8),
+                                  y.i=y.i, x.i=x.i, mu.i=mu.i)
+                } else if(optim.method == "bfgs") {
+                    out <- optim(par = START, fn = f.eta.i,
+                                 gr = NULL, 
+                                 control = list(reltol = 1e-8, fnscale = 1.1),
+                                 method = "BFGS",
+                                 y.i = y.i, x.i = x.i, mu.i = mu.i)
+                }
+                if(out$convergence == 0L) {
+                    eta.i <- out$par
+                } else {
+                    eta.i <- rep(as.numeric(NA), nfac)
+                }
             } else {
-                eta.i <- rep(as.numeric(NA), nfac2)
+                eta.i <- rep(as.numeric(NA), nfac)
             }
 
             # add dummy ov.y lv values
@@ -535,6 +844,7 @@ lav_predict_yhat <- function(lavobject = NULL, # for convience
                              # sub objects
                              lavmodel = NULL, lavdata = NULL,
                              lavsamplestats = NULL,
+                             lavimplied = NULL,
                              # new data
                              data.obs = NULL, eXo = NULL,
                              # ETA values
@@ -542,16 +852,17 @@ lav_predict_yhat <- function(lavobject = NULL, # for convience
                              # options
                              method = "EBM", 
                              duplicate = FALSE,
-                             optim.method = "nlminb") {
+                             optim.method = "bfgs") {
 
     # full object?
     if(inherits(lavobject, "lavaan")) {
         lavmodel       <- lavobject@Model
         lavdata        <- lavobject@Data
         lavsamplestats <- lavobject@SampleStats
+        lavimplied     <- lavobject@implied
     } else {
         stopifnot(!is.null(lavmodel), !is.null(lavdata), 
-                  !is.null(lavsamplestats))
+                  !is.null(lavsamplestats), !is.null(lavimplied))
     }
 
     # new data?
@@ -566,13 +877,26 @@ lav_predict_yhat <- function(lavobject = NULL, # for convience
     if(is.null(ETA)) {
         ETA <- lav_predict_eta(lavobject = NULL, lavmodel = lavmodel,
                    lavdata = lavdata, lavsamplestats = lavsamplestats,
+                   lavimplied = lavimplied,
                    data.obs = data.obs, eXo = eXo, method = method,
                    optim.method = optim.method)
     } else {
-        # list
+        # matrix
         if(is.matrix(ETA)) { # user-specified?
-            tmp <- ETA; ETA <- vector("list", length=lavdata@ngroups)
-            ETA[seq_len(lavdata@ngroups)] <- list(tmp)
+            if(nrow(ETA) == 1L) {
+                tmp <- matrix(ETA, lavsamplestats@ntotal, length(ETA), 
+                              byrow = TRUE)
+            } else if(nrow(ETA) != lavsamplestats@ntotal) {
+                stop("lavaan ERROR: nrow(ETA) != lavsamplestats@ntotal")
+            } else {
+                tmp <- ETA
+            }
+            ETA <- lapply(1:lavdata@ngroups, function(i) tmp[lavdata@case.idx[[i]],])
+        # vector: just 1 row of factor-scores
+        } else if(is.numeric(ETA)) {
+            # convert to matrix
+            tmp <- matrix(ETA, lavsamplestats@ntotal, length(ETA), byrow = TRUE)
+            ETA <- lapply(1:lavdata@ngroups, function(i) tmp[lavdata@case.idx[[i]],])
         } else if(is.list(ETA)) {
             stopifnot(lavdata@ngroups == length(ETA))
         }
@@ -580,6 +904,7 @@ lav_predict_yhat <- function(lavobject = NULL, # for convience
 
     YHAT <- computeYHAT(lavmodel = lavmodel, GLIST = NULL,
                         lavsamplestats = lavsamplestats, eXo = eXo,
+                        nobs = lavdata@norig,
                         ETA = ETA, duplicate = duplicate)
 
     # if conditional.x, paste eXo
@@ -599,6 +924,7 @@ lav_predict_fy <- function(lavobject = NULL, # for convience
                            # sub objects
                            lavmodel = NULL, lavdata = NULL,
                            lavsamplestats = NULL,
+                           lavimplied = NULL,
                            # new data
                            data.obs = NULL, eXo = NULL,
                            # ETA values
@@ -606,16 +932,17 @@ lav_predict_fy <- function(lavobject = NULL, # for convience
                            # options
                            method = "EBM",
                            log. = FALSE,
-                           optim.method = "nlminb") {
+                           optim.method = "bfgs") {
 
     # full object?
     if(inherits(lavobject, "lavaan")) {
         lavmodel       <- lavobject@Model
         lavdata        <- lavobject@Data
         lavsamplestats <- lavobject@SampleStats
+        lavimplied     <- lavobject@implied
     } else {
         stopifnot(!is.null(lavmodel), !is.null(lavdata),
-                  !is.null(lavsamplestats))
+                  !is.null(lavsamplestats), !is.null(lavimplied))
     }
 
     # new data?
@@ -629,6 +956,7 @@ lav_predict_fy <- function(lavobject = NULL, # for convience
     # we need the YHATs (per group)
     YHAT <- lav_predict_yhat(lavobject = NULL, lavmodel = lavmodel,
                 lavdata = lavdata, lavsamplestats = lavsamplestats,
+                lavimplied = lavimplied,
                 data.obs = data.obs, eXo = eXo, ETA = ETA, method = method,
                 duplicate = FALSE, optim.method = optim.method)
 
