@@ -59,7 +59,7 @@ function(object, type="raw", labels=TRUE) {
         stop("lavaan ERROR: can not compute standardized residuals if there are no free parameters in the model")
     }
 
-    G <- object@Data@ngroups
+    G <- object@Model@ngroups
     meanstructure <- object@Model@meanstructure
     ov.names <- object@Data@ov.names
 
@@ -69,7 +69,7 @@ function(object, type="raw", labels=TRUE) {
         # fixed.x idx?
         x.idx <- integer(0)
         if(object@Options$fixed.x) {
-            x.idx <- match(vnames(object@ParTable, "ov.x", group=1L),
+            x.idx <- match(vnames(object@ParTable, "ov.x", block=1L),
                            object@Data@ov.names[[1L]]) ### FIXME!!!! will not
                                                        ### work for different
         }                                              ### models in groups
@@ -86,16 +86,16 @@ function(object, type="raw", labels=TRUE) {
             augUser$free[      idx ] <- max(augUser$free) + 1:length(idx) 
             #augUser$unco[idx ] <- max(augUser$unco) + 1:length(idx) 
             augModel <- lav_model(lavpartable    = augUser,
-                            representation = object@Options$representation,
-                            conditional.x  = object@Options$conditional.x,
-                            parameterization = object@Options$parameterization,
-                            link           = object@Options$link,
-                            debug          = object@Options$debug)
+                                  lavoptions     = object@Options,
+                                  cov.x          = object@SampleStats@cov.x,
+                                  mean.x         = object@SampleStats@mean.x)
             VarCov <- lav_model_vcov(lavmodel       = augModel, 
                                      lavsamplestats = object@SampleStats,
                                      lavdata        = object@Data,
                                      lavpartable    = object@ParTable,
-                                     lavoptions     = object@Options)
+                                     lavoptions     = object@Options,
+                                     lavimplied     = object@implied,
+                                     lavh1          = object@h1)
             # set cov between free and fixed.x elements to zero
             ###
             ### FIXME: should we not do this on the information level,
@@ -112,7 +112,9 @@ function(object, type="raw", labels=TRUE) {
                                      lavdata        = object@Data,
                                      lavpartable    = object@ParTable,
                                      lavsamplestats = object@SampleStats,
-                                     lavoptions     = object@Options)
+                                     lavoptions     = object@Options,
+                                     lavimplied     = object@implied,
+                                     lavh1          = object@h1)
             Delta  <- computeDelta(lavmodel = object@Model)
         }   
     }
@@ -141,7 +143,7 @@ function(object, type="raw", labels=TRUE) {
         }
         nvar <- ncol(S)
 
-        # residuals (for this group)
+        # residuals (for this block)
         if(type == "cor.bollen") {
             if(object@Model@conditional.x) {
                 R[[g]]$cov  <- cov2cor(S) - cov2cor(object@implied$res.cov[[g]])
@@ -196,7 +198,7 @@ function(object, type="raw", labels=TRUE) {
                 R[[g]]$th <- R[[g]]$th[ -NUM.idx ]
             }
             if(labels) {
-                names(R[[g]]$th) <- vnames(object@ParTable, type="th", group=g)
+                names(R[[g]]$th) <- vnames(object@ParTable, type="th", block=g)
             }
         }
 
@@ -235,6 +237,7 @@ function(object, type="raw", labels=TRUE) {
                     # data complete, under h1, expected == observed
                     A1 <- lav_mvnorm_h1_information_observed_samplestats(
                       sample.cov     = lavsamplestats@cov[[g]],
+                      x.idx          = lavsamplestats@x.idx[[g]],
                       sample.cov.inv = lavsamplestats@icov[[g]])
                 }
 
@@ -248,6 +251,7 @@ function(object, type="raw", labels=TRUE) {
                     B1 <- lav_mvnorm_h1_information_firstorder(
                           Y     = lavdata@X[[g]],
                           Gamma = lavsamplestats@NACOV[[g]],
+                          x.idx = lavsamplestats@x.idx[[g]],
                           sample.cov = lavsamplestats@cov[[g]],
                           sample.cov.inv = lavsamplestats@icov[[g]])
                 }
@@ -266,6 +270,7 @@ function(object, type="raw", labels=TRUE) {
                     B1 <- lav_mvnorm_h1_information_firstorder(
                           Y     = lavdata@X[[g]],
                           Gamma = lavsamplestats@NACOV[[g]],
+                          x.idx = lavsamplestats@x.idx[[g]],
                           sample.cov = lavsamplestats@cov[[g]],
                           sample.cov.inv = lavsamplestats@icov[[g]])
                 }
