@@ -6,8 +6,7 @@ standardize.est.lv.x <- function(x, lavobject, partable = NULL, cov.std = TRUE)
     est[free.idx] <- x
 
     # take care of setResidualElements...
-    lavmodel <- lav_model_set_parameters(lavmodel = lavobject@Model, x = x, 
-                                         estimator = lavobject@Options$estimator)
+    lavmodel <- lav_model_set_parameters(lavmodel = lavobject@Model, x = x)
     GLIST <- lavmodel@GLIST
 
     standardize.est.lv(lavobject = lavobject, partable = partable, est = est, 
@@ -22,8 +21,7 @@ standardize.est.all.x <- function(x, lavobject, partable = NULL, cov.std = TRUE)
     est[free.idx] <- x
 
     # take care of setResidualElements...
-    lavmodel <- lav_model_set_parameters(lavmodel = lavobject@Model, x = x, 
-                                         estimator = lavobject@Options$estimator)
+    lavmodel <- lav_model_set_parameters(lavmodel = lavobject@Model, x = x)
     GLIST <- lavmodel@GLIST
 
     standardize.est.all(lavobject = lavobject, partable = partable, est = est,
@@ -38,8 +36,7 @@ standardize.est.all.nox.x <- function(x, lavobject, partable = NULL, cov.std = T
     est[free.idx] <- x
 
     # take care of setResidualElements...
-    lavmodel <- lav_model_set_parameters(lavmodel = lavobject@Model, x = x, 
-                                         estimator = lavobject@Options$estimator)
+    lavmodel <- lav_model_set_parameters(lavmodel = lavobject@Model, x = x)
     GLIST <- lavmodel@GLIST
 
     standardize.est.all.nox(lavobject = lavobject, partable = partable, est = est, 
@@ -67,20 +64,19 @@ standardize.est.lv <- function(lavobject, partable=NULL, est=NULL, GLIST=NULL,
 
     # compute ETA
     LV.ETA <- computeVETA(lavmodel       = lavobject@Model,
-                          GLIST          = GLIST,
-                          lavsamplestats = lavobject@SampleStats)
+                          GLIST          = GLIST)
     
-    for(g in 1:lavobject@Data@ngroups) {
+    for(g in 1:lavobject@Model@nblocks) {
 
-        ov.names <- vnames(lavobject@ParTable, "ov", group=g) # not user, 
+        ov.names <- vnames(lavobject@ParTable, "ov", block=g) # not user, 
                                                        # which may be incomplete
-        lv.names <- vnames(lavobject@ParTable, "lv", group=g)
+        lv.names <- vnames(lavobject@ParTable, "lv", block=g)
        
-        # shortcut: no latents in this group, nothing to do
+        # shortcut: no latents in this block, nothing to do
         if(length(lv.names) == 0L)
             next
 
-        # which mm belong to group g?
+        # which mm belong to block g?
         mm.in.group <- 1:nmat[g] + cumsum(c(0,nmat))[g]
         MLIST <- GLIST[ mm.in.group ]
 
@@ -89,34 +85,34 @@ standardize.est.lv <- function(lavobject, partable=NULL, est=NULL, GLIST=NULL,
 
         # 1a. "=~" regular indicators
         idx <- which(partable$op == "=~" & !(partable$rhs %in% lv.names) & 
-                     partable$group == g)
+                     partable$block == g)
         out[idx] <- out[idx] * ETA[ match(partable$lhs[idx], lv.names) ]
 
         # 1b. "=~" regular higher-order lv indicators
         idx <- which(partable$op == "=~" & !(partable$rhs %in% ov.names) &
-                     partable$group == g)
+                     partable$block == g)
         out[idx] <- ( out[idx] * ETA[ match(partable$lhs[idx], lv.names) ]
                                / ETA[ match(partable$rhs[idx], lv.names) ] )
 
         # 1c. "=~" indicators that are both in ov and lv
         #idx <- which(partable$op == "=~" & partable$rhs %in% ov.names
         #                             & partable$rhs %in% lv.names &
-        #             partable$group == g)
+        #             partable$block == g)
 
         # 2. "~" regressions (and "<~")
         idx <- which((partable$op == "~" | partable$op == "<~") & 
                      partable$lhs %in% lv.names &
-                     partable$group == g)
+                     partable$block == g)
         out[idx] <- out[idx] / ETA[ match(partable$lhs[idx], lv.names) ] 
 
         idx <- which((partable$op == "~" | partable$op == "<~") & 
                      partable$rhs %in% lv.names &
-                     partable$group == g)
+                     partable$block == g)
         out[idx] <- out[idx] * ETA[ match(partable$rhs[idx], lv.names) ]
 
         # 3a. "~~" ov
         #idx <- which(partable$op == "~~" & !(partable$lhs %in% lv.names) & 
-        #             partable$group == g)
+        #             partable$block == g)
 
         # 3b. "~~" lv
         # ATTENTION: in Mplus 4.1, the off-diagonal residual covariances 
@@ -129,7 +125,7 @@ standardize.est.lv <- function(lavobject, partable=NULL, est=NULL, GLIST=NULL,
         # variances
         rv.idx <- which(partable$op == "~~" & partable$rhs %in% lv.names &
                         partable$lhs == partable$rhs &
-                        partable$group == g)
+                        partable$block == g)
         out[rv.idx] <- ( out[rv.idx] / ETA[ match(partable$lhs[rv.idx], lv.names) ]
                                      / ETA[ match(partable$rhs[rv.idx], lv.names) ] )
 
@@ -151,7 +147,7 @@ standardize.est.lv <- function(lavobject, partable=NULL, est=NULL, GLIST=NULL,
         idx.lhs <- which(partable$op == "~~" &
                          partable$lhs %in% lv.names &
                          partable$lhs != partable$rhs &
-                         partable$group == g)
+                         partable$block == g)
         if(length(idx.lhs) > 0L) {
             if(cov.std == FALSE) {
                 out[idx.lhs] <- 
@@ -166,7 +162,7 @@ standardize.est.lv <- function(lavobject, partable=NULL, est=NULL, GLIST=NULL,
         idx.rhs <- which(partable$op == "~~" & 
                          partable$rhs %in% lv.names &
                          partable$lhs != partable$rhs &
-                         partable$group == g)
+                         partable$block == g)
         if(length(idx.rhs) > 0L) {
             if(cov.std == FALSE) {
                 out[idx.rhs] <- 
@@ -180,11 +176,11 @@ standardize.est.lv <- function(lavobject, partable=NULL, est=NULL, GLIST=NULL,
 
         # 4a. "~1" ov
         #idx <- which(partable$op == "~1" & !(partable$lhs %in% lv.names) &
-        #             partable$group == g)
+        #             partable$block == g)
 
         # 4b. "~1" lv
         idx <- which(partable$op == "~1" & partable$lhs %in% lv.names &
-                     partable$group == g)
+                     partable$block == g)
         out[idx] <- out[idx] / ETA[ match(partable$lhs[idx], lv.names) ]
     }
 
@@ -227,26 +223,41 @@ standardize.est.all <- function(lavobject, partable=NULL, est=NULL, est.std=NULL
     stopifnot(N == length(partable$lhs))
 
     VY <- computeVY(lavmodel = lavobject@Model, GLIST = GLIST,
-                    lavsamplestats = lavobject@SampleStats, 
                     diagonal.only = TRUE)
 
-    for(g in 1:lavobject@Data@ngroups) {
 
-        ov.names <- vnames(lavobject@ParTable, "ov", group=g) # not user
-        lv.names <- vnames(lavobject@ParTable, "lv", group=g)
+    for(g in 1:lavobject@Model@nblocks) {
 
-        OV  <- sqrt(VY[[g]])
+        ov.names <- vnames(lavobject@ParTable, "ov", block = g) # not user
+        lv.names <- vnames(lavobject@ParTable, "lv", block = g)
+
+        OV2 <- VY[[g]]
+        # replace zero values by NA (but keep negative values)
+        zero.idx <- which(abs(OV2) < .Machine$double.eps)
+        if(length(zero.idx) > 0L) {
+            OV2[zero.idx] <- as.numeric(NA)
+        }
+     
+        # replace negative values by NA (for sqrt)
+        tmp.OV2 <- OV2
+        neg.idx <- which(tmp.OV2 < 0)
+        if(length(neg.idx) > 0L) {
+            tmp.OV2[neg.idx] <- as.numeric(NA)
+        }
+        OV  <- sqrt(tmp.OV2)
 
         if(lavobject@Model@conditional.x) {
             # extend OV with ov.names.x
-            ov.names.x <- vnames(lavobject@ParTable, "ov.x", group=g)
-            ov.names <- c(ov.names, ov.names.x)
+            ov.names.x <- vnames(lavobject@ParTable, "ov.x", block = g)
+            ov.names.nox <- vnames(lavobject@ParTable, "ov.nox", block = g)
+            ov.names <- c(ov.names.nox, ov.names.x)
+            OV2 <- c(OV2, diag(lavobject@SampleStats@cov.x[[g]]))
             OV <- c(OV, sqrt(diag(lavobject@SampleStats@cov.x[[g]])))
         }
 
         # 1a. "=~" regular indicators
         idx <- which(partable$op == "=~" & !(partable$rhs %in% lv.names) &
-                     partable$group == g)
+                     partable$block == g)
         out[idx] <- out[idx] / OV[ match(partable$rhs[idx], ov.names) ]
 
         # 1b. "=~" regular higher-order lv indicators
@@ -254,17 +265,17 @@ standardize.est.all <- function(lavobject, partable=NULL, est=NULL, est.std=NULL
         # 1c. "=~" indicators that are both in ov and lv
         #idx <- which(partable$op == "=~" & partable$rhs %in% ov.names
         #                             & partable$rhs %in% lv.names &
-        #             partable$group == g)
+        #             partable$block == g)
 
         # 2. "~" regressions (and "<~")
         idx <- which((partable$op == "~" | partable$op == "<~") & 
                      partable$lhs %in% ov.names &
-                     partable$group == g)
+                     partable$block == g)
         out[idx] <- out[idx] / OV[ match(partable$lhs[idx], ov.names) ]
 
         idx <- which((partable$op == "~" | partable$op == "<~") & 
                      partable$rhs %in% ov.names &
-                     partable$group == g)
+                     partable$block == g)
         out[idx] <- out[idx] * OV[ match(partable$rhs[idx], ov.names) ]
 
         # 3a. "~~" ov
@@ -278,9 +289,11 @@ standardize.est.all <- function(lavobject, partable=NULL, est=NULL, est.std=NULL
         # variances
         rv.idx <- which(partable$op == "~~" & !(partable$lhs %in% lv.names) & 
                         partable$lhs == partable$rhs &
-                        partable$group == g)
-        out[rv.idx] <- ( out[rv.idx] / OV[ match(partable$lhs[rv.idx], ov.names) ]
-                                     / OV[ match(partable$rhs[rv.idx], ov.names) ] )
+                        partable$block == g)
+        #out[rv.idx] <- ( out[rv.idx] / OV[ match(partable$lhs[rv.idx], ov.names) ]
+        #                             / OV[ match(partable$rhs[rv.idx], ov.names) ] )
+        out[rv.idx] <- ( out[rv.idx] / 
+                             OV2[ match(partable$lhs[rv.idx], ov.names) ] )
 
         # covariances ov
         # three types:
@@ -300,7 +313,7 @@ standardize.est.all <- function(lavobject, partable=NULL, est=NULL, est.std=NULL
         idx.lhs <- which(partable$op == "~~" &
                          !(partable$lhs %in% lv.names) &
                          partable$lhs != partable$rhs &
-                         partable$group == g)
+                         partable$block == g)
         if(length(idx.lhs) > 0L) {
             if(cov.std == FALSE) {
                 out[idx.lhs] <- 
@@ -315,7 +328,7 @@ standardize.est.all <- function(lavobject, partable=NULL, est=NULL, est.std=NULL
         idx.rhs <- which(partable$op == "~~" & 
                          !(partable$rhs %in% lv.names) &
                          partable$lhs != partable$rhs &
-                         partable$group == g)
+                         partable$block == g)
         if(length(idx.rhs) > 0L) {
             if(cov.std == FALSE) {
                 out[idx.rhs] <- 
@@ -328,25 +341,25 @@ standardize.est.all <- function(lavobject, partable=NULL, est=NULL, est.std=NULL
 
         # 3b. "~~" lv
         #idx <- which(partable$op == "~~" & partable$rhs %in% lv.names &
-        #             partable$group == g)
+        #             partable$block == g)
 
         # 4a. "~1" ov
         idx <- which(partable$op == "~1" & !(partable$lhs %in% lv.names) &
-                     partable$group == g)
+                     partable$block == g)
         out[idx] <- out[idx] / OV[ match(partable$lhs[idx], ov.names) ]
 
         # 4b. "~1" lv
         #idx <- which(partable$op == "~1" & partable$lhs %in% lv.names &
-        #             partable$group == g)
+        #             partable$block == g)
 
         # 4c. "|" thresholds
         idx <- which(partable$op == "|" & !(partable$lhs %in% lv.names) &
-                     partable$group == g)
+                     partable$block == g)
         out[idx] <- out[idx] / OV[ match(partable$lhs[idx], ov.names) ]
 
         # 4d. "~*~" scales
         idx <- which(partable$op == "~*~" & !(partable$lhs %in% lv.names) &
-                     partable$group == g)
+                     partable$block == g)
         out[idx] <- 1.0
     }
 
@@ -391,28 +404,42 @@ standardize.est.all.nox <- function(lavobject, partable=NULL, est=NULL,
     stopifnot(N == length(partable$lhs))
 
     VY <- computeVY(lavmodel = lavobject@Model, GLIST = GLIST,
-                    lavsamplestats = lavobject@SampleStats,
                     diagonal.only = TRUE)
 
-    for(g in 1:lavobject@Data@ngroups) {
 
-        ov.names     <- vnames(lavobject@ParTable, "ov",     group=g) # not user
-        ov.names.x   <- vnames(lavobject@ParTable, "ov.x",   group=g)
-        ov.names.nox <- vnames(lavobject@ParTable, "ov.nox", group=g)
-        lv.names     <- vnames(lavobject@ParTable, "lv",     group=g)
+    for(g in 1:lavobject@Model@nblocks) {
 
-        OV  <- sqrt(VY[[g]])
+        ov.names     <- vnames(lavobject@ParTable, "ov",     block = g) 
+        ov.names.x   <- vnames(lavobject@ParTable, "ov.x",   block = g)
+        ov.names.nox <- vnames(lavobject@ParTable, "ov.nox", block = g)
+        lv.names     <- vnames(lavobject@ParTable, "lv",     block = g)
+
+        OV2 <- VY[[g]]
+        # replace zero values by NA (but keep negative values)
+        zero.idx <- which(abs(OV2) < .Machine$double.eps)
+        if(length(zero.idx) > 0L) {
+            OV2[zero.idx] <- as.numeric(NA)
+        }
+
+        # replace negative values by NA (for sqrt)
+        tmp.OV2 <- OV2
+        neg.idx <- which(tmp.OV2 < 0)
+        if(length(neg.idx) > 0L) {
+            tmp.OV2[neg.idx] <- as.numeric(NA)
+        }
+        OV  <- sqrt(tmp.OV2)
 
         if(lavobject@Model@conditional.x) {
             # extend OV with ov.names.x
-            ov.names.x <- vnames(lavobject@ParTable, "ov.x", group=g)
-            ov.names <- c(ov.names, ov.names.x)
+            ov.names.x <- vnames(lavobject@ParTable, "ov.x", block = g)
+            ov.names <- c(ov.names.nox, ov.names.x)
+            OV2 <- c(OV2, diag(lavobject@SampleStats@cov.x[[g]]))
             OV <- c(OV, sqrt(diag(lavobject@SampleStats@cov.x[[g]])))
         }
 
         # 1a. "=~" regular indicators
         idx <- which(partable$op == "=~" & !(partable$rhs %in% lv.names) &
-                     partable$group == g)
+                     partable$block == g)
         out[idx] <- out[idx] / OV[ match(partable$rhs[idx], ov.names) ]
 
         # 1b. "=~" regular higher-order lv indicators
@@ -420,17 +447,17 @@ standardize.est.all.nox <- function(lavobject, partable=NULL, est=NULL,
         # 1c. "=~" indicators that are both in ov and lv
         #idx <- which(partable$op == "=~" & partable$rhs %in% ov.names
         #                             & partable$rhs %in% lv.names &
-        #             partable$group == g)
+        #             partable$block == g)
 
         # 2. "~" regressions (and "<~")
         idx <- which((partable$op == "~" | partable$op == "<~") & 
                      partable$lhs %in% ov.names &
-                     partable$group == g)
+                     partable$block == g)
         out[idx] <- out[idx] / OV[ match(partable$lhs[idx], ov.names) ]
 
         idx <- which((partable$op == "~" | partable$op == "<~") & 
                      partable$rhs %in% ov.names.nox &
-                     partable$group == g)
+                     partable$block == g)
         out[idx] <- out[idx] * OV[ match(partable$rhs[idx], ov.names.nox) ]
 
         # 3a. "~~" ov
@@ -445,9 +472,11 @@ standardize.est.all.nox <- function(lavobject, partable=NULL, est=NULL,
         rv.idx <- which(partable$op == "~~" & !(partable$lhs %in% lv.names) & 
                         !(partable$lhs %in% ov.names.x) &
                         partable$lhs == partable$rhs &
-                        partable$group == g)
-        out[rv.idx] <- ( out[rv.idx] / OV[ match(partable$lhs[rv.idx], ov.names) ]
-                                     / OV[ match(partable$rhs[rv.idx], ov.names) ] )
+                        partable$block == g)
+        #out[rv.idx] <- ( out[rv.idx] / OV[ match(partable$lhs[rv.idx], ov.names) ]
+         #                            / OV[ match(partable$rhs[rv.idx], ov.names) ] )
+        out[rv.idx] <- ( out[rv.idx] / 
+                             OV2[ match(partable$lhs[rv.idx], ov.names) ] )
 
         # covariances ov
         # three types:
@@ -468,7 +497,7 @@ standardize.est.all.nox <- function(lavobject, partable=NULL, est=NULL,
                          !(partable$lhs %in% lv.names) &
                          !(partable$lhs %in% ov.names.x) &
                          partable$lhs != partable$rhs &
-                         partable$group == g)
+                         partable$block == g)
         if(length(idx.lhs) > 0L) {
             if(cov.std == FALSE) {
                 out[idx.lhs] <- 
@@ -484,7 +513,7 @@ standardize.est.all.nox <- function(lavobject, partable=NULL, est=NULL,
                          !(partable$rhs %in% lv.names) &
                          !(partable$rhs %in% ov.names.x) &
                          partable$lhs != partable$rhs &
-                         partable$group == g)
+                         partable$block == g)
         if(length(idx.rhs) > 0L) {
             if(cov.std == FALSE) {
                 out[idx.rhs] <- 
@@ -497,26 +526,26 @@ standardize.est.all.nox <- function(lavobject, partable=NULL, est=NULL,
 
         # 3b. "~~" lv
         #idx <- which(partable$op == "~~" & partable$rhs %in% lv.names &
-        #             partable$group == g)
+        #             partable$block == g)
 
         # 4a. "~1" ov
         idx <- which(partable$op == "~1" & !(partable$lhs %in% lv.names) &
                      !(partable$lhs %in% ov.names.x) &
-                     partable$group == g)
+                     partable$block == g)
         out[idx] <- out[idx] / OV[ match(partable$lhs[idx], ov.names) ]
 
         # 4b. "~1" lv
         #idx <- which(partable$op == "~1" & partable$lhs %in% lv.names &
-        #             partable$group == g)
+        #             partable$block == g)
 
         # 4c. "|" thresholds
         idx <- which(partable$op == "|" & !(partable$lhs %in% lv.names) &
-                     partable$group == g)
+                     partable$block == g)
         out[idx] <- out[idx] / OV[ match(partable$lhs[idx], ov.names) ]
 
         # 4d. "~*~" scales
         idx <- which(partable$op == "~*~" & !(partable$lhs %in% lv.names) &
-                     partable$group == g)
+                     partable$block == g)
         out[idx] <- 1.0
     }
 
@@ -550,45 +579,35 @@ unstandardize.est.ov <- function(partable, ov.var=NULL, cov.std=TRUE) {
     if(is.null(partable$ustart)) 
         partable$ustart <- partable$est
   
-    # check if group is missing
-    if(is.null(partable$group)) 
-        partable$group <- rep(1L, length(partable$ustart))
+    # check if block is missing
+    if(is.null(partable$block))  {
+        partable$block <- rep(1L, length(partable$ustart))
+    }
 
     stopifnot(!any(is.na(partable$ustart)))
     est <- out <- partable$ustart
     N <- length(est)
 
-    # ngroup
-    if(is.null(partable$group)) {
-        partable$group <- rep(1L, length(partable$lhs))
-        ngroups <- 1L
-    } else {
-        if(is.character(partable$group)) {
-            group.label <- unique(partable$group)
-            group.label <- group.label[ nchar(group.label) > 0L ]
-            ngroups <- length(group.label)
-        } else {
-            ngroups <- max(partable$group)
-        }
-    }
+    # nblocks
+    nblocks <- lav_partable_nblocks(partable)
 
     # if ov.var is NOT a list, make a list
     if(!is.list(ov.var)) {
         tmp <- ov.var
-        ov.var <- vector("list", length=ngroups)
-        ov.var[1:ngroups] <- list(tmp)
+        ov.var <- vector("list", length=nblocks)
+        ov.var[1:nblocks] <- list(tmp)
     }
 
-    for(g in 1:ngroups) {
+    for(g in 1:nblocks) {
 
-        ov.names <- vnames(partable, "ov", group=g) # not user
-        lv.names <- vnames(partable, "lv", group=g)
+        ov.names <- vnames(partable, "ov", block = g) # not user
+        lv.names <- vnames(partable, "lv", block = g)
 
         OV  <- sqrt(ov.var[[g]])
 
         # 1a. "=~" regular indicators
         idx <- which(partable$op == "=~" & !(partable$rhs %in% lv.names) &
-                     partable$group == g)
+                     partable$block == g)
         out[idx] <- out[idx] * OV[ match(partable$rhs[idx], ov.names) ]
 
         # 1b. "=~" regular higher-order lv indicators
@@ -596,17 +615,17 @@ unstandardize.est.ov <- function(partable, ov.var=NULL, cov.std=TRUE) {
         # 1c. "=~" indicators that are both in ov and lv
         #idx <- which(partable$op == "=~" & partable$rhs %in% ov.names
         #                             & partable$rhs %in% lv.names &
-        #             partable$group == g)
+        #             partable$block == g)
 
         # 2. "~" regressions (and "<~")
         idx <- which((partable$op == "~" | partable$op == "<~") & 
                      partable$lhs %in% ov.names &
-                     partable$group == g)
+                     partable$block == g)
         out[idx] <- out[idx] * OV[ match(partable$lhs[idx], ov.names) ]
 
         idx <- which((partable$op == "~" | partable$op == "<~") & 
                      partable$rhs %in% ov.names &
-                     partable$group == g)
+                     partable$block == g)
         out[idx] <- out[idx] / OV[ match(partable$rhs[idx], ov.names) ]
 
         # 3a. "~~" ov
@@ -620,14 +639,14 @@ unstandardize.est.ov <- function(partable, ov.var=NULL, cov.std=TRUE) {
         # variances
         rv.idx <- which(partable$op == "~~" & !(partable$lhs %in% lv.names) & 
                         partable$lhs == partable$rhs &
-                        partable$group == g)
+                        partable$block == g)
         out[rv.idx] <- ( out[rv.idx] * OV[ match(partable$lhs[rv.idx], ov.names) ]
                                      * OV[ match(partable$rhs[rv.idx], ov.names) ] )
 
         # covariances
         idx <- which(partable$op == "~~" & !(partable$lhs %in% lv.names) &
                      partable$lhs != partable$rhs &
-                     partable$group == g)
+                     partable$block == g)
         if(length(idx) > 0L) {
             if(cov.std == FALSE) {
                 out[idx] <- ( out[idx] * OV[ match(partable$lhs[idx], ov.names) ]
@@ -646,16 +665,16 @@ unstandardize.est.ov <- function(partable, ov.var=NULL, cov.std=TRUE) {
 
         # 3b. "~~" lv
         #idx <- which(partable$op == "~~" & partable$rhs %in% lv.names &
-        #             partable$group == g)
+        #             partable$block == g)
 
         # 4a. "~1" ov
         idx <- which(partable$op == "~1" & !(partable$lhs %in% lv.names) &
-                     partable$group == g)
+                     partable$block == g)
         out[idx] <- out[idx] * OV[ match(partable$lhs[idx], ov.names) ]
 
         # 4b. "~1" lv
         #idx <- which(partable$op == "~1" & partable$lhs %in% lv.names &
-        #             partable$group == g)
+        #             partable$block == g)
 
     }
 
