@@ -1,19 +1,26 @@
 # factor score regression
 
-# three methods:
+# four methods:
 #  - naive (regression or Bartlett)
 #  - Skrondal & Laake (2001) (regression models only)
 #  - Croon (2002) (general + robust SE)
+#  - simple: always use Bartlett, replace var(f) by psi estimate
+#
+# TODO
+#  - Hishino & Bentler: this is simple + WLS
 
 fsr <- function(model      = NULL, 
                 data       = NULL, 
                 cmd        = "sem",
-                fsr.method = "Croon", 
-                fs.method  = "Bartlett", 
+                fsr.method = "Croon",
+                fs.method  = "Bartlett",
                 fs.scores  = FALSE,
+                mm.options = list(se = "standard", test = "standard"),
                 Gamma.NT   = TRUE,
                 lvinfo     = FALSE,
-                ...) {
+                mm.list    = NULL,
+                ...,
+                output     = "fsr") {
    
     # we need full data
     if(is.null(data)) {
@@ -29,6 +36,9 @@ fsr <- function(model      = NULL,
         fsr.method <- "skrondal.laake"
     } else if(fsr.method == "croon") {
         # nothing to do
+    } else if(fsr.method == "simple") {
+        # force fs.method to Bartlett!
+        fs.method <- "Bartlett"
     } else {
         stop("lavaan ERROR: invalid option for argument fsr.method: ",
              fsr.method)
@@ -44,13 +54,17 @@ fsr <- function(model      = NULL,
         stop("lavaan ERROR: invalid option for argument fs.method: ",
              fs.method)
     }
+
+    if(output %in% c("scores", "fs.scores", "fsr.scores")) {
+        fs.scores <- TRUE
+    }
     
     # dot dot dot
     dotdotdot <- list(...)
 
     # change 'default' values for fsr
     if(is.null(dotdotdot$se)) {
-        dotdotdot$se <- "none"
+        dotdotdot$se <- "robust.sem"
     }
     if(is.null(dotdotdot$test)) {
         dotdotdot$test <- "satorra.bentler"
@@ -82,8 +96,9 @@ fsr <- function(model      = NULL,
     # restore
     lavoptions$se   <- dotdotdot$se
     lavoptions$test <- dotdotdot$test
-    ngroups    <- lavInspect(FIT, "ngroups")
-    lavpta     <- FIT@pta
+    ngroups     <- lavInspect(FIT, "ngroups")
+    lavpta      <- FIT@pta
+    lavpartable <- FIT@ParTable
 
     # FIXME: not ready for multiple groups yet
     if(ngroups > 1L) {
@@ -100,8 +115,24 @@ fsr <- function(model      = NULL,
 
     # any `regular' latent variables?
     lv.names <- unique(unlist(FIT@pta$vnames$lv.regular))
+    ov.names <- unique(unlist(FIT@pta$vnames$ov))
+
+    # check for higher-order factors
+    good.idx <- logical( length(lv.names) )
+    for(f in seq_len(length(lv.names))) {
+        # check the indicators
+        FAC <- lv.names[f]
+        IND <- lavpartable$rhs[ lavpartable$lhs == FAC &
+                                lavpartable$op  == "=~" ]
+        if(all(IND %in% ov.names)) {
+            good.idx[f] <- TRUE
+        }
+        # FIXME: check for mixed lv/ov indicators
+    }
+    lv.names <- lv.names[ good.idx ]
+
     if(length(lv.names) == 0L) {
-        stop("lavaan ERROR: model does not contain any latent variables")
+        stop("lavaan ERROR: model does not contain any (measured) latent variables")
     }
     nfac     <- length(lv.names)
 
@@ -110,11 +141,12 @@ fsr <- function(model      = NULL,
     PT$est <- PT$se <- NULL
 
     # find the structural regressions in the parameter table
-    eqs.idx <- which(PT$op == "~" & (PT$lhs %in% lv.names |
-                                     PT$rhs %in% lv.names))
-    if(length(eqs.idx) == 0L) {
-        stop("lavaan ERROR: regressions do not involve any latent variables")
-    }
+    #eqs.idx <- which(PT$op == "~" & (PT$lhs %in% lv.names |
+    #                                 PT$rhs %in% lv.names))
+    # FIXME: we should allow for just correlations too?
+    #if(length(eqs.idx) == 0L) {
+    #    stop("lavaan ERROR: regressions do not involve any latent variables")
+    #}
 
     # determine eqs.y and eqs.x names
     eqs.x.names <- unlist(FIT@pta$vnames$eqs.x)
@@ -129,8 +161,33 @@ fsr <- function(model      = NULL,
     }
 
 
-    # STEP 1a: compute factor scores for each latent variable
+    # STEP 1a: compute factor scores for each measurement model (block)
 
+    # how many measurement models?
+    if(!is.null(mm.list)) {
+
+        if(fsr.method != "simple") {
+            stop("lavaan ERROR: mm.list only available if fsr.method = \"simple\"")
+        }
+
+        nblocks <- length(mm.list)
+        # check each measurement block
+        for(b in seq_len(nblocks)) {
+            if(!all(mm.list[[b]] %in% lv.names)) {
+              stop("lavaan ERROR: mm.list contains unknown latent variable(s):", 
+                paste( mm.list[[b]][ mm.list[[b]] %in% lv.names ], sep = " "),
+                "\n")
+            }
+        }
+    } else {
+        # TODO: here comes the automatic 'detection' of linked
+        #       measurement models
+        #
+        # for now we take a single latent variable per measurement model block
+        mm.list <- as.list(lv.names)
+        nblocks <- length(mm.list)
+    }
+   
     # compute factor scores, per latent variable
     FS.SCORES  <- vector("list", length = ngroups)
     LVINFO     <- vector("list", length = ngroups)
@@ -138,10 +195,10 @@ fsr <- function(model      = NULL,
         names(FS.SCORES) <- names(LVINFO) <- lavInspect(FIT, "group.label")
     }
     for(g in 1:ngroups) {
-        FS.SCORES[[g]]  <- vector("list", length = nfac)
-        names(FS.SCORES[[g]]) <-  lv.names
-        LVINFO[[g]] <- vector("list", length = nfac)
-        names(LVINFO[[g]]) <- lv.names
+        FS.SCORES[[g]]  <- vector("list", length = nblocks)
+        #names(FS.SCORES[[g]]) <-  lv.names
+        LVINFO[[g]] <- vector("list", length = nblocks)
+        #names(LVINFO[[g]]) <- lv.names
     }
 
     # adjust options
@@ -151,23 +208,30 @@ fsr <- function(model      = NULL,
     dotdotdot2$debug <- FALSE
     dotdotdot2$verbose <- FALSE
     dotdotdot2$auto.cov.lv.x <- TRUE # allow correlated exogenous factors
+
+    # override with mm.options
+    dotdotdot2 <- modifyList(dotdotdot2, mm.options)
  
     # we assume the same number/names of lv's per group!!!
-    for(f in 1:nfac) {
+    MM.FIT <- vector("list", nblocks)
+    for(b in 1:nblocks) {
 
-        # create parameter table for this factor only
-        PT.1fac <- lav_partable_subset_measurement_model(PT = PT,
-                                                         lavpta = lavpta,
-                                                         lv.names = lv.names[f])
+        # create parameter table for this measurement block only
+        PT.block <- 
+            lav_partable_subset_measurement_model(PT = PT,
+                                                  lavpta = lavpta,
+                                                  add.lv.cov = TRUE,
+                                                  lv.names = mm.list[[b]])
         # fit 1-factor model
-        fit.1fac <- do.call("lavaan",
-                            args =  c(list(model  = PT.1fac,
+        fit.block <- do.call("lavaan",
+                            args =  c(list(model  = PT.block,
                                            data   = data), dotdotdot2) )
+        MM.FIT[[b]] <- fit.block
 
         # fs.method?
         if(fsr.method == "skrondal.laake") {
             # dependent -> Bartlett
-            if(lv.names[f] %in% eqs.y.names) {
+            if(lv.names[b] %in% eqs.y.names) {
                 fs.method <- "Bartlett"
             } else {
                 fs.method <- "regression"
@@ -175,28 +239,36 @@ fsr <- function(model      = NULL,
         }
 
         # compute factor scores
-        if(fsr.method %in% c("croon") || 
-           lavoptions$se == "robust.sem") {
-            SC <- lav_predict_eta(fit.1fac, method = fs.method, fsm = TRUE)
+        if(fsr.method %in% c("croon", "simple") ||
+           lavoptions$se == "robust.sem") { 
+            # we use lavPredict() here to remove unwanted dummy lv's, if any
+            SC <- lavPredict(fit.block, method = fs.method, fsm = TRUE)
             FSM <- attr(SC, "fsm"); attr(SC, "fsm") <- NULL
-            LAMBDA <- computeLAMBDA(fit.1fac@Model)
-            THETA  <- computeTHETA(fit.1fac@Model)
+            LAMBDA <- computeLAMBDA(fit.block@Model) # FIXME: remove dummy lv's?
+            THETA  <- computeTHETA(fit.block@Model)  # FIXME: remove not used ov?
+            PSI <- computeVETA(fit.block@Model)
         } else {
-            SC <- lav_predict_eta(fit.1fac, method = fs.method, fsm = FALSE)
+            SC <- lavPredict(fit.block, method = fs.method, fsm = FALSE)
         }
+        # if ngroups = 1, make list again
+        if(ngroups == 1L) {
+            # because lavPredict() drops the list
+            SC <- list(SC)
+        }
+  
 
         # store results
         for(g in 1:ngroups) {
-            FS.SCORES[[g]][[f]] <- SC[[g]]
-            if(fsr.method %in% c("croon") ||
-               lavoptions$se == "robust.sem") {
-                LVINFO[[g]][[f]] <- list(fsm = FSM[[g]], lambda = LAMBDA[[g]],
-                                                         theta  = THETA[[g]])
+            FS.SCORES[[g]][[b]] <- SC[[g]]
+            if(fsr.method %in% c("croon", "simple")) {
+                LVINFO[[g]][[b]] <- list(fsm = FSM[[g]], 
+                                         lambda = LAMBDA[[g]],
+                                         psi    = PSI[[g]],
+                                         theta  = THETA[[g]])
             }
         } # g
 
-    } # nfac
-
+    } # measurement block
 
     # cbind factor scores
     FS.SCORES <- lapply(1:ngroups, function(g) {
@@ -224,14 +296,21 @@ fsr <- function(model      = NULL,
          FSR.COV <- lav_fsr_croon_correction(FS.COV    = FS.COV,
                                              LVINFO    = LVINFO,
                                              fs.method = fs.method)
+    } else if(fsr.method == "simple") {
+        FSR.COV <- lav_fsr_simple_correction(FS.COV    = FS.COV,
+                                             LVINFO    = LVINFO,
+                                             mm.list   = mm.list,
+                                             force.pd  = FALSE)
     } else {
         FSR.COV <- FS.COV 
     }
 
+    
+
     # STEP 1c: do we need full set of factor scores?
     if(fs.scores) {
         # transform?
-        if(fsr.method == "croon") {
+        if(fsr.method %in% c("croon", "simple")) {
             for(g in 1:ngroups) {
                 OLD.inv <- solve(FS.COV[[g]])
                 OLD.inv.sqrt <- lav_matrix_symmetric_sqrt(OLD.inv)
@@ -259,12 +338,23 @@ fsr <- function(model      = NULL,
 
     PT.PA <- lav_partable_subset_structural_model(PT, lavpta = lavpta)
 
-    # free all means/intercepts
-    int.idx <- which(PT.PA$op == "~1")
+    # free all means/intercepts (of observed variables only)
+    lv.names.pa <- lavNames(PT.PA, "lv")
+    int.idx <- which(PT.PA$op == "~1" & !PT.PA$lhs %in% lv.names.pa)
     PT.PA$free[int.idx] <- 1L
     PT.PA$ustart[int.idx] <- NA
-
     # adjust lavoptions
+
+    #if(is.null(dotdotdot$missing)) {
+    #    lavoptions$missing <- "listwise" # factor scores are always complete
+    #} else {
+    #    lavoptions$missing <- dotdotdot$missing
+    #}
+    #if(is.null(dotdotdot$information)) {
+    #    lavoptions$information <- "expected"
+    #} else {
+    #    lavoptions$information <- dotdotdot$information
+    #}
     if(is.null(dotdotdot$do.fit)) {
         lavoptions$do.fit <- TRUE
     } else {
@@ -285,6 +375,7 @@ fsr <- function(model      = NULL,
     } else {
         lavoptions$sample.cov.rescale <- dotdotdot$sample.cov.rescale
     }
+
     # take care of NACOV, in case we want correct standard errors
     if(lavoptions$se == "robust.sem") {
         Omega.f <- vector("list", length = ngroups)
@@ -382,8 +473,6 @@ fsr <- function(model      = NULL,
                    FSR.COV = FSR.COV,
                    LVINFO = LVINFO)
 
-    PE <- parameterEstimates(fit, add.attributes = TRUE)
-
     # standard errors
     #lavsamplestats <- fit@SampleStats
     #lavsamplestats@NACOV <- Omega.f
@@ -400,14 +489,33 @@ fsr <- function(model      = NULL,
     #    }
     #}
    
-    out <- list(header = "This is fsr (0.1) -- factor score regression.",
-                PE = PE)
+    if(output == "fsr") {
+        #PE <- parameterEstimates(fit, add.attributes = TRUE, ci = FALSE)
+        HEADER <- paste("This is fsr (0.2) -- factor score regression using ",
+                        "fsr.method = ", fsr.method, sep = "")
+        out <- list(header = HEADER, MM.FIT = MM.FIT, STRUC.FIT = fit)
+        if(lvinfo) {
+            out$lvinfo <- extra
+        }
 
-    if(lvinfo) {
-        out$lvinfo <- extra
-    }
+        class(out) <- c("lavaan.fsr", "list")
 
-    class(out) <- c("lavaan.fsr", "list")
+    } else if(output %in% c("lavaan", "fit")) {
+        out <- fit
+    } else if(output == "extra") {
+        out <- extra
+    } else if(output == "lvinfo") {
+        out <- LVINFO
+    } else if(output %in% c("scores", "f.scores", "fs.scores")) {
+        out <- FS.SCORES
+    } else if(output %in% c("FSR.COV", "fsr.cov", "croon", "cov.croon", 
+                            "croon.cov", "COV", "cov")) {
+        out <- FSR.COV
+    } else if(output %in% c("FS.COV", "fs.cov")) {
+        out <- FS.COV
+    } else {
+        stop("lavaan ERROR: unknown output= argument: ", output)
+    } 
 
     out
 }
